#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""FAL command for remotelink3 package."""

import os
import time

from ionas.request import (XmlConfNode, Entry,)
from iobase.exception import (InvalidArgError, NotFoundError,)
from iobase.xmldata import (ANY_CHILD_NAME,)

from libfal import (falcmd, util)
from libfal.util import (RequestError,)
from fal_info import (ReqTypes, Results)
from pal_info import (PAL_CONFFILE_PATH, PAL_REMOTELINK3_PATH,
                        PAL_REMOTEACCESS_PATH, PAL_QRCODE_PATH,
                        PAL_SYSINFO_PATH, PAL_MAILSEND_PATH,)
from product_conf import (TEMP_DIR,)

from fal_config import (SHARES_CONF_FILE, NOTICE_CONF_FILE,)
from pkgaddon.fal_remotelink3_config import (CONFIG_FILE, REMOTELINK3_SERVICE,
                                                PINCODE_26_DIGIT_NUM,
                                                SEND_PINCODE_MAIL_XML_TMPL,
                                                NORMAL_PINCODE_TYPE,
                                                CLIENTLIMIT_PINCODE_TYPE,
                                                QR_CODE_IMAGE_TYPE,
                                                QR_CODE_IMAGE_SUBTYPE,)
from remotelink3.pincode import (to_qrcode_str,)

# Add the service entry for the eset package
util.config.FUNC_TO_SERVICE_MAP[REMOTELINK3_SERVICE] = REMOTELINK3_SERVICE


REMOTELINK3_NAME = "remotelink3"
PINCODE_NAME = "pincode"
REMOTEACCESS_NAME = "remoteaccess"
IOPORTAL_NAME = "ioportal"
ACCOUNT_NAME_PATH = "account/name"
KEY_NAME = "key"
TIMELIMITLIST_NAME = "timelimitlist"
TIMELIMIT_NAME = "timelimit"
TIMELIMIT_PATH = TIMELIMITLIST_NAME + "/" + TIMELIMIT_NAME
CLIENTLIMITLIST_NAME = "clientlimitlist"
CLIENTLIMIT_NAME = "clientlimit"
CLIENTLIMIT_PATH = CLIENTLIMITLIST_NAME + "/" + CLIENTLIMIT_NAME
CONNECTIONLIST_NAME = "connectionlist"

TIMELIMIT_PINCODE_PREFIX = "RLFT."
CLIENTLIMIT_PINCODE_PREFIX = "RLFD."

APPLY_TAG_LIST = ["terms", REMOTEACCESS_NAME, "connection", "ioportal"]
DELETE_INFO_LIST_MAP = {
        REMOTEACCESS_NAME: [("sharelist", "*")],
    }

SHARE_PATH = "remoteaccess/sharelist/share"
NAME_NAME = "name"
SHARE_PATH_XPATH_TMPL = """sharelist/share/name[text()="%s"]/../path"""

SYSINFO_NAME = "sysinfo"
SYSINFO_MODEL_PATH = "model"
SYSINFO_NAME_PATH = "hostdesc/name"

QR_MODEL_TMPL = " (%s)"

QRCODE_PATH = "qrcode"
QRCODE_DIRECTORY_PATH = "qrcode/directory"
QRCODE_PINCODE_PATH = "qrcode/pincode"

QRCODE_REQ_TMPL = "<qrcode><data>%s</data></qrcode>"
QRCODE_FILE = "pincode_qr.png"
FILENAME_XML_TMPL = "<filename>%s</filename>"

MAILADDR_NAME = "mailaddr"
MAILCONF_PATH = "mailbasic/mailconf"

TOADDR_XML_TMPL = "<toaddrlist><toaddr>%s</toaddr></toaddrlist>"

KEY_LEN = 4

WAIT_SERVICE_TIME = 2

GET_PINCODE_REQ = "get_pincode"
CHANGE_PINCODE_REQ = "change_pincode"
DELETE_PINCODE_REQ = "delete_pincode"
GET_TIMELIMIT_LIST_REQ = "get_timelimit_list"
GET_CLIENTLIMIT_LIST_REQ = "get_clientlimit_list"
GET_TIMELIMIT_REQ = "get_timelimit"
ADD_TIMELIMIT_REQ = "add_timelimit"
MODIFY_TIMELIMIT_REQ = "modify_timelimit"
DELETE_TIMELIMIT_REQ = "delete_timelimit"
ADD_CLIENTLIMIT_REQ = "add_clientlimit"
MODIFY_CLIENTLIMIT_REQ = "modify_clientlimit"
DELETE_CLIENTLIMIT_REQ = "delete_clientlimit"
CREATE_QRCODE_REQ = "create_qrcode"
SEND_PINCODE_REQ = "send_pincode"
GET_CONNECTION_LIST_REQ = "get_connection_list"
MODIFY_CONNECTION_REQ = "modify_connection"
DELETE_CONNECTION_REQ = "delete_connection"


class NoPinCodeError(Exception):
    """No Pincode Error."""

    pass


class FALRemoteLink3(falcmd.FalCmd):
    """FAL remotelink3 command implementation class."""

    def _check_body_node(self, node, type=None):
        """Check if entry's body node is valid."""

        if node.get_node_name() != REMOTELINK3_NAME:
            raise InvalidArgError(node.get_node_name())

    def _send_request(self, target, entry, path=None):
        """Send "entry" request to "target" PAL commands."""

        if path:
            response_entry = self.send_entry_with_file(entry, target, path)
        else:
            response_entry = self.send_entry(entry, target)

        if response_entry.get_result() != Results.OK:
            raise RequestError(response_entry.get_result())

        return response_entry

    def _send_to_pal_remoteaccess(self, base_entry):
        """Send request base on "base_entry" to pal_remoteaccess."""

        req_entry = base_entry.get_clone()
        req_entry.get().set_node_name(REMOTEACCESS_NAME)
        return self._send_request(PAL_REMOTEACCESS_PATH, req_entry)

    def _delete_all_timelimit(self):
        """Delete all timelimit entry."""

        req_entry = Entry(GET_TIMELIMIT_LIST_REQ, REMOTEACCESS_NAME)
        res_entry = self._send_request(PAL_REMOTEACCESS_PATH, req_entry)
        for timelimit in res_entry.get().get_node_list(TIMELIMIT_PATH):
            del_req_entry = Entry(DELETE_TIMELIMIT_REQ, REMOTEACCESS_NAME)
            del_req_entry.get().append_child(timelimit)
            self._send_request(PAL_REMOTEACCESS_PATH, del_req_entry)

    def _delete_all_clientlimit(self):
        """Delete all clientlimit entry."""

        req_entry = Entry(GET_CLIENTLIMIT_LIST_REQ, REMOTEACCESS_NAME)
        res_entry = self._send_request(PAL_REMOTEACCESS_PATH, req_entry)
        for clientlimit in res_entry.get().get_node_list(CLIENTLIMIT_PATH):
            del_req_entry = Entry(DELETE_CLIENTLIMIT_REQ, REMOTEACCESS_NAME)
            del_req_entry.get().append_child(clientlimit)
            self._send_request(PAL_REMOTEACCESS_PATH, del_req_entry)

    def _to_26_decimal_char(self, value):
        """Transform "value" to 26 decimal charactor."""

        return chr(ord("A") + int(value))

    def _encode_pincode(self, pincode):
        """Encode "pincode" for extended PIN code."""

        value = int(pincode, 16)
        encoded = ""
        for digit in range(PINCODE_26_DIGIT_NUM - 1, -1, -1):
            base_value = pow(26, digit)
            encoded += self._to_26_decimal_char(value / base_value)
            value %= base_value
        return encoded

    def _change_key_to_pincode(self, limit_node, is_timelimit):
        """Transform "key" tag to "pincode" in "limit_node."""

        res_entry = self._send_request(PAL_REMOTELINK3_PATH,
                            Entry(GET_PINCODE_REQ, REMOTELINK3_NAME))
        pincode = res_entry.get().get_value(PINCODE_NAME)
        if not pincode:
            raise NoPinCodeError

        old_key_node = limit_node.get_node(KEY_NAME)
        key = old_key_node.get_text()

        ext_pincode = CLIENTLIMIT_PINCODE_PREFIX
        if is_timelimit:
            ext_pincode = TIMELIMIT_PINCODE_PREFIX

        ext_pincode += self._encode_pincode(pincode) + key

        old_key_node.set_node_name(PINCODE_NAME)
        old_key_node.set_text(ext_pincode)

    def _change_pincode_to_key(self, limit_node):
        """Transform "pincode" tag to "key" in "limit_node."""

        old_pincode_node = limit_node.get_node(PINCODE_NAME)
        pincode = old_pincode_node.get_text()

        old_pincode_node.set_node_name(KEY_NAME)
        old_pincode_node.set_text(pincode[-1 * KEY_LEN:])

    def _try_deleting_old_setting(self, old_tag):
        """Try deleting old setting in "old_tag"."""

        del_info_list = DELETE_INFO_LIST_MAP.get(old_tag.get_node_name(), [])
        for del_info in del_info_list:
            old_tag.delete_nodes(del_info[0], del_info[1])

    def _get_pal_apply_entry(self, entry):
        """Return entry for "apply" request to pal."""

        req_entry = entry.get_clone()
        share_config = util.load_config(SHARES_CONF_FILE)
        for share in req_entry.get().get_node_list(SHARE_PATH):
            try:
                path = share_config.get_node(
                            SHARE_PATH_XPATH_TMPL % share.get_value(NAME_NAME))
                share.append_child(path.get_clone())
            except:
                # Ignore error
                pass

        return req_entry

    def _get_qr_dev_name(self, sysinfo):
        """Return the DEVNAME data for QR code.
           None is returned if getting fail.
        """

        try:
            model = sysinfo.get_value(SYSINFO_MODEL_PATH)
            dev_name = sysinfo.get_value(SYSINFO_NAME_PATH)
            if model:
                dev_name += QR_MODEL_TMPL % (model,)
            if dev_name:
                return dev_name
        except:
            pass
        return None

    def _create_qr_code_image(self, pincode, image_path, sysinfo):
        """Create QR code image and save it to "image_path"."""

        user = None
        password = None
        dev_name = self._get_qr_dev_name(sysinfo)
        if pincode.startswith(TIMELIMIT_PINCODE_PREFIX):
            user = ""
            password = ""

        qrcode_str = to_qrcode_str(pincode, user, password, dev_name)

        self._send_request(PAL_QRCODE_PATH,
                Entry(ReqTypes.ENCODE, QRCODE_REQ_TMPL % (qrcode_str,)),
                image_path)

    def _send_mail_qrocde(self, pincode, mail_addr, host_name, image_path):
        """Send mail include "pincode" with QR code image file."""

        tmpl_dict = {
                "landisk_name": host_name,
                "mail_to": mail_addr,
                "pincode": pincode,
                "attach_type": QR_CODE_IMAGE_TYPE,
                "attach_subtype": QR_CODE_IMAGE_SUBTYPE,
                "attach_path": image_path,
        }

        if pincode.startswith(CLIENTLIMIT_PINCODE_PREFIX):
            tmpl_dict["pincode_type"] = CLIENTLIMIT_PINCODE_TYPE
        else:
            tmpl_dict["pincode_type"] = NORMAL_PINCODE_TYPE

        mailsend = XmlConfNode(SEND_PINCODE_MAIL_XML_TMPL % tmpl_dict)

        notice_config = util.load_config(NOTICE_CONF_FILE)
        mailconf = notice_config.get_node(MAILCONF_PATH)
        mailconf.append_child(XmlConfNode(TOADDR_XML_TMPL % (mail_addr,)))
        mailsend.append_child(mailconf)

        self._send_request(PAL_MAILSEND_PATH, Entry(ReqTypes.SEND, mailsend))

    def _change_pincode_handler(self, entry):
        """Handle change_pincode request."""

        self._check_body_node(entry.get())

        try:
            self._delete_all_timelimit()
        except:
            self.exception("_change_pincode_handler: delete timelimit error")
            entry.set_result(Results.CHANGE_ERROR)
            return entry

        try:
            self._delete_all_clientlimit()
        except:
            self.exception("_change_pincode_handler: delete clientlimit error")
            entry.set_result(Results.CHANGE_ERROR)
            return entry

        config = util.load_config(CONFIG_FILE)
        request_entry = entry.get_clone()
        request_entry.get().append_child(config.get_node(IOPORTAL_NAME))
        response_entry = self.send_entry(request_entry, PAL_REMOTELINK3_PATH)
        entry.set_result(response_entry.get_result())
        util.servicectrl_suspend([], [REMOTELINK3_SERVICE])
        util.servicectrl_resume([], [REMOTELINK3_SERVICE])
        return entry

    def _delete_pincode_handler(self, entry):
        """Handle delete_pincode request."""

        self._check_body_node(entry.get())

        config = util.load_config(CONFIG_FILE)
        request_entry = entry.get_clone()
        request_entry.get().append_child(config.get_node(IOPORTAL_NAME))
        response_entry = self.send_entry(request_entry, PAL_REMOTELINK3_PATH)
        util.servicectrl_suspend([], [REMOTELINK3_SERVICE])
        util.servicectrl_resume([], [REMOTELINK3_SERVICE])
        entry.set_result(response_entry.get_result())
        return entry

    def _get_timelimit_list_handler(self, entry):
        """Handle get_timelimit_list request."""

        self._check_body_node(entry.get())
        try:
            res_entry = self._send_to_pal_remoteaccess(entry)
            for timelimit in res_entry.get().get_node_list(TIMELIMIT_PATH):
                self._change_key_to_pincode(timelimit, True)

            entry.get().append_child(
                            res_entry.get().get_node(TIMELIMITLIST_NAME))
            entry.set_result(Results.OK)
        except:
            self.exception("get_timelimit_list fail")
            entry.set_result(Results.GET_ERROR)
        return entry

    def _get_clientlimit_list_handler(self, entry):
        """Handle get_clientlimit_list request."""

        self._check_body_node(entry.get())
        try:
            res_entry = self._send_to_pal_remoteaccess(entry)
            for clientlimit in res_entry.get().get_node_list(CLIENTLIMIT_PATH):
                self._change_key_to_pincode(clientlimit, False)

            entry.get().append_child(
                            res_entry.get().get_node(CLIENTLIMITLIST_NAME))
            entry.set_result(Results.OK)
        except:
            self.exception("get_clientlimit_list fail")
            entry.set_result(Results.GET_ERROR)
        return entry

    def _get_timelimit_handler(self, entry):
        """Handle get_timelimit request."""

        self._check_body_node(entry.get())
        try:
            req_entry = entry.get_clone()
            try:
                self._change_pincode_to_key(
                        req_entry.get().get_node(TIMELIMIT_NAME))
            except:
                # username and pathlist may be used
                pass

            res_entry = self._send_to_pal_remoteaccess(req_entry)
            timelimit = res_entry.get().get_node(TIMELIMIT_NAME)
            self._change_key_to_pincode(timelimit, True)

            entry.get().delete_children(ANY_CHILD_NAME)
            entry.get().append_child(timelimit)
            entry.set_result(Results.OK)
        except:
            self.exception("get_timelimit fail")
            entry.set_result(Results.GET_ERROR)
        return entry

    def _add_timelimit_handler(self, entry):
        """Handle add_timelimit request."""

        self._check_body_node(entry.get())
        try:
            res_entry = self._send_to_pal_remoteaccess(entry)
            res_timelimit = res_entry.get().get_node(TIMELIMIT_NAME)

            timelimit = entry.get().get_node(TIMELIMIT_NAME)
            timelimit.append_child(res_timelimit.get_node(KEY_NAME))
            self._change_key_to_pincode(timelimit, True)

            entry.set_result(Results.OK)
        except:
            self.exception("add_timelimit fail")
            entry.set_result(Results.ADD_ERROR)
        return entry

    def _modify_timelimit_handler(self, entry):
        """Handle modify_timelimit request."""

        self._check_body_node(entry.get())
        try:
            req_entry = entry.get_clone()
            self._change_pincode_to_key(
                    req_entry.get().get_node(TIMELIMIT_NAME))

            self._send_to_pal_remoteaccess(req_entry)
            entry.set_result(Results.OK)
        except:
            self.exception("modify_timelimit fail")
            entry.set_result(Results.ADD_ERROR)
        return entry

    def _delete_timelimit_handler(self, entry):
        """Handle delete_timelimit request."""

        self._check_body_node(entry.get())
        try:
            req_entry = entry.get_clone()
            self._change_pincode_to_key(
                    req_entry.get().get_node(TIMELIMIT_NAME))

            self._send_to_pal_remoteaccess(req_entry)
            entry.set_result(Results.OK)
        except:
            self.exception("delete_timelimit fail")
            entry.set_result(Results.ADD_ERROR)
        return entry

    def _add_clientlimit_handler(self, entry):
        """Handle add_clientlimit request."""

        self._check_body_node(entry.get())
        try:
            res_entry = self._send_to_pal_remoteaccess(entry)
            res_clientlimit = res_entry.get().get_node(CLIENTLIMIT_NAME)

            clientlimit = entry.get().get_node(CLIENTLIMIT_NAME)
            clientlimit.append_child(res_clientlimit.get_node(KEY_NAME))
            self._change_key_to_pincode(clientlimit, False)

            entry.set_result(Results.OK)
        except:
            self.exception("add_clientlimit fail")
            entry.set_result(Results.ADD_ERROR)
        return entry

    def _modify_clientlimit_handler(self, entry):
        """Handle modify_clientlimit request."""

        self._check_body_node(entry.get())
        try:
            req_entry = entry.get_clone()
            self._change_pincode_to_key(
                    req_entry.get().get_node(CLIENTLIMIT_NAME))

            self._send_to_pal_remoteaccess(req_entry)
            entry.set_result(Results.OK)
        except:
            self.exception("modify_clientlimit fail")
            entry.set_result(Results.ADD_ERROR)
        return entry

    def _delete_clientlimit_handler(self, entry):
        """Handle delete_clientlimit request."""

        self._check_body_node(entry.get())
        try:
            req_entry = entry.get_clone()
            self._change_pincode_to_key(
                    req_entry.get().get_node(CLIENTLIMIT_NAME))

            self._send_to_pal_remoteaccess(req_entry)
            entry.set_result(Results.OK)
        except:
            self.exception("delete_clientlimit fail")
            entry.set_result(Results.ADD_ERROR)
        return entry

    def _apply_handler(self, entry):
        """Handle apply request."""

        self._check_body_node(entry.get(), ReqTypes.APPLY)
        try:
            config = util.load_config(CONFIG_FILE)
            new_config = entry.get()

            is_service_enabled = None
            do_create_pincode = False
            util.servicectrl_suspend([], [REMOTELINK3_SERVICE])
            for tag_name in APPLY_TAG_LIST:
                try:
                    new_target = new_config.get_node(tag_name)
                    if tag_name == REMOTEACCESS_NAME:
                        is_service_enabled = new_target.is_enabled()

                    try:
                        target = config.get_node(tag_name)
                        if tag_name == IOPORTAL_NAME:
                            if (not target.get_value(ACCOUNT_NAME_PATH, "") and
                                new_target.get_value(ACCOUNT_NAME_PATH, "")):

                                do_create_pincode = True

                        self._try_deleting_old_setting(target)
                        target.update(new_target, override=True)
                    except NotFoundError:
                        config.append_child(new_target.get_clone())
                except:
                    # Not change setting about "tag_name" tag entry.
                    continue

            if do_create_pincode:
                is_service_enabled = True
                config.get_node(REMOTEACCESS_NAME).enable()

            self._send_request(PAL_REMOTELINK3_PATH,
                                self._get_pal_apply_entry(entry))
            util.servicectrl_resume([], [REMOTELINK3_SERVICE])

            if is_service_enabled is True:
                util.servicectrl_start([], [REMOTELINK3_SERVICE])
            elif is_service_enabled is False:
                util.servicectrl_stop([], [REMOTELINK3_SERVICE])

            if do_create_pincode:
                time.sleep(WAIT_SERVICE_TIME)
                request_entry = Entry(CHANGE_PINCODE_REQ, REMOTELINK3_NAME)
                request_entry.get().append_child(
                                config.get_node(IOPORTAL_NAME).get_clone())
                self._send_request(PAL_REMOTELINK3_PATH, request_entry)
                util.servicectrl_suspend([], [REMOTELINK3_SERVICE])
                util.servicectrl_resume([], [REMOTELINK3_SERVICE])

            util.save_config(CONFIG_FILE, config)
            entry.set_result(Results.OK)
        except RequestError, err:
            self.exception("Request error")
            entry.set_result(err.args[0])
        except:
            self.exception("apply fail")
            entry.set_result(Results.APPLY_ERROR)

        return entry

    def _create_qrcode_handler(self, entry):
        """Handle create_qrcode request."""

        self._check_body_node(entry.get())
        try:
            remotelink3 = entry.get()
            directory = remotelink3.get_node(QRCODE_DIRECTORY_PATH).get_text()
            image_path = os.path.join(directory, QRCODE_FILE)
            pincode = remotelink3.get_node(QRCODE_PINCODE_PATH).get_text()

            sysinfo = self._send_request(PAL_SYSINFO_PATH,
                                    Entry(ReqTypes.QUERY, SYSINFO_NAME)).get()

            self._create_qr_code_image(pincode, image_path, sysinfo)

            remotelink3.get_node(QRCODE_PATH).append_child(
                            XmlConfNode(FILENAME_XML_TMPL % (QRCODE_FILE,)))
            entry.set_result(Results.OK)
        except:
            self.exception("Create QR Code fail!!")
            entry.set_result(Results.CREATE_ERROR)

        return entry

    def _send_pincode_handler(self, entry):
        """Handle send_pincode request."""

        self._check_body_node(entry.get())
        try:
            remotelink3 = entry.get()
            pincode = remotelink3.get_node(PINCODE_NAME).get_text()
            mail_addr = remotelink3.get_node(MAILADDR_NAME).get_text()
            image_path = os.path.join(TEMP_DIR, QRCODE_FILE)

            sysinfo = self._send_request(PAL_SYSINFO_PATH,
                                    Entry(ReqTypes.QUERY, SYSINFO_NAME)).get()

            self._create_qr_code_image(pincode, image_path, sysinfo)
        except:
            self.exception("Create QR Code fail!!")
            entry.set_result(Results.CREATE_ERROR)
            return entry

        try:
            host_name = sysinfo.get_value(SYSINFO_NAME_PATH)
            self._send_mail_qrocde(pincode, mail_addr, host_name, image_path)
            entry.set_result(Results.OK)
        except:
            self.exception("Send mail fail!!")
            entry.set_result(Results.SEND_ERROR)

        return entry

    def _get_connection_list_handler(self, entry):
        """Handle get_connection_list request."""

        self._check_body_node(entry.get())
        try:
            res_entry = self._send_to_pal_remoteaccess(entry)
            entry.get().append_child(
                            res_entry.get().get_node(CONNECTIONLIST_NAME))
            entry.set_result(Results.OK)
        except:
            self.exception("get_connection_list fail")
            entry.set_result(Results.GET_ERROR)
        return entry

    def _modify_connection_handler(self, entry):
        """Handle modify_connection request."""

        self._check_body_node(entry.get())
        try:
            res_entry = self._send_to_pal_remoteaccess(entry)
            entry.set_result(Results.OK)
        except:
            self.exception("modify_connection fail")
            entry.set_result(Results.MODIFY_ERROR)
        return entry

    def _delete_connection_handler(self, entry):
        """Handle get_connection_list request."""

        self._check_body_node(entry.get())
        try:
            res_entry = self._send_to_pal_remoteaccess(entry)
            entry.set_result(Results.OK)
        except:
            self.exception("delete_connection fail")
            entry.set_result(Results.DELETE_ERROR)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        falcmd.FalCmd.__init__(self, option=option)

        load_save_handler = lambda entry: self.send_entry_with_file(
            entry, PAL_CONFFILE_PATH, CONFIG_FILE)

        self.set_handler(ReqTypes.LOAD, load_save_handler)
        self.set_handler(ReqTypes.SAVE, load_save_handler)

        handler_pass_to_pal = lambda entry: \
                            self.send_entry(entry, PAL_REMOTELINK3_PATH)

        self.set_handler(ReqTypes.QUERY, handler_pass_to_pal)
        self.set_handler(GET_PINCODE_REQ, handler_pass_to_pal)

        self.set_handler(CHANGE_PINCODE_REQ, self._change_pincode_handler)
        self.set_handler(DELETE_PINCODE_REQ, self._delete_pincode_handler)

        self.set_handler(GET_TIMELIMIT_LIST_REQ,
                                self._get_timelimit_list_handler)
        self.set_handler(GET_CLIENTLIMIT_LIST_REQ,
                                self._get_clientlimit_list_handler)
        self.set_handler(GET_TIMELIMIT_REQ, self._get_timelimit_handler)
        self.set_handler(ADD_TIMELIMIT_REQ, self._add_timelimit_handler)
        self.set_handler(MODIFY_TIMELIMIT_REQ, self._modify_timelimit_handler)
        self.set_handler(DELETE_TIMELIMIT_REQ, self._delete_timelimit_handler)
        self.set_handler(ADD_CLIENTLIMIT_REQ,
                                self._add_clientlimit_handler)
        self.set_handler(MODIFY_CLIENTLIMIT_REQ,
                                self._modify_clientlimit_handler)
        self.set_handler(DELETE_CLIENTLIMIT_REQ,
                                self._delete_clientlimit_handler)

        self.set_handler(ReqTypes.APPLY, self._apply_handler)
        self.set_handler(CREATE_QRCODE_REQ, self._create_qrcode_handler)
        self.set_handler(SEND_PINCODE_REQ, self._send_pincode_handler)
        self.set_handler(GET_CONNECTION_LIST_REQ,
                            self._get_connection_list_handler)
        self.set_handler(MODIFY_CONNECTION_REQ,
                            self._modify_connection_handler)
        self.set_handler(DELETE_CONNECTION_REQ,
                            self._delete_connection_handler)


#
# Main
#

if __name__ == "__main__":
    FALRemoteLink3(option=falcmd.FalCmdOption()).handle_request()
