#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Define utility functions for PAL commands."""


from __future__ import with_statement

import os
import re

from iobase.cmdexec import (CmdExec, PIPE, NULL)
from iobase.exception import (NotFoundError)
import product_conf
import product_conf.devinfo as devinfo
import pal_config as config
import iobase.fileut as fileut


PROC_MOUNTS = os.path.join(product_conf.PROC_DIR, "mounts")
MNT_POINT_KEY = "mnt_point"
ALT_MNT_POINT_KEY = "alt_mnt_point"
MNT_OPT_SEP = ","
RO_OPT = "ro"
ENC_DEV_DIR_NAME = "mapper"


class _NotMatch:
    """Exception class for representing that "re.match()" return None."""

    pass


def _dev_to_port_info(dev_name, pattern):
    """Return tuple of port I/F name and number if "dev_name" match with
       "pattern".
    """

    match = re.match(pattern, dev_name.strip())
    if match == None:
        raise _NotMatch

    return (match.group(1), match.group(2))


# For public

class RemountRoot:
    """Class for remounting the root filesystem as read-only or write."""

    def __enter__(self):
        """Enter context for remounting the root filesystem as write."""

        self.read_write()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit context for remounting the root filesystem as write."""

        self.read_only()
        if exc_type != None:
            return False
        return True

    def read_only(self):
        """Remount the root filesystem as read-only."""

        if product_conf.REMOUNT_ROOT_READ_ONLY_CMD:
            CmdExec(product_conf.REMOUNT_ROOT_READ_ONLY_CMD,
                        stdout=NULL, stderr=NULL).execute()

    def read_write(self):
        """Remount the root filesystem as write."""

        if product_conf.REMOUNT_ROOT_WRITE_CMD:
            CmdExec(product_conf.REMOUNT_ROOT_WRITE_CMD,
                    stdout=NULL, stderr=NULL).execute()

    def remount(self, read_only=False):
        """Remount the root filesystem as "read_only" flag."""

        if read_only:
            self.read_only()
        else:
            self.read_write()


class RaidOperation(fileut.FLock):

    def __init__(self, op_value):

        self.__op_value = op_value
        fileut.FLock.__init__(self,
                              config.LOCK_DIR,
                              config.RAID_OPERATION_LOCK_FILE,
                              exclusive=True, timeout=None, suffix=".lock")

    def __enter__(self):

        fileut.FLock.__enter__(self)

        try:
            self.__set_raid_operation()
        except:
            pass

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        try:
            self.__clear_raid_operation()
        except:
            pass

        fileut.FLock.__exit__(self, exc_type, exc_val, exc_tb)

        if exc_type != None:
            return False
        return True

    def __set_raid_operation(self):

        self.__save_raid_operation(self.__op_value)
        return

    def __clear_raid_operation(self):

        self.__save_raid_operation(config.NON_RAID_OPERATION)
        return

    def __save_raid_operation(self, op_value):

        dir_path = os.path.dirname(config.RAID_CUR_OPERATION_FILE)
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

        with open(config.RAID_CUR_OPERATION_FILE, "w") as fh:
            fh.write(op_value)
        return


def dev_to_port_info(dev_name):
    """Return tuple of port I/F name and number."""

    dev_name = os.path.basename(dev_name)
    try:
        return _dev_to_port_info(dev_name, "(.+)d([0-9]+).*")
    except:
        pass

    try:
        return _dev_to_port_info(dev_name, "(.+)ups([0-9]+).*")
    except:
        pass

    try:
        return _dev_to_port_info(dev_name, "(.+)lp([0-9]+).*")
    except:
        pass

    return ("", "")


def path_to_vol(path):
    """Return the volume name which contains "path"."""

    for vol in product_conf.VOL_LIST:
        if (path.startswith(devinfo.VOL_INFO[vol][MNT_POINT_KEY]) or
            path.startswith(devinfo.VOL_INFO[vol][ALT_MNT_POINT_KEY])):

            return vol

    raise NotFoundError("no volume", path)


def get_mount_info(dev_or_mnt_point):
    """Return the tuple of mount information about "dev_or_mnt_point":
       (DEV, MOUNT_POINT, FS, MOUNT_OPTS).
       Note: "dev_or_mnt_point" must not include a space.
    """

    MOUNT_INFO_RE = "([^ ]+) +([^ ]+) +([^ ]+) +([^ ]+) +.*"
    if dev_or_mnt_point[0] != "/":
        # Assume that the device name is specified.
        dev_or_mnt_point = \
                    os.path.join(product_conf.DEV_DIR, dev_or_mnt_point)

    with open(PROC_MOUNTS) as proc_mounts:
        for line in proc_mounts:
            match = re.match(MOUNT_INFO_RE, line)
            if match == None:
                continue

            if (match.group(1) == dev_or_mnt_point or
                match.group(2) == dev_or_mnt_point):

                return match.group(1, 2, 3, 4)

    raise NotFoundError("for mount information", dev_or_mnt_point)


def is_set_to_read_only(mnt_opts):
    """Return if "mnt_opts" include "ro" option."""

    try:
        if RO_OPT in mnt_opts.split(MNT_OPT_SEP):
            return True
        return False
    except:
        return False


def is_mounted_as_read_only(dev_or_mnt_point):
    """Return if "dev_or_mnt_point" is mounted as read-only."""

    try:
        (dev, mnt_point, fs, opts) = get_mount_info(dev_or_mnt_point)
        return is_set_to_read_only(opts)
    except:
        return False


def is_mounted(dev_or_mnt_point):
    """Return if "dev_or_mnt_point" is mounted."""

    try:
        get_mount_info(dev_or_mnt_point)
        return True
    except:
        return False


def is_enc_dev_name(dev_name):

    return dev_name in devinfo.DEV_TO_ENC_DEV.values()


def symbl_dev_name2dev_path(dev_name):

    if is_enc_dev_name(dev_name):
        return os.path.join(product_conf.DEV_DIR, ENC_DEV_DIR_NAME, dev_name)

    return os.path.join(product_conf.DEV_DIR, dev_name)


def get_sata_disk_info(disk_path):

    MODEL_PREFIX_VENDOR_MAP = \
        {"ST": "Seagate",
         "WD": "Western Digital",
         "Hitachi": "Hitachi",
         "SAMSUNG": "Samsung"}

    KEY_RE_PAIR = \
        {"model": "Device Model:\s*(.*)",
         "firmware": "Firmware Version:\s*(.*)",
         "serialnumber": "Serial Number:\s*(.*)"}

    def model2vendor(model):
        for prefix, vendor in MODEL_PREFIX_VENDOR_MAP.iteritems():
            if model.startswith(prefix):
                return vendor
        return "Unknown"

    def get_smartctl_identify_output(disk_path):
        cmd = [product_conf.SMARTCTL_CMD, "-i", disk_path]
        proc = CmdExec(cmd, stdout=PIPE, stderr=NULL).execute()
        return proc.stdout

    info_dict = {}
    for line in get_smartctl_identify_output(disk_path):
        for key, re_str in KEY_RE_PAIR.iteritems():
            match = re.match(re_str, line)
            if not match:
                continue

            info_dict[key] = match.group(1)
            if "model" == key:
                info_dict["vendor"] = model2vendor(match.group(1))

    return info_dict


def get_cur_raid_operation():

    if not os.path.isfile(config.RAID_CUR_OPERATION_FILE):
        return config.NON_RAID_OPERATION

    try:
        with open(config.RAID_CUR_OPERATION_FILE, "r") as fh:
            return fh.read().rstrip()
    except:
        return config.NON_RAID_OPERATION


def get_cuurent_array_num(vol_name=config.INTERNAL_VOL):

    file_path = config.CUR_ARRAY_NUM_TMPL % vol_name

    if not os.path.isfile(file_path):
        return config.RAID_MODE_INFO["vfsext"]["array_num"]

    with open(file_path, "r") as fh:
        array_num = fh.read().rstrip()

    return int(array_num)


def get_raid_element_info(raid_mode=None):

    if raid_mode is None:
        # 現在の設定から
        return config.RAID_TO_ELEMENT[get_cuurent_array_num()]

    # 指定されたraid modeから
    array_num = config.RAID_MODE_INFO[raid_mode]["array_num"]
    return config.RAID_TO_ELEMENT[array_num]


def get_proxies(proxy_node):
    """Get proxies dic from XML proxy node."""

    proxies = None

    try:
        for server_node in proxy_node.get_node_list("serverlist/server"):
            try:
                host = server_node.get_value("host")
                port = server_node.get_value("port")
                proxies = {
                    "http": "http://%s:%u/" % (host, int(port)),
                    "https": "https://%s:%u/" % (host, int(port))}

                # only get first one
                break
            except:
                continue
    except:
        pass

    return proxies
