#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for windows azure blob storage."""
from __future__ import with_statement

import shutil
import threading

from iobase.cmdexec import (CmdExec, NULL)
from iobase.exception import ExecError
import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)

import pal_config as config

from azure.storage import BlobService, WindowsAzureError

SHARENAME_NAME = "sharename"
RESULT_NAME = "result"
ACCOUNTNAME_NAME = "accountname"
ACCESSKEY_NAME = "accesskey"
CONTAINER_NAME = "container"
PROXY_SERVER_PATH = "proxy/serverlist/server"
PROXYHOST_PATH = PROXY_SERVER_PATH + "/host"
PROXYPORT_PATH = PROXY_SERVER_PATH + "/port"
HTTPMETHOD_NAME = "httpmethod"

SEND_DOWNLOAD_SIGNAL_ARGS = [config.NASDSYNC_INITD, "reload"]

DOWNLOAD_FAILED_MSG = "Download Failed!!"
CHECK_FAILED_MSG = "Check Failed!!"

CONTAINER_NOT_FOUND_MSG = 'Not found (The specified container does not exist.)'
BLOB_NOT_FOUND_MSG = 'Not found (The specified blob does not exist.)'
FAIL_AUTH_MSG = 'Unknown error ' + \
            '(Server failed to authenticate the request.' + \
            ' Make sure the value of Authorization header is' + \
            ' formed correctly including the signature.)'
INCORRECT_PADDING_MSG = 'Incorrect padding'

SUCCESS_RESULT = "Success"
FAILED_RESULT = "Failed"
TIMEOUT_RESULT = "TimeOut"
NO_SUCH_CONTAINER_RESULT = "NoSuchContainer"
INVALID_ACCESS_KEY_RESULT = "InvalidAccessKey"

ERROR_TABLE = {CONTAINER_NOT_FOUND_MSG: NO_SUCH_CONTAINER_RESULT,
               FAIL_AUTH_MSG: INVALID_ACCESS_KEY_RESULT,
               INCORRECT_PADDING_MSG: INVALID_ACCESS_KEY_RESULT,
               BLOB_NOT_FOUND_MSG: SUCCESS_RESULT}

ACCESS_TEST_NAME = "access_test.I-O_DATA.access_test"

CHECK_TIMEOUT_SEC = 20

chk_result = FAILED_RESULT


def _exec_check(client, container_name):
    global chk_result

    try:
        client.get_blob_properties(container_name, ACCESS_TEST_NAME)
        chk_result = SUCCESS_RESULT

    except WindowsAzureError, ins:
        chk_result = ERROR_TABLE.get(ins.args[0], FAILED_RESULT)

    except TypeError, ins:
        try:
            chk_result = ERROR_TABLE.get(ins.args[0].args[0], FAILED_RESULT)
        except:
            pass

    except:
        pass


class PalAzure(palcmd.PalCmd):
    """PAL Azure command implementation class."""

    def __make_download_share_file(self, share_name):

        with open(config.AZURE_DOWNLOAD_TARGET_TEMP, "wb") as file:
            file.write(share_name)
        shutil.move(config.AZURE_DOWNLOAD_TARGET_TEMP,
            config.AZURE_DOWNLOAD_TARGET)

    def __download_handler(self, entry):
        """Handle download request."""

        try:
            share_name = entry.get().get_node(SHARENAME_NAME).get_text()
            self.__make_download_share_file(share_name)

            CmdExec(SEND_DOWNLOAD_SIGNAL_ARGS).execute(
                stdout=NULL, stderr=NULL)
            entry.set_result(Results.OK)

        except ExecError:
            self.exception(DOWNLOAD_FAILED_MSG)
            entry.set_result(Results.DOWNLOAD_ERROR)

        return entry

    def __check_handler(self, entry):
        """Handle check request."""

        global chk_result

        try:
            entry.set_result(Results.OK)

            body = entry.get()

            account_name = body.get_value(ACCOUNTNAME_NAME)
            access_key = body.get_value(ACCESSKEY_NAME)
            container_name = body.get_value(CONTAINER_NAME)
            protocol = body.get_value(HTTPMETHOD_NAME)
            proxy_host = body.get_value(PROXYHOST_PATH, None)
            proxy_port = body.get_value(PROXYPORT_PATH, None)

            client = BlobService(account_name,
                                 access_key,
                                 protocol)

            if proxy_host and proxy_port:
                client.set_proxy(proxy_host, proxy_port)

            t = threading.Thread(target=_exec_check,
                                 args=(client, container_name))
            t.setDaemon(True)
            t.start()
            t.join(CHECK_TIMEOUT_SEC)
            if t.isAlive():
                chk_result = TIMEOUT_RESULT

        except:
            chk_result = FAILED_RESULT
            entry.set_result(Results.CHECK_ERROR)

        finally:
            end_result = chk_result

            if end_result != SUCCESS_RESULT:
                self.exception(CHECK_FAILED_MSG)

            entry.get().append_child(RESULT_NAME, end_result)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.DOWNLOAD, self.__download_handler)
        self.set_handler(ReqTypes.CHECK, self.__check_handler)

#
# Main
#

if __name__ == "__main__":
    PalAzure(option=palcmd.PalCmdOption()).handle_request()
