#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for system initialization."""


import shutil
import re
import os

import libpal.palcmd as palcmd
import libpal.util as util

from iobase.exception import (NotFoundError)
from iobase.cmdexec import (CmdExec, NULL, PIPE)
from pal_info import (ReqTypes, Results)
import pal_config as config


MNT_DEVLIST = "devlist"
MNT_POINT = "mnt_point"


class PalInitialize(palcmd.PalCmd):
    """Pal initialize implementation class."""

    def __is_zeroclear(self, entry):
        try:
            if entry.get().get_node("option").is_set_to_true("zeroclear"):
                return True
            else:
                return False
        except NotFoundError:
            return False

    def __get_filesystem_type(self, part_path):
        """Get filesystem type of specified partition."""

        cmd = [config.BLKID_CMD, part_path]
        proc = CmdExec(cmd, stdout=PIPE, stderr=NULL).execute()

        for line in proc.stdout:
            line = line.rstrip()
            match = re.search("TYPE=\"([^\"]*)\"", line)
            return match.group(1)
        return None

    def __get_dev_path(self, dev_list):
        for dev in dev_list:
            dev_path = os.path.join(config.DEV_DIR, dev)
            if os.path.exists(dev_path):
                return dev_path
        raise NotFoundError("No devices!")

    def __initialize_handler(self, entry):
        """Handle initialize request and set entry to result."""

        mnt_info = config.SYS_VOL_INFO[config.INITIALIZE_MNT]
        try:
            array_num = util.get_cuurent_array_num()
            dev_path = self.__get_dev_path(mnt_info[MNT_DEVLIST][array_num])
            filesystype = self.__get_filesystem_type(dev_path)
            cmd = [config.MOUNT_CMD,
                    "-t", filesystype,
                    "-o", config.MOUNT_OPT_INFO[filesystype],
                    dev_path, mnt_info[MNT_POINT]]
            CmdExec(cmd, stdout=NULL, stderr=NULL).execute()

            if self.__is_zeroclear(entry):
                shutil.copy(config.ZEROCLEAR_SCRIPT_PATH,
                            config.INITIALIZE_SCRIPT_PATH)
            else:
                shutil.copy(config.NO_ZEROCLEAR_SCRIPT_PATH,
                            config.INITIALIZE_SCRIPT_PATH)

            shutil.copy(config.RC_LOCAL_PATH, config.INITIALIZE_RC_LOCAL_PATH)
        except:
            entry.set_result(Results.INITIALIZE_ERROR)
            self.exception("Initialize Fail!")
            return entry
        entry.set_result(Results.OK)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.INITIALIZE, self.__initialize_handler)

#
# Main
#

if __name__ == "__main__":
    PalInitialize(option=palcmd.PalCmdOption()).handle_request()
