#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for MAIL."""

from __future__ import with_statement

import os
import re
import smtplib
import poplib
import pal_config as config
import libpal.palcmd as palcmd
from email.MIMEText import MIMEText
from email.Header import Header
from email.Utils import formatdate
from email.mime.multipart import (MIMEMultipart,)
from email.mime.image import (MIMEImage,)
from datetime import datetime
from iobase.exception import (InvalidArgError, NotFoundError)
from pal_info import ReqTypes
from pal_info import Results

### XML Tag Name ###
MAILSEND_NAME = "mailsend"
MAILCONF_NAME = "mailconf"
HOST_NAME = "host"
PORT_NAME = "port"
FROMADDR_NAME = "fromaddr"
TOADDR_NAME = "toaddr"
ENCODE_NAME = "encode"
AUTH_NAME = "auth"
USESSL_NAME = "usessl"
MESSAGE_NAME = "message"
HEADER_NAME = "header"
SUBJECT_NAME = "subject"
BODY_NAME = "body"
TOADDRLIST_NAME = "toaddrlist"
USER_NAME = "user"
PASSWORD_NAME = "password"
POP3_NAME = "pop3"

ATTACH_PATH = "attachlist/attach"
PATH_NAME = "path"


### XML Attribute ###
TYPE_ATTR = "type"
TEST_MODE_ATTR = "test"
SUBTYPE_ATTR = "subtype"

### XML Attribute value ###

SMTP_ATTR_VAL = "smtp"
NONE_AUTH_ATTR_VAL = "no"
POP_B4_SMTP_AUTH_ATTR_VAL = "pop"
PLAIN_AUTH_ATTR_VAL = "plain"
LOGIN_AUTH_ATTR_VAL = "login"
MD5_AUTH_ATTR_VAL = "crammd5"
ENABLE_ATTR_VAL = "enable"
NORESEND_ATTR = "noresend"

### Some Value ###
AUTH_PLAIN = "PLAIN"
AUTH_LOGIN = "LOGIN"
AUTH_CRAM_MD5 = "CRAM-MD5"

IMAGE_TYPE = "image"


class PalSmtp(smtplib.SMTP):
    """Mail class based on SMTP class"""

    #
    # Override method
    #

    def __init__(self, host='', port=25, pop3_host='', pop3_port=110):
        """Constructor."""

        smtplib.SMTP.__init__(self, host=host, port=port)
        self.__pop3_server_addr = pop3_host
        self.__pop3_port = pop3_port

    def __convert_ioauth_to_actualauth(self, target_ioauth):
        """Convert IODATA authtype to actual authtype"""

        ioauth_actauth = {PLAIN_AUTH_ATTR_VAL: AUTH_PLAIN,
                          LOGIN_AUTH_ATTR_VAL: AUTH_LOGIN,
                          MD5_AUTH_ATTR_VAL: AUTH_CRAM_MD5}
        for ioauth, actauth in ioauth_actauth.items():
            if target_ioauth == ioauth:
                return actauth

    def login(self, user, password, auth_type):
        """Overriden login method
           Add PopBeforeSmtp auth.
           Change smtp auth to user can specify authtype.
        """

        if POP_B4_SMTP_AUTH_ATTR_VAL == auth_type:

            self.__pop_b4_smtp_login(self.__pop3_server_addr,
                                     self.__pop3_port,
                                     user, password)
            return

        self.esmtp_features['auth'] \
            = self.__convert_ioauth_to_actualauth(auth_type)

        smtplib.SMTP.login(self, user, password)

    def __pop_b4_smtp_login(self, server_addr, port, user, password):
        """Login for Pop Before Smtp."""

        pop3 = poplib.POP3(server_addr, port)
        pop3.user(user)
        pop3.pass_(password)
        return


class PalMail(palcmd.PalCmd):
    """Pal Mail command implementation class."""

    #
    # Sub functions
    #

    def __validate_root_tag_name(self, entry):
        """Validate the root tag name."""

        node_name = entry.get().get_node_name()
        if MAILSEND_NAME != node_name:
            self.error("Invalid root node name %s" % node_name)
            raise InvalidArgError("invalid root tag..")

    def __is_test_mode(self, mailsend_node):
        """Check TEST MODE.
           This method return True when pal_mail test module is done.
        """

        try:
            if ENABLE_ATTR_VAL == mailsend_node.get_attr_value(TEST_MODE_ATTR):
                return True
        except NotFoundError:
            return False

        return False

    def __attach_parts(self, msg, attach_info_list):
        """Attach parts of multipart message to "msg"."""

        for attach_info in attach_info_list:
            # Only image is supported
            if attach_info[TYPE_ATTR] == IMAGE_TYPE:
                with open(attach_info[PATH_NAME], 'rb') as attach:
                    image_part = \
                        MIMEImage(attach.read(),
                            attach_info[SUBTYPE_ATTR],
                            name=os.path.basename(attach_info[PATH_NAME]))
                    msg.attach(image_part)

    def __create_message(self, fromaddr, toaddr, subject, body, charset,
            attach_info_list):
        """Create mail massage"""

        def convert_line_feed(body):

            body = body.replace("\r\n", "\n")
            return body.replace("\n", "\r\n")

        body = convert_line_feed(body)

        text = MIMEText(unicode(body, "utf-8").encode(charset, "replace"),
                       "plain", charset)
        if attach_info_list:
            msg = MIMEMultipart()
            msg.attach(text)
            self.__attach_parts(msg, attach_info_list)
        else:
            msg = text

        encode_subject = Header(unicode(subject, "utf-8").encode(charset),
                       charset)
        msg['Subject'] = encode_subject
        msg['From'] = fromaddr
        msg['To'] = toaddr
        msg['Date'] = formatdate(localtime=True)

        return msg

    def __send_mail(self, smtp_server, smtp_port, fromaddr, encode,
                    auth_type, userid, password, subject, body, toaddr_list,
                    pop3_server, pop3_port, usessl, attach_info_list=[]):
        """Send mail"""

        msg = self.__create_message(fromaddr, toaddr_list[0],
                                    subject, body, encode,
                                    attach_info_list)

        if POP_B4_SMTP_AUTH_ATTR_VAL == auth_type:
            pal_smtp = PalSmtp(host=smtp_server, port=smtp_port,
                               pop3_host=pop3_server, pop3_port=pop3_port)
        else:
            pal_smtp = PalSmtp(host=smtp_server, port=smtp_port)

        if usessl:
            pal_smtp.ehlo()
            pal_smtp.starttls()

        pal_smtp.ehlo()
        if NONE_AUTH_ATTR_VAL != auth_type:
            pal_smtp.login(userid, password, auth_type)
        pal_smtp.sendmail(fromaddr, toaddr_list, msg.as_string())
        pal_smtp.close()

    def __get_smtp_server(self, mailconf_node):
        """Get SMTP server from "mailconf" node"""

        server_node = mailconf_node.get_node(HOST_NAME)
        if SMTP_ATTR_VAL != server_node.get_attr_value(TYPE_ATTR):
            raise InvalidArgError("fail to get smtp server info..")
        return server_node.get_text()

    def __get_smtp_port(self, mailconf_node):
        """Get SMTP port from "mailconf" node"""

        port_node = mailconf_node.get_node(PORT_NAME)
        if SMTP_ATTR_VAL != port_node.get_attr_value(TYPE_ATTR):
            raise InvalidArgError("fail to get smtp port..")
        return int(port_node.get_text())

    def __get_fromaddr(self, mailconf_node):
        """Get source mail address from "mailconf" node"""

        return mailconf_node.get_node(FROMADDR_NAME).get_text()

    def __get_encode(self, mailconf_node):
        """Get encode from "mailconf" node"""

        return mailconf_node.get_node(ENCODE_NAME).get_text()

    def __get_authinfo(self, mailconf_node):
        """Get auth type of mail from "mailconf" node"""

        auth_node = mailconf_node.get_node(AUTH_NAME)
        auth_type = auth_node.get_attr_value(TYPE_ATTR)
        if NONE_AUTH_ATTR_VAL == auth_type:
            return auth_type, "", ""

        userid = auth_node.get_node(USER_NAME).get_text(do_strip=False)
        password = auth_node.get_node(PASSWORD_NAME).get_text(do_strip=False)
        return auth_type, userid, password

    def __get_usessl(self, mailconf_node):
        """Get "usessl" setting from "mailconf" node"""

        try:
            return mailconf_node.get_node(USESSL_NAME).is_enabled()
        except:
            return False

    def __get_toaddr_list(self, mailconf_node):
        """Get destination mail address list body mail from
           "mailconf_node" node
        """

        toaddr_node_list = \
            mailconf_node.get_node(TOADDRLIST_NAME).get_node_list(TOADDR_NAME)

        toaddr_list = []
        for toaddr_node in toaddr_node_list:
            toaddr_list += [toaddr_node.get_text()]

        return toaddr_list

    def __get_message_subject(self, mailsend_node):
        """Get message subject mail from "mail" node"""

        node = mailsend_node.get_node(MESSAGE_NAME).get_node(HEADER_NAME)
        return node.get_node(SUBJECT_NAME).get_text()

    def __get_message_body(self, mailsend_node):
        """Get message body mail from "mail" node"""

        body_node = mailsend_node.get_node(MESSAGE_NAME).get_node(BODY_NAME)
        return body_node.get_text()

    def __get_attach_info_list(self, mailsend_node):
        """Get attach info list."""

        attach_info_list = []
        for attach in mailsend_node.get_node_list(ATTACH_PATH):
            attach_info = {}
            attach_info[TYPE_ATTR] = attach.get_attr_value(TYPE_ATTR)
            attach_info[SUBTYPE_ATTR] = attach.get_attr_value(SUBTYPE_ATTR)
            attach_info[PATH_NAME] = attach.get_node(PATH_NAME).get_text()

            attach_info_list.append(attach_info)
        return attach_info_list

    def __get_pop3_param(self, mailconf_node):
        """Get pop3 tag params"""

        pop3_node = mailconf_node.get_node(AUTH_NAME).get_node(POP3_NAME)
        pop3_server = pop3_node.get_node(HOST_NAME).get_text()
        pop3_port = pop3_node.get_node(PORT_NAME).get_text()
        return pop3_server, int(pop3_port)

    def __is_resend(self, mailsend_node):

        if mailsend_node.get_node(MESSAGE_NAME).is_set_to_true(NORESEND_ATTR):
            return False
        return True

    def __pool_resend_mail(self, mailsend_node):

        def get_current_datetime():

            now = datetime.now()
            return now.strftime("%Y%m%d%H%M%S")

        def remove_landisk_name_from_subject():
            # メールのタイトルから[landisk***]を取り除く
            # landiks名はメール送信する時に自動的につく。
            # 削除しないとそのままだと2重につくことになる
            node = mailsend_node.get_node(MESSAGE_NAME).get_node(HEADER_NAME)
            subject = node.get_node(SUBJECT_NAME).get_text()
            match = re.match("^(\[.*\])(.*)$", subject)
            node.get_node(SUBJECT_NAME).set_text(match.group(2).lstrip())
            return

        remove_landisk_name_from_subject()

        try:
            message_node = mailsend_node.get_node(MESSAGE_NAME)
            try:
                message_node.get_attr_value("retrycount")
            except:
                message_node.set_attrs({"retrycount": "0"})
        except:
            pass

        mail_file_path = os.path.join(config.MAIL_SPOOL_DIR,
                                      get_current_datetime())

        with open(mail_file_path, "w") as fh:
            fh.write(str(mailsend_node))

    #
    # Request handlers
    #

    def __send_handler(self, entry):
        """Handle "send" request."""

        self.__validate_root_tag_name(entry)

        mailsend_node = entry.get()
        mailconf_node = mailsend_node.get_node(MAILCONF_NAME)

        test_flg = self.__is_test_mode(mailsend_node)
        smtp_server = self.__get_smtp_server(mailconf_node)
        smtp_port = self.__get_smtp_port(mailconf_node)
        fromaddr = self.__get_fromaddr(mailconf_node)
        encode = self.__get_encode(mailconf_node)
        auth_type, userid, password = self.__get_authinfo(mailconf_node)
        usessl = self.__get_usessl(mailconf_node)
        toaddr_list = self.__get_toaddr_list(mailconf_node)

        subject = self.__get_message_subject(mailsend_node)
        body = self.__get_message_body(mailsend_node)
        attach_info_list = self.__get_attach_info_list(mailsend_node)

        pop3_server = ""
        pop3_port = 110
        if POP_B4_SMTP_AUTH_ATTR_VAL == auth_type:
            pop3_server, pop3_port = self.__get_pop3_param(mailconf_node)

        if test_flg:
            entry.set_result(Results.OK)
            return entry

        try:
            self.__send_mail(smtp_server, smtp_port, fromaddr, encode,
                             auth_type, userid, password, subject, body,
                             toaddr_list, pop3_server, pop3_port, usessl,
                             attach_info_list)
        except:
            # リトライ属性ありならリトライ
            if self.__is_resend(mailsend_node):
                self.__pool_resend_mail(mailsend_node.get_clone())

            entry.set_result(Results.SEND_ERROR)
            return entry

        entry.set_result(Results.OK)
        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.SEND, self.__send_handler)

#
# Main
#

if __name__ == "__main__":
    PalMail(option=palcmd.PalCmdOption()).handle_request()
