#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Samba."""


import os
import shutil
import re
import glob
import select

from ionas.fileaccess import SmbAccess
from ionas.exception import PermissionError
from libpal import (palcmd, util)
from pal_info import (ReqTypes, Results)
from iobase.cmdexec import (CmdExec, NULL, PIPE)
from iobase.exception import (NotFoundError, InvalidArgError, ExecError)
from iobase.xmldata import XmlNode
from ionas.confut import (build_config_file, disable_setting, enable_setting)
import pal_config as config

PRINTER_FLAG_PATH = config.TEMP_DIR + "/printer/printer_enable"
# See "evh_printer"
USB_ORIG_SUFFIX = ".real"
# See "evh_ports"
USB_SHARE_RE = "^usb\d$"

TMP_SUFFIX = ".tmp"

TO_NULL_OPTS = {"stdout": NULL, "stderr": NULL}

ON_AUDIT_FILE = config.SMB_PART_GLOBAL_AUDIT_FILE + ".on"
OFF_AUDIT_FILE = config.SMB_PART_GLOBAL_AUDIT_FILE + ".off"
DOMAIN_NSSWITCH = config.NSSWITCH_CONF_FILE + ".domain"
NODOMAIN_NSSWITCH = config.NSSWITCH_CONF_FILE + ".nodomain"

HOSTDESC_NAME = "hostdesc"
SHARES_NAME = "shares"
MSNETWORK_NAME = "msnetwork"
NAME_NAME = "name"
ASSIGN_NAME = "assign"
DOMAIN_NAME = "domain"
HOST_NAME = "host"
SHARE_NAME = "share"
DOMAINLIST_NAME = "domainlist"
HOSTLIST_NAME = "hostlist"
SHARELIST_NAME = "sharelist"
DOMAINLIST_NODE_STR = "<shares><domainlist/></shares>"
DOMAIN_FMT_STR = "<domain><name>%s</name><hostlist/></domain>"
HOST_FMT_STR = "<host><name>%s</name><sharelist/></host>"
SHARE_FMT_STR = "<share><name>%s</name></share>"

NAME_PATH = "share/name"
DESC_PATH = "share/desc"
PATH_PATH = "share/path"
UNAME_PATH = "share/uname"
UID_PATH = "share/uid"
GID_PATH = "share/gid"
PERMISSION_PATH = "share/permission"
READONLY_PATH = PERMISSION_PATH + "/readonly"
BACKUP_PATH = PERMISSION_PATH + "/backup"
DETAILPERMISSION_PATH = PERMISSION_PATH + "/detailpermission"
READONLYMEMBER_PATH = DETAILPERMISSION_PATH + "/readonlymember"
RO_USER_LIST_PATH = READONLYMEMBER_PATH + "/userlist/*"
RO_GROUP_LIST_PATH = READONLYMEMBER_PATH + "/grouplist/*"
READWRITEMEMBER_PATH = DETAILPERMISSION_PATH + "/readwritemember"
RW_USER_LIST_PATH = READWRITEMEMBER_PATH + "/userlist/*"
RW_GROUP_LIST_PATH = READWRITEMEMBER_PATH + "/grouplist/*"
SHARESERVICE_PATH = "share/shareservice"
NOGUEST_PATH = SHARESERVICE_PATH + "/noguest"
MSSHARE_PATH = SHARESERVICE_PATH + "/msshare"
HIDDEN_PATH = MSSHARE_PATH + "/hidden"
TRASH_PATH = MSSHARE_PATH + "/trash"
WG_PATH = "assign/wg"
WORKGROUP_PATH = WG_PATH + "/workgroup"
NT_PATH = "assign/nt"
NT_DOMAIN_PATH = NT_PATH + "/domain"
NT_DCNAME_PATH = NT_PATH + "/dcname"
AD_PATH = "assign/ad"
NETBIOS_PATH = AD_PATH + "/netbios"
FULLDNS_PATH = AD_PATH + "/fulldns"
AD_DCNAME_PATH = AD_PATH + "/dcname"
AD_ADMINNAME_PATH = AD_PATH + "/adminname"
AD_ADMINPASSWD_PATH = AD_PATH + "/adminpasswd"
WINSSRV_PATH = "./winssrv"

METHOD_ATTR = "method"

PATH_KEY = "path"
COMMENT_KEY = "comment"
READ_ONLY_KEY = "read only"
VALID_USERS_KEY = "valid users"
READ_LIST_KEY = "read list"
WRITE_LIST_KEY = "write list"
BROWSEABLE_KEY = "browseable"
PUBLIC_KEY = "public"
ADMIN_USERS_KEY = "admin users"
FORCE_USER_KEY = "force user"
FORCE_GROUP_KEY = "force group"
RECYCLE_REPOSITORY_KEY = "recycle:repository"
RECYCLE_DIRECTORY_MODE_KEY = "recycle:directory_mode"
RECYCLE_KEEPTREE_KEY = "recycle:keeptree"
RECYCLE_VERSIONS_KEY = "recycle:versions"
RECYCLE_TOUCH_KEY = "recycle:touch"
VFS_OBJECTS_KEY = "vfs objects"
SECURITY_KEY = "security"
WORKGROUP_KEY = "workgroup"
PASSWORD_SERVER_KEY = "password server"
WINBIND_RPC_ONLY_KEY = "winbind rpc only"
REALM_KEY = "realm"
WINS_SERVER_KEY = "wins server"
FULLDNS_KEY = "fulldns"
DOMAIN_CONTROLER_KEY = "domain_controler"
DOMAIN_NAME_KEY = "domain_name"

YES_STR = "yes"
NO_STR = "no"
NOBODY_STR = "nobody"
NOGROUP_STR = "nogroup"
ADMIN_STR = "admin"
ADMIN_QUOT_STR = '"admin"'
RECYCLE_STR = "recycle"
FULL_AUDIT_STR = "full_audit"
USER_STR = "user"
DOMAIN_STR = "domain"
ADS_STR = "ads"
SATARAID1 = "sataraid1"

RECYCLE_REPOSITORY = "trashbox"
RECYCLE_DIRECTORY_MODE = "0777"
RECYCLE_KEEPTREE = "yes"
RECYCLE_VERSIONS = "yes"
RECYCLE_TOUCH = "yes"

WRITE_MODE = "w"

VOLUME_RE = r"/mnt/([^/]+)"

USER_FMT_STR = '"%s"'
GROUP_FMT_STR = '"@%s"'
SMB_CONF_SECTION_FMT_STR = "[%s]\n"
SMB_CONF_ITEM_FMT_STR = "\t%s = %s\n"
KRB5_CONF_FMT_STR = \
        "[libdefaults]\n" + \
        "\tdefault_realm = %(" + FULLDNS_KEY + ")s\n" + \
        "\n" + \
        "[realms]\n" + \
        "\t%(" + FULLDNS_KEY + ")s = {\n" + \
        "\t\tkdc = %(" + DOMAIN_CONTROLER_KEY + ")s\n" + \
        "\t}\n" + \
        "\n" + \
        "[domain_realms]\n" + \
        "\t.%(" + DOMAIN_NAME_KEY + ")s = %(" + FULLDNS_KEY + ")s\n" + \
        "\t%(" + DOMAIN_NAME_KEY + ")s  = %(" + FULLDNS_KEY + ")s\n"
ADMIN_INFO_FMT_STR = "%s%%%s"
HOSTDESC_STR = "\tserver string = %s\n"

FILE_NAME_SEPARATOR = "."
DOMAIN_SEPARATOR = "."
ANY_NAME = "*"

RW = "rw"
RO = "ro"

WG = "wg"
NT = "nt"
AD = "ad"

NT_OLDJOIN_ARGS = [config.NET_CMD, "rpc", "oldjoin"]
NT_TESTJOIN_ARGS = [config.NET_CMD, "rpc", "testjoin"]
AD_JOIN_ARGS = [config.NET_CMD, "ads", "join",
        "-S", "dc_name", "-U", "admin_info"]
DC_NAME_OPT_IDX = 4
ADMIN_INFO_OPT_IDX = 6
NOT_SHARE_PART_LIST = [config.SMB_PART_GLOBAL_BASE_FILE,
                    config.SMB_PART_GLOBAL_HOSTDESC_FILE,
                    config.SMB_PART_GLOBAL_DOMAIN_FILE,
                    config.SMB_PART_GLOBAL_PRINTER_FILE,
                    config.SMB_PART_GLOBAL_AUDIT_FILE,
                    config.SMB_PART_PRINTERS_VALIABLE_FILE]

SMB_PREFIX = "smb://"

TRASHBOX_NAME = "trashbox"

TIMEOUT = 30

MKDIR_MODE = 0777

VFS_OBJECTS_RE = VFS_OBJECTS_KEY + " = "


class PalSamba(palcmd.PalCmd):
    """Pal Samba command implementation class."""

    def __check_root_node_name(self, entry, name):
        """Root node name check."""

        if entry.get().get_node_name() != name:
            raise InvalidArgError(\
                    "root_node_name = " + entry.get().get_node_name())

    def __is_enabled(self, node, path):
        """Check enabled. In the case of NotFoundError, return False"""

        try:
            result = node.get_node(path).is_enabled()
            return result

        except NotFoundError:
            return False

    def __get_browseable_setting(self, entry):
        """Get browseable setting."""

        value_dict = {}

        if self.__is_enabled(entry.get(), HIDDEN_PATH):
            value_dict[BROWSEABLE_KEY] = NO_STR
        else:
            value_dict[BROWSEABLE_KEY] = YES_STR

        return value_dict

    def __get_detail_permission_list(self, entry, format_str, path):
        """Get detail permission list."""

        perm_list = []
        # FIXME get_value()が""かまで見る？
        # それはvalidateの範疇では？
        # 見るとしたらshare_nameも見る必要あるが?
        try:
            perm_list = [format_str % node.get_value(NAME_NAME) \
                         for node in entry.get().get_node_list(path) \
                         if node.get_value(NAME_NAME) != ""]

        except NotFoundError:
            pass

        return perm_list

    def __is_read_only(self, node, suffix):
        """Is read only check."""

        if suffix == RO:
            return True
        if self.__is_enabled(node, READONLY_PATH):
            return True
        if self.__is_enabled(node, BACKUP_PATH):
            return True
        return False

    def __get_detail_permission_share_setting(self, entry, suffix):
        """Get detail permission share setting."""

        value_dict = {}
        msshare_node = entry.get().get_node(MSSHARE_PATH)
        try:
            quota = msshare_node.get_value("quota/limit")
            if len(quota) > 0 and float(quota) > 0:
                uname = entry.get().get_value(UNAME_PATH)
                value_dict[FORCE_USER_KEY] = uname
                value_dict[FORCE_GROUP_KEY] = NOGROUP_STR
        except NotFoundError:
            pass

        value_dict[PUBLIC_KEY] = NO_STR

        rou_list = rog_list = rwu_list = rwg_list = []

        rou_list = self.__get_detail_permission_list(entry,
                USER_FMT_STR, RO_USER_LIST_PATH)

        rog_list = self.__get_detail_permission_list(entry,
                GROUP_FMT_STR, RO_GROUP_LIST_PATH)

        rwu_list = self.__get_detail_permission_list(entry,
                USER_FMT_STR, RW_USER_LIST_PATH)

        rwg_list = self.__get_detail_permission_list(entry,
                GROUP_FMT_STR, RW_GROUP_LIST_PATH)

        ug_valid = ",".join(
            [ADMIN_QUOT_STR] + rou_list + rwu_list + rog_list + rwg_list)

        value_dict[VALID_USERS_KEY] = ug_valid

        # FIXME smb.confに"read_list = "といった
        # 値のない記述がされる可能性あるが大丈夫か
        if self.__is_read_only(entry.get(), suffix):
            value_dict[READ_ONLY_KEY] = YES_STR
            value_dict[READ_LIST_KEY] = ug_valid
            value_dict[WRITE_LIST_KEY] = ADMIN_QUOT_STR

        else:
            value_dict[READ_ONLY_KEY] = NO_STR
            value_dict[READ_LIST_KEY] = ",".join(rou_list + rog_list)
            value_dict[WRITE_LIST_KEY] = ",".join(
                [ADMIN_QUOT_STR] + rwu_list + rwg_list)

        return value_dict

    def __get_all_share_setting(self, entry, suffix):
        """Get all share setting."""

        value_dict = {}

        if self.__is_read_only(entry.get(), suffix):
            value_dict[READ_ONLY_KEY] = YES_STR
            value_dict[WRITE_LIST_KEY] = ADMIN_QUOT_STR

        else:
            value_dict[READ_ONLY_KEY] = NO_STR

        msshare_node = entry.get().get_node(MSSHARE_PATH)
        try:
            quota = msshare_node.get_value("quota/limit")
            if len(quota) > 0 and float(quota) > 0:
                uname = entry.get().get_value(UNAME_PATH)
                value_dict[FORCE_USER_KEY] = uname
            else:
                value_dict[FORCE_USER_KEY] = NOBODY_STR
        except NotFoundError:
            value_dict[FORCE_USER_KEY] = NOBODY_STR
        value_dict[FORCE_GROUP_KEY] = NOGROUP_STR

        if self.__is_enabled(entry.get(), NOGUEST_PATH):
            value_dict[PUBLIC_KEY] = NO_STR
        else:
            value_dict[PUBLIC_KEY] = YES_STR

        return value_dict

    def __get_permission_setting(self, entry,  suffix):
        """Get permission setting."""

        value_dict = {}

        if self.__is_enabled(entry.get(), DETAILPERMISSION_PATH):
            value_dict.update(
                self.__get_detail_permission_share_setting(entry, suffix))

        else:
            value_dict.update(self.__get_all_share_setting(entry, suffix))

        return value_dict

    def __is_enable_accesslog(self):
        """Check access log availability."""

        try:
            path_current = os.path.join(
                config.SMB_CONF_PART_DIR, config.SMB_PART_GLOBAL_AUDIT_FILE)
            path_audit_on = os.path.join(
                config.SMB_CONF_PART_DIR, ON_AUDIT_FILE)
            file_current = os.path.basename(os.readlink(path_current))
            file_audit_on = os.path.basename(path_audit_on)
            if file_current == file_audit_on:
                return True
        except:
            pass

        return False

    def __get_trash_box_setting(self, entry):
        """Get trash box setting."""

        value_dict = {}

        if self.__is_enabled(entry.get(), TRASH_PATH):
            value_dict[RECYCLE_REPOSITORY_KEY] = RECYCLE_REPOSITORY
            value_dict[RECYCLE_DIRECTORY_MODE_KEY] = RECYCLE_DIRECTORY_MODE
            value_dict[RECYCLE_KEEPTREE_KEY] = RECYCLE_KEEPTREE
            value_dict[RECYCLE_VERSIONS_KEY] = RECYCLE_VERSIONS
            value_dict[RECYCLE_TOUCH_KEY] = RECYCLE_TOUCH

            if self.__is_enable_accesslog():
                vfs = [RECYCLE_STR, FULL_AUDIT_STR]
            else:
                vfs = [RECYCLE_STR, ]

            value_dict[VFS_OBJECTS_KEY] = " ".join(vfs)

        return value_dict

    def __get_other_setting(self, entry):
        """Get other setting."""

        value_dict = {}
        value_dict[PATH_KEY] = entry.get().get_value(PATH_PATH)
        value_dict[ADMIN_USERS_KEY] = ADMIN_STR

        try:
            value_dict[COMMENT_KEY] = entry.get().get_value(DESC_PATH)

        except NotFoundError:
            pass

        return value_dict

    def __get_share_value_dict_setting(self, entry, suffix):
        """Get share value dict setting."""

        value_dict = {}

        value_dict.update(self.__get_browseable_setting(entry))

        value_dict.update(self.__get_permission_setting(entry, suffix))

        value_dict.update(self.__get_trash_box_setting(entry))

        value_dict.update(self.__get_other_setting(entry))

        return value_dict

    def __create_temp_share_part_file(self, entry, share_name, suffix):
        """Create temp samba share part file."""

        value_dict = self.__get_share_value_dict_setting(entry, suffix)

        tmp_file_name = share_name + FILE_NAME_SEPARATOR + suffix + \
                FILE_NAME_SEPARATOR + str(os.getpid())
        tmp_file_path = os.path.join(config.TEMP_DIR, tmp_file_name)

        tmp_file = open(tmp_file_path, WRITE_MODE)

        tmp_file.write(SMB_CONF_SECTION_FMT_STR % share_name)
        for key, value in value_dict.items():
            tmp_file.write(SMB_CONF_ITEM_FMT_STR % (key, value))
        tmp_file.write("\n")
        tmp_file.close()

        return tmp_file_path

    def __move_temp_share_part_file(
        self, tmp_file_path, file_dir, share_name, suffix):
        """Move temp share part file."""

        link_path = ""
        file_name = share_name + FILE_NAME_SEPARATOR + suffix
        usb_match = re.match(USB_SHARE_RE, share_name)
        if usb_match:
            link_path = os.path.join(file_dir, file_name)
            file_name += USB_ORIG_SUFFIX
        file_path = os.path.join(file_dir, file_name)

        shutil.move(tmp_file_path, file_path)
        if link_path != "":
            if not os.path.exists(link_path):
                os.symlink(file_path, link_path)

        return

    def __create_share_part_file(self, entry):
        """Create samba share part file."""

        share_name = entry.get().get_value(NAME_PATH)
        share_volume = re.match(VOLUME_RE,
                entry.get().get_value(PATH_PATH)).group(1)
        src_dir = os.path.join(config.SMB_CONF_PART_DIR, share_volume)
        if not os.path.exists(src_dir):
            os.makedirs(src_dir)
        for suffix in [RW, RO]:
            tmp_file_path = self.__create_temp_share_part_file(
                entry, share_name, suffix)
            self.__move_temp_share_part_file(
                tmp_file_path, src_dir, share_name, suffix)

        return

    def __remove_share_part_file(self, file_dir, share_name, suffix):
        """Remove share part file."""

        del_list = []
        del_list.append(share_name + FILE_NAME_SEPARATOR + suffix)

        usb_match = re.match(USB_SHARE_RE, share_name)
        if usb_match:
            orig_name = share_name + FILE_NAME_SEPARATOR + \
                        suffix + USB_ORIG_SUFFIX
            del_list.append(orig_name)

        for file_name in del_list:
            file_path = os.path.join(file_dir, file_name)
            if os.path.exists(file_path):
                os.remove(file_path)
        return

    def __delete_share_part_file(self, entry):
        """Delete samba share part file."""

        share_name = entry.get().get_value(NAME_PATH)
        share_volume = re.match(VOLUME_RE, entry.get().\
                get_value(PATH_PATH)).group(1)
        file_dir = os.path.join(config.SMB_CONF_PART_DIR, share_volume)

        for suffix in [RW, RO]:
            self.__remove_share_part_file(file_dir, share_name, suffix)
        return

    def __is_trashbox_enable(self, share_part_file_path):
        """Check trashbox config."""

        fpr = open(share_part_file_path, "r")
        for line in fpr.readlines():
            if re.search(VFS_OBJECTS_RE, line):
                if re.search(RECYCLE_STR, line):
                    return True
        fpr.close()
        return False

    def __update_share_part_files(self):
        """Update samba share part files."""

        CONFLINE_VFS_AT = "\t%s = %s %s\n\n" % (
            VFS_OBJECTS_KEY, RECYCLE_STR, FULL_AUDIT_STR)
        CONFLINE_VFS_A = "\t%s = %s\n\n" % (VFS_OBJECTS_KEY, FULL_AUDIT_STR)
        CONFLINE_VFS_T = "\t%s = %s\n\n" % (VFS_OBJECTS_KEY, RECYCLE_STR)

        accesslog_enable = self.__is_enable_accesslog()

        file_dir_list = []
        for share_volume in config.VOL_LIST:
            file_dir_list.append(
                os.path.join(config.SMB_CONF_PART_DIR, share_volume))

        for file_dir in file_dir_list:
            if not os.path.exists(file_dir):
                continue

            for dir_file in os.listdir(file_dir):
                dir_file_path = os.path.join(file_dir, dir_file)

                if os.path.islink(dir_file_path):
                    # Ignore a symblic link
                    continue

                trashbox_enable = self.__is_trashbox_enable(dir_file_path)

                fpr = open(dir_file_path, "r")
                fpw = open(dir_file_path + TMP_SUFFIX, "w")
                for line in fpr.readlines():
                    if not re.search(VFS_OBJECTS_RE, line) and line != "\n":
                        fpw.write(line)
                fpr.close()
                if accesslog_enable:
                    if trashbox_enable:
                        fpw.write(CONFLINE_VFS_AT)
                    else:
                        fpw.write(CONFLINE_VFS_A)
                else:
                    if trashbox_enable:
                        fpw.write(CONFLINE_VFS_T)
                    else:
                        fpw.write("\n")
                fpw.close()

                basename = os.path.basename(dir_file_path)
                share_name = os.path.splitext(basename)[0]
                usb_match = re.match(USB_SHARE_RE, share_name)
                if usb_match:
                    shutil.move(
                        dir_file_path + TMP_SUFFIX,
                        dir_file_path + USB_ORIG_SUFFIX)
                else:
                    shutil.move(dir_file_path + TMP_SUFFIX, dir_file_path)

    def __get_mount_info(self, vol):
        """Get the suffix list of 'vol'."""

        mnt_point = config.VOL_INFO[vol]["mnt_point"]
        if util.is_mounted(mnt_point):
            if util.is_mounted_as_read_only(mnt_point):
                return [".ro"]
            else:
                return [".rw"]
        else:
            return None

    def __sort_dir_list(self, setting_dir):

        path_list = []
        for part_name in NOT_SHARE_PART_LIST:
            path_list.append(os.path.join(setting_dir, part_name))
        all_path = glob.glob(os.path.join(setting_dir, ANY_NAME))
        all_path.sort()
        for pt in all_path:
            if not pt in path_list:
                path_list.append(pt)
        return path_list

    def __update_global_printer_link(self, src_dir, setting_dir):

        if os.path.isfile(PRINTER_FLAG_PATH):
            opts = [".on"]
            enable = True
        else:
            opts = [".off"]
            enable = False
        enable_setting([config.SMB_PART_GLOBAL_PRINTER_FILE],
                opts, src_dir, setting_dir)
        return enable

    def __update_printer_valiable_link(self, enable, src_dir, setting_dir):

        opts = [".noport"]
        if enable:
            for dev in config.USB_LP_DEV_LIST:
                dev_path = os.path.join(config.DEV_DIR, dev)
                if os.path.exists(dev_path):
                    opts = ["." + dev]
                    break

        enable_setting([config.SMB_PART_PRINTERS_VALIABLE_FILE],
                opts, src_dir, setting_dir)
        return

    def __create_smb_conf_file(self):
        """Create smb.conf file."""

        conf_dir = config.SMB_CONF_REAL_DIR
        if not os.path.exists(conf_dir):
            os.makedirs(conf_dir)
        conf_path = os.path.join(conf_dir, config.SMB_CONF_FILE)
        setting_dir = config.SMB_SETTING_DIR
        if os.path.exists(setting_dir):
            for part_file in glob.glob(os.path.join(setting_dir, ANY_NAME)):
                if os.path.islink(part_file):
                    os.remove(part_file)
        else:
            os.makedirs(setting_dir)

        src_dir = config.SMB_CONF_PART_DIR
        enable_setting([config.SMB_PART_GLOBAL_BASE_FILE,
                        config.SMB_PART_GLOBAL_HOSTDESC_FILE,
                        config.SMB_PART_GLOBAL_DOMAIN_FILE,
                        config.SMB_PART_GLOBAL_AUDIT_FILE],
                        [], src_dir, setting_dir)
        enable = self.__update_global_printer_link(src_dir, setting_dir)
        self.__update_printer_valiable_link(enable, src_dir, setting_dir)
        self.__create_or_delete_share_link(src_dir, setting_dir)
        build_config_file(conf_path, self.__sort_dir_list(setting_dir))

        return

    def __create_or_delete_share_link(self, src_dir, setting_dir):

        for vol in config.VOL_INFO.keys():
            vol_dir = os.path.join(src_dir, vol)
            share_list = []
            for path in glob.glob(os.path.join(vol_dir, ANY_NAME)):
                if not os.path.exists(path):
                    # Ignore "path", becase it may be a symblic link
                    # to a file which doesn't exist.
                    continue

                base = os.path.basename(path)
                match = re.match("(.*)\.r[ow]$", base)
                if not match:
                    continue
                if match.group(1) in share_list:
                    continue
                share_list.append(match.group(1))
            mnt_opt = self.__get_mount_info(vol)
            if share_list != []:
                if mnt_opt:
                    enable_setting(share_list, mnt_opt, vol_dir, setting_dir)
                else:
                    disable_setting(share_list, setting_dir)

        return

    def __get_wg_value_dict_setting(self, entry):
        """Get workgroup value dict set."""

        value_dict = {}
        value_dict[SECURITY_KEY] = USER_STR
        value_dict[WORKGROUP_KEY] = entry.get().get_value(WORKGROUP_PATH)

        try:
            value_dict[WINS_SERVER_KEY] = entry.get().get_value(WINSSRV_PATH)

        # FIXME "wins server = "の場合があるが大丈夫か?
        except NotFoundError:
            pass

        return value_dict

    def __get_nt_value_dict_setting(self, entry):
        """Get ntdomain value dict setting."""

        value_dict = {}
        value_dict[SECURITY_KEY] = DOMAIN_STR
        value_dict[WORKGROUP_KEY] = entry.get().get_value(NT_DOMAIN_PATH)
        value_dict[PASSWORD_SERVER_KEY] = entry.get().get_value(NT_DCNAME_PATH)
        value_dict[WINBIND_RPC_ONLY_KEY] = YES_STR

        try:
            value_dict[WINS_SERVER_KEY] = entry.get().get_value(WINSSRV_PATH)

        except NotFoundError:
            pass

        return value_dict

    def __get_ad_dc_name(self, entry):
        """Get active directory domain controler name."""

        ad_dcname = entry.get().get_value(AD_DCNAME_PATH)

        try:
            dc_hostname, dc_domainname = ad_dcname.split(DOMAIN_SEPARATOR, 1)
            """ad_dcname is FQDN. """
            return ad_dcname.lower()

        except ValueError:
            """ad_dcname is short host name. """
            return ad_dcname + DOMAIN_SEPARATOR + \
                    entry.get().get_value(FULLDNS_PATH).lower()

    def __get_ad_value_dict_setting(self, entry):
        """Get active directory value dict setting."""

        value_dict = {}
        value_dict[SECURITY_KEY] = ADS_STR
        value_dict[WORKGROUP_KEY] = entry.get().get_value(NETBIOS_PATH)
        value_dict[REALM_KEY] = entry.get().get_value(FULLDNS_PATH).upper()
        value_dict[PASSWORD_SERVER_KEY] = self.__get_ad_dc_name(entry)

        try:
            value_dict[WINS_SERVER_KEY] = entry.get().get_value(WINSSRV_PATH)

        except NotFoundError:
            pass

        return value_dict

    def __get_domain_value_dict_setting(self, entry, method):
        """Get domain value dict setting."""

        if method == WG:
            return self.__get_wg_value_dict_setting(entry)
        if method == NT:
            return self.__get_nt_value_dict_setting(entry)
        if method == AD:
            return self.__get_ad_value_dict_setting(entry)

    def __create_domain_part_file(self, entry, method):
        """Create samba domain part file."""

        tmp_file_name = config.SMB_PART_GLOBAL_DOMAIN_FILE + \
                FILE_NAME_SEPARATOR + str(os.getpid())
        tmp_file_path = os.path.join(config.TEMP_DIR, tmp_file_name)

        tmp_file = open(tmp_file_path, WRITE_MODE)

        value_dict = self.__get_domain_value_dict_setting(entry, method)

        for key, value in value_dict.items():
            tmp_file.write(SMB_CONF_ITEM_FMT_STR % (key, value))
        tmp_file.write("\n")
        tmp_file.close()

        # FIXME 部品ファイル保管場所はTBD
        file_dir = config.SMB_CONF_PART_DIR
        if not os.path.exists(file_dir):
            os.mkdir(file_dir)
        file_path = os.path.join(file_dir, config.SMB_PART_GLOBAL_DOMAIN_FILE)
        reserve_file = config.SMB_PART_GLOBAL_DOMAIN_FILE + ".reserve"
        reserve_path = os.path.join(file_dir, reserve_file)

        if os.path.exists(file_path):
            shutil.move(file_path, reserve_path)
        shutil.move(tmp_file_path, file_path)

        return

    def __regain_domain_part_file(self):

        file_dir = config.SMB_CONF_PART_DIR
        file_path = os.path.join(file_dir, config.SMB_PART_GLOBAL_DOMAIN_FILE)
        reserve_file = config.SMB_PART_GLOBAL_DOMAIN_FILE + ".reserve"
        reserve_path = os.path.join(file_dir, reserve_file)
        if not os.path.exists(reserve_path):
            return WG
        conf = open(reserve_path).read()
        if re.match("^\W*security\W+=\W+user", conf):
            method = WG
        if re.match("^\W*security\W+=\W+domain", conf):
            method = NT
        if re.match("^\W*security\W+=\W+ads", conf):
            method = AD
        shutil.move(reserve_path, file_path)
        return method

    def __create_nsswitch_conf_file(self, method):
        """Create nsswtich.conf for domain."""

        if method == WG:
            src_path = os.path.join(config.SMB_CONF_PART_DIR,
                                    NODOMAIN_NSSWITCH)
        else:
            src_path = os.path.join(config.SMB_CONF_PART_DIR, DOMAIN_NSSWITCH)
        if not os.path.exists(src_path):
            raise NotFoundError("no file", src_path)

        dest_path = os.path.join(config.REAL_CONF_DIR,
                                config.NSSWITCH_CONF_FILE)
        if os.path.exists(dest_path):
            os.remove(dest_path)
        os.symlink(src_path, dest_path)
        return

    def __delete_krb5_conf_file(self):
        """Delete krb5.conf."""

        file_path = os.path.join(config.REAL_CONF_DIR, config.KRB5_CONF_FILE)
        conf_file = open(file_path, "w")
        conf_file.write("")
        conf_file.close()

        return

    def __get_kdc_name(self, entry):
        """Get kerberos domain controler name."""

        return self.__get_ad_dc_name(entry)

    def __get_domain_name(self, entry):
        """Get domain name."""

        ad_dcname = entry.get().get_value(AD_DCNAME_PATH)

        try:
            dc_hostname, dc_domainname = ad_dcname.split(DOMAIN_SEPARATOR, 1)
            """ad_dcname is FQDN. """
            return dc_domainname

        except ValueError:
            """ad_dcname is short host name. """
            return entry.get().get_value(FULLDNS_PATH).lower()

    def __get_krb5_value_dict_setting(self, entry):
        """Krb5 value dict set."""

        value_dict = {}
        value_dict[FULLDNS_KEY] = entry.get().get_value(FULLDNS_PATH)
        value_dict[DOMAIN_CONTROLER_KEY] = self.__get_kdc_name(entry)
        value_dict[DOMAIN_NAME_KEY] = self.__get_domain_name(entry)

        return value_dict

    def __create_krb5_conf_file(self, entry):
        """Create krb5.conf."""

        value_dict = self.__get_krb5_value_dict_setting(entry)

        krb5_conf_str = KRB5_CONF_FMT_STR % \
                {FULLDNS_KEY: value_dict[FULLDNS_KEY],
                DOMAIN_CONTROLER_KEY: value_dict[DOMAIN_CONTROLER_KEY],
                DOMAIN_NAME_KEY: value_dict[DOMAIN_NAME_KEY]}

        # FIXME 部品ファイルtmp保管場所はTBD
        tmp_file_dir = config.TEMP_DIR
        tmp_file_name = config.KRB5_CONF_FILE + \
                FILE_NAME_SEPARATOR + str(os.getpid())
        tmp_file_path = os.path.join(tmp_file_dir, tmp_file_name)
        tmp_file = open(tmp_file_path, WRITE_MODE)
        tmp_file.write(krb5_conf_str)
        tmp_file.close()

        # FIXME 部品ファイル保管場所はTBD
        file_dir = config.REAL_CONF_DIR
        file_name = config.KRB5_CONF_FILE
        file_path = os.path.join(file_dir, file_name)
        shutil.move(tmp_file_path, file_path)

        return

    def __kill_process(self, cmd):
        ps_cmd = [config.PS_CMD, "-ef"]
        proc = CmdExec(ps_cmd, stdout=PIPE, stderr=NULL).execute()
        if proc.wait() != 0:
            raise ExecError
        for line in proc.stdout.readlines():
            if line.find(cmd) == -1:
                continue
            else:
                line_vars = line.split()
                pid = line_vars[1]
                kill_cmd = [config.KILL_CMD, pid]
                CmdExec(kill_cmd, stdout=NULL, stderr=NULL).execute()
        return

    def __logon_to_nt(self):
        """Logon to NT domain."""

        try:
            proc = CmdExec(NT_OLDJOIN_ARGS, stdout=PIPE, stderr=NULL).execute()
            rdlist, wtlist, exclist = select.select(
                                [proc.stdout], [], [], TIMEOUT)
            if not (rdlist or wtlist or exclist):
                self.__kill_process(config.NET_CMD)
                raise
            if proc.wait() != 0:
                raise ExecError

        except ExecError:   # FIXME XRはexec関数でやるせいなのか、
                            # 8ビットシフトしている
            proc = CmdExec(NT_TESTJOIN_ARGS,
                           stdout=PIPE,
                           stderr=NULL).execute()
            rdlist, wtlist, exclist = select.select(
                                [proc.stdout], [], [], TIMEOUT)
            if not (rdlist or wtlist or exclist):
                self.__kill_process(config.NET_CMD)
                raise
            if proc.wait() != 0:
                raise ExecError

    def __logon_to_ad(self, entry):
        """Logon to Active Directory."""

        cmd_args = list(AD_JOIN_ARGS)
        cmd_args[DC_NAME_OPT_IDX] = self.__get_ad_dc_name(entry)
        cmd_args[ADMIN_INFO_OPT_IDX] = ADMIN_INFO_FMT_STR % \
                (entry.get().get_value(AD_ADMINNAME_PATH),
                 entry.get().get_value(AD_ADMINPASSWD_PATH, do_strip=False))

        proc = CmdExec(cmd_args, stdout=PIPE, stderr=NULL).execute()
        rdlist, wtlist, exclist = select.select([proc.stdout], [], [], TIMEOUT)
        if not (rdlist or wtlist or exclist):
            self.__kill_process(config.NET_CMD)
            raise ExecError

        if proc.wait() != 0:
            raise ExecError

    def __assign_to_workgroup(self, entry, method):
        """Assign to WorkGroup"""

        self.__create_domain_part_file(entry, method)
        self.__create_smb_conf_file()

        self.__create_nsswitch_conf_file(method)

        self.__delete_krb5_conf_file()
        entry.set_result(Results.OK)

        return entry

    def __assign_to_ntdomain(self, entry, method):
        """Assign to NT Domain."""

        self.__create_domain_part_file(entry, method)
        self.__create_smb_conf_file()

        self.__create_nsswitch_conf_file(method)

        self.__delete_krb5_conf_file()

        try:
            self.__logon_to_nt()
            entry.set_result(Results.OK)
        except:
            assign = self.__regain_domain_part_file()
            self.__create_nsswitch_conf_file(assign)
            self.__create_smb_conf_file()
            entry.set_result(Results.APPLY_ERROR)

        return entry

    def __assign_to_activedirectory(self, entry, method):
        """Assign to Active Directory."""

        self.__create_domain_part_file(entry, method)
        self.__create_smb_conf_file()

        self.__create_nsswitch_conf_file(method)

        self.__create_krb5_conf_file(entry)

        try:
            self.__logon_to_ad(entry)
            entry.set_result(Results.OK)
        except:
            assign = self.__regain_domain_part_file()
            self.__create_nsswitch_conf_file(assign)
            self.__create_smb_conf_file()
            self.__delete_krb5_conf_file()
            entry.set_result(Results.APPLY_ERROR)

        return entry

    def __set_sharelist(self, smbaccess, host_node):

        host = host_node.get_value(NAME_NAME)
        sharelist_node = host_node.get_node(SHARELIST_NAME)

        try:
            for share in smbaccess.listdir(
                SMB_PREFIX + host, [smbaccess.FILE_SHARE_TYPE]):
                if share.endswith("$"):
                    continue
                sharelist_node.append_child(XmlNode(SHARE_FMT_STR % share))

        except PermissionError:
            pass

    def __set_hostlist(self, smbaccess, domain_node):

        domain = domain_node.get_value(NAME_NAME)
        hostlist_node = domain_node.get_node(HOSTLIST_NAME)

        try:
            for host in smbaccess.listdir(SMB_PREFIX + domain):
                hostlist_node.append_child(XmlNode(HOST_FMT_STR % host))
                host_node = hostlist_node.get_node(
                    HOST_NAME + "[name/text()='%s']" % host)

                try:
                    self.__set_sharelist(smbaccess, host_node)
                except:
                    pass

        except PermissionError:
            pass

        except:
            # FIXME
            # 取得したドメイン一覧に、以前存在したが
            # なくなったドメインがある場合、
            # RuntimeError: (111, 'Connection refused')
            # が発生する。マスタブラウザが
            # あると思ってしまって返してくる?
            pass

    def __set_domainlist(self, smbaccess, domainlist_node):

        for domain in smbaccess.listdir(SMB_PREFIX):
            domainlist_node.append_child(XmlNode(DOMAIN_FMT_STR % domain))
            domain_node = domainlist_node.get_node(
                DOMAIN_NAME + "[name/text()='%s']" % domain)

            self.__set_hostlist(smbaccess, domain_node)

    def __update_global_audit_link(self, accesslog_node):

        link_des = os.path.join(config.SMB_CONF_PART_DIR,
                        config.SMB_PART_GLOBAL_AUDIT_FILE)
        if os.path.exists(link_des):
            os.remove(link_des)

        if accesslog_node.is_enabled():
            link_src = os.path.join(config.SMB_CONF_PART_DIR, ON_AUDIT_FILE)
        else:
            link_src = os.path.join(config.SMB_CONF_PART_DIR, OFF_AUDIT_FILE)

        os.symlink(link_src, link_des)
        return

    def __create_hostdesc_part_file(self, desc):

        tmp_file_name = config.SMB_PART_GLOBAL_HOSTDESC_FILE + \
                FILE_NAME_SEPARATOR + str(os.getpid())
        tmp_file_path = os.path.join(config.TEMP_DIR, tmp_file_name)
        tmp_file = open(tmp_file_path, WRITE_MODE)
        tmp_file.write(HOSTDESC_STR % desc)
        tmp_file.close()

        file_path = os.path.join(config.SMB_CONF_PART_DIR,
                                    config.SMB_PART_GLOBAL_HOSTDESC_FILE)
        shutil.move(tmp_file_path, file_path)
        return

    def __set_hostdesc(self, entry):
        """Host description settings."""

        desc = entry.get().get_value("hostdesc/desc")
        self.__create_hostdesc_part_file(desc)
        self.__create_smb_conf_file()
        entry.set_result(Results.OK)

        return entry

    def __set_msnetwork(self, entry):
        """MSnetwork settings."""

        self.__check_root_node_name(entry, MSNETWORK_NAME)

        method = entry.get().\
                get_node(ASSIGN_NAME).get_attr_value(METHOD_ATTR)

        if method == WG:
            entry = self.__assign_to_workgroup(entry, method)
        elif method == NT:
            entry = self.__assign_to_ntdomain(entry, method)
        elif method == AD:
            entry = self.__assign_to_activedirectory(entry, method)
        else:
            raise InvalidArgError("method = " + method)

        return entry

    def __create_trashbox(self, shares_node):

        # 使用量制限の限界以上使用している状態で
        # ファイル削除すると
        # trashboxフォルダーを作成できないため
        # ファイル削除できない問題の対処

        share_path = shares_node.get_value(PATH_PATH)
        share_uid = shares_node.get_value(UID_PATH)
        share_gid = shares_node.get_value(GID_PATH)
        trashbox_path = os.path.join(share_path, TRASHBOX_NAME)

        if os.path.exists(trashbox_path):
            return

        os.mkdir(trashbox_path)
        os.chmod(trashbox_path, MKDIR_MODE)
        os.chown(trashbox_path, int(share_uid), int(share_gid))

    def __delete_trashbox(self, shares_node):

        share_path = shares_node.get_value(PATH_PATH)
        trashbox_path = os.path.join(share_path, TRASHBOX_NAME)

        if not os.path.exists(trashbox_path):
            return

        shutil.rmtree(trashbox_path, True)

    def __maintenance_trashbox_dir(self, entry):

        shares_node = entry.get()

        try:
            if shares_node.get_node(TRASH_PATH).is_enabled():
                self.__create_trashbox(shares_node)
            else:
                self.__delete_trashbox(shares_node)

        except:
            self.__delete_trashbox(shares_node)

    def __add_handler(self, entry):
        """Add Handler."""

        try:
            self.__check_root_node_name(entry, SHARES_NAME)

            self.__maintenance_trashbox_dir(entry)

            self.__create_share_part_file(entry)
            self.__create_smb_conf_file()

            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.ADD_ERROR)
            self.exception("Add fail!!")

        return entry

    def __delete_handler(self, entry):
        """Delete Handler."""

        try:
            self.__check_root_node_name(entry, SHARES_NAME)

            self.__maintenance_trashbox_dir(entry)

            self.__delete_share_part_file(entry)
            self.__create_smb_conf_file()

            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.DELETE_ERROR)
            self.exception("Delete fail!!")

        return entry

    def __apply_handler(self, entry):
        """Apply hander."""

        try:
            set_type = entry.get().get_node(ANY_NAME).get_node_name()
            if set_type == HOSTDESC_NAME:
                entry = self.__set_hostdesc(entry)
            elif set_type == ASSIGN_NAME:
                entry = self.__set_msnetwork(entry)
            else:
                raise InvalidArgError(set_type)
#            entry.set_result(Results.OK)

        except (IOError, OSError, ExecError):
            entry.set_result(Results.APPLY_ERROR)
            self.exception("Apply fail!!")

        return entry

    def __list_handler(self, entry):

        smbaccess = SmbAccess("", "")

        entry.set(DOMAINLIST_NODE_STR)
        domainlist_node = entry.get().get_node(DOMAINLIST_NAME)

        try:
            self.__set_domainlist(smbaccess, domainlist_node)

        except:
            # FIXME 例外種類、とくになし?
            entry.set_result(Results.QUERY_ERROR)
            self.exception("Query fail!!")

        entry.set_result(Results.OK)

        return entry

    def __update_handler(self, entry):
        """Update Handler."""

        try:
            self.__create_smb_conf_file()
            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.UPDATE_ERROR)
            self.exception("Update fail!!")

        return entry

    def __systemlog_handler(self, entry):
        """Set systemlog Handler."""

        try:
            accesslog_node = entry.get().get_node("accesslog")
            self.__update_global_audit_link(accesslog_node)
            self.__update_share_part_files()
            self.__create_smb_conf_file()
            entry.set_result(Results.OK)

        except (IOError, OSError):
            entry.set_result(Results.SET_ERROR)
            self.exception("Set system log fail!!")

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)
        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.LIST, self.__list_handler)
        self.set_handler(ReqTypes.UPDATE, self.__update_handler)
        self.set_handler(ReqTypes.SET_SYSTEMLOG, self.__systemlog_handler)


#
# Main
#

if __name__ == "__main__":
    PalSamba(option=palcmd.PalCmdOption()).handle_request()
