#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for SHARE configuration."""

import os
import shutil

from iobase.exception import (ExecError, InvalidArgError)

import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)

import pal_config as config


SHARES_NAME = "shares"
SHARE_NAME = "share"
PATH_NAME = "path"
UID_NAME = "uid"
GID_NAME = "gid"

MKDIR_MODE = 0777

INVALID_ROOT_TAG_MSG = "Invalid tag name!!"
ADD_FAILED_MSG = "Fail adding share!!"
MODIFY_FAILED_MSG = "Fail modifying share!!"
DELETE_FAILED_MSG = "Fail deleting share!!"


class PalShares(palcmd.PalCmd):
    """PAL Shares command implementation class."""

    def __chown(self, path, uid, gid):
        os.chown(path, uid, gid)
        for item in os.listdir(path):
            itempath = os.path.join(path, item)
            if os.path.isfile(itempath):
                os.chown(itempath, uid, gid)
            elif os.path.isdir(itempath):
                os.chown(itempath, uid, gid)
                self.__chown(itempath, uid, gid)

    def __check_root_tag(self, entry):
        """Check root tag."""

        root_name = entry.get().get_node_name()

        if root_name != SHARES_NAME:
            raise InvalidArgError(INVALID_ROOT_TAG_MSG)

    def __add_handler(self, entry):
        """Handle add request."""

        try:
            self.__check_root_tag(entry)

            share_node = entry.get().get_node(SHARE_NAME)
            path = share_node.get_node(PATH_NAME).get_text()
            uid = int(share_node.get_value(UID_NAME, default=-1))
            gid = int(share_node.get_value(GID_NAME, default=-1))

            os.mkdir(path, MKDIR_MODE)
            os.chmod(path, MKDIR_MODE)
            try:
                self.__chown(path, uid, gid)
            except:
                pass

            try:
                share_node.delete_child(UID_NAME)
            except:
                pass
            try:
                share_node.delete_child(GID_NAME)
            except:
                pass
            stat = os.stat(path)
            share_node.append_child(UID_NAME, text=str(stat.st_uid))
            share_node.append_child(GID_NAME, text=str(stat.st_gid))
            entry.set_result(Results.OK)

        except (IOError, OSError):
            self.exception(ADD_FAILED_MSG)
            entry.set_result(Results.ADD_ERROR)

        return entry

    def __modify_handler(self, entry):
        """Handle modify request."""

        try:
            self.__check_root_tag(entry)

            share_node = entry.get().get_node(SHARE_NAME)
            path = share_node.get_node(PATH_NAME).get_text()
            uid = int(share_node.get_value(UID_NAME, default=-1))
            gid = int(share_node.get_value(GID_NAME, default=-1))

            try:
                self.__chown(path, uid, gid)
            except:
                pass

            try:
                share_node.delete_child(UID_NAME)
            except:
                pass
            try:
                share_node.delete_child(GID_NAME)
            except:
                pass
            stat = os.stat(path)
            share_node.append_child(UID_NAME, text=str(stat.st_uid))
            share_node.append_child(GID_NAME, text=str(stat.st_gid))
            entry.set_result(Results.OK)

        except (IOError, OSError):
            self.exception(MODIFY_FAILED_MSG)
            entry.set_result(Results.MODIFY_ERROR)

        return entry

    def __delete_handler(self, entry):
        """Handle delete request."""

        try:
            self.__check_root_tag(entry)

            share_node = entry.get().get_node(SHARE_NAME)
            path = share_node.get_node(PATH_NAME).get_text()

            if os.path.isdir(path):
                shutil.rmtree(path)

            entry.set_result(Results.OK)

        except (IOError, OSError):
            self.exception(DELETE_FAILED_MSG)
            entry.set_result(Results.DELETE_ERROR)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.MODIFY, self.__modify_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)

#
# Main
#

if __name__ == "__main__":
    PalShares(option=palcmd.PalCmdOption()).handle_request()
