#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for system setting."""

import os
import re
import stat

from iobase.xmldata import (XmlNode, XmlData, ANY_CHILD_NAME)
from iobase.cmdexec import (CmdExec, PIPE, NULL)
from iobase.exception import (NotFoundError, InvalidArgError)

import libpal.palcmd as palcmd
from libpal.util import RemountRoot

import pal_config as config
from pal_info import (ReqTypes, Results)


### XML Path ###
HOSTNAME_PATH = "./hostdesc/name"

### XML Tag Name ###
SYSINFO_NAME = "sysinfo"
OPTION_NAME = "option"
NAME_NAME = "name"
MODEL_NAME = "model"
SERIES_NAME = "series"
FANRPM_NAME = "fanrpm"
HOST_DESC_NAME = "hostdesc"
FWVERSION_NAME = "fwversion"
TEMPERATURE_NAME = "temperature"
UPTIME_NAME = "uptime"
POWER_NAME = "power"
FAN_NAME = "fan"
DEVICE_NAME = "device"
CONDITION_NAME = "condition"
CONDITION_INSTALLED = "installed"
CONDITION_UPDATED = "updated"
CONDITION_NORMAL = "normal"

### XML Attribute ###
TYPE_ATTR = "type"
SHORTVER_ATTR = "shortver"
DETAIL_ATTR = "detail"

### XML Attribute Value ###
DIMM_ATTR_VAL = "dimm"
MB_ATTR_VAL = "mb"
CPU_ATTR_VAL = "cpu"


### Command and argment ###
#
# XXXXX_ARG_FORMAT is the prototype list of the command which inclues VAR.
# VAR is replaced dynamically in __execute_cmd() method using
# XXXXX_VAR_INDEX_LIST and some variables._FORMAT
#
VAR = "%s"
# FORMAT: df --block-size=1 [volume path]
GET_TEMP_RPM_CMD = [config.GET_SYSTEM_TEMP_CMD]

# FORMAT: hostname
GET_HOSTNAME_CMD = [config.HOSTNAME_CMD]

# FORMAT: hostname
SET_HOSTNAME_ARG_FORMAT = [config.HOSTNAME_CMD, VAR]
SET_HOSTNAME_VAR_INDEX_LIST = [1]

# FORMAT: eeupdate --fanputime
GET_FAN_UPTIME_ARG_FORMAT = [config.EEDATA_CMD, "--fantime"]

# FORMAT: eeupdate --poweruptime
GET_POWER_UPTIME_ARG_FORMAT = [config.EEDATA_CMD, "--powertime"]

# FORMAT: eeupdate --uptime
GET_DEVICE_UPTIME_ARG_FORMAT = [config.EEDATA_CMD, "--uptime"]

# Message
INVALID_ROOT_TAG_MSG = "Specified root tag is invalid !!"

HOSTNAME_FILE = os.path.join(config.SYS_ETC_DIR, "hostname")
HOSTS_FILE = os.path.join(config.SYS_ETC_DIR, "hosts")
KERNEL_HOSTNAME_FILE = "/proc/sys/kernel/hostname"


class PalSysinfo(palcmd.PalCmd):
    """PAL sysinfo module implementation class."""

    def __validate_root_tag_name(self, entry):
        """Validate the root tag name."""

        if SYSINFO_NAME != entry.get().get_node_name():
            raise InvalidArgError("INVALID_ROOT_TAG_MSG")

    def __is_exist_node(self, node, node_path):
        """Check whether node_path exists in specified node"""

        try:
            node.get_node(node_path)
        except NotFoundError:
            return False
        return True

    def __load_value(self, file_path):

        with open(file_path) as fh:
            return fh.read().rstrip()

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_out, std_err,
                      cmd_fmt, cmd_var_idx_list=[], *variables):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_args - variables corresponding to cmd_var_idx_list
        """

        cmd = cmd_fmt[:]
        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            cmd[var_idx] = variables[loop_idx]

        proc = CmdExec(cmd, stdout=std_out, stderr=std_err,
                       raise_exc=is_enable_exception).execute()

        if is_wait:
            proc.wait()

        return proc

    #
    #  Getter funcs
    #

    def __get_series(self):
        """get landisk series name"""

        return self.__load_value(config.SYS_FW_MODEL_PATH)

    def __get_model(self):
        """get landisk model name"""

        return self.__load_value(config.SYS_FULL_PRODUCT__PATH)

    def __get_fw_ver(self):
        """get firmware version"""

        beta_ver = ""
        if os.path.isfile(config.SYS_FW_BETA_VERSION_PATH):
            beta_ver = self.__load_value(config.SYS_FW_BETA_VERSION_PATH)
            if beta_ver:
                beta_ver = "(" + beta_ver + ")"

        return self.__load_value(config.SYS_FW_VERSION_PATH) + beta_ver

    def __get_nowarranty(self):
        """get firmware version"""

        return self.__load_value(config.SYS_NOWARRANTY_PATH)

    def __get_host_name(self):
        """get host name"""

        proc = self.__execute_cmd(True, True, PIPE, NULL, GET_HOSTNAME_CMD)
        return proc.stdout.read().rstrip()

    def __get_temperature_and_rpm(self):
        """get temperature of board"""

        ave_rpm = 3000
        temp = 0
        proc = self.__execute_cmd(True, True, PIPE, NULL, GET_TEMP_RPM_CMD, [])
        for line in proc.stdout:
            try:
                temp = int(line)
                break
            except:
                pass

        return str(temp), str(temp), str(temp), str(ave_rpm)

    #
    #  Append funcs
    #

    def __append_series(self, sysinfo_node):
        """Append "series" node into specified node"""

        sysinfo_node.append_child(SERIES_NAME, self.__get_series())
        return

    def __append_model(self, sysinfo_node):
        """Append "model" node into specified node"""

        sysinfo_node.append_child(MODEL_NAME, self.__get_model())
        return

    def __append_fw_ver(self, sysinfo_node):
        """Append "fwversion" node into specified node"""

        sysinfo_node.append_child(FWVERSION_NAME, self.__get_fw_ver())
        return

    def __append_temperature_and_rpm(self, sysinfo_node):
        """Append "temperature" node into specified node"""

        dimm_temp, mb_temp, cpu_tmp, rpm = self.__get_temperature_and_rpm()

        dimm_temp = str(int(float(dimm_temp)))
        mb_temp = str(int(float(mb_temp)))
        cpu_tmp = str(int(float(cpu_tmp)))

        sysinfo_node.append_child(TEMPERATURE_NAME, dimm_temp,
                                  {TYPE_ATTR: DIMM_ATTR_VAL})

        sysinfo_node.append_child(TEMPERATURE_NAME, mb_temp,
                                  {TYPE_ATTR: MB_ATTR_VAL})

        sysinfo_node.append_child(TEMPERATURE_NAME, cpu_tmp,
                                  {TYPE_ATTR: CPU_ATTR_VAL})

        sysinfo_node.append_child(FANRPM_NAME, rpm)
        return

    def __append_host_info(self, sysinfo_node):
        """Append "hostdesc" node into specified node"""

        hostname = self.__get_host_name()
        sysinfo_node.append_child(HOST_DESC_NAME)
        hostdesc_node = sysinfo_node.get_node(HOST_DESC_NAME)
        hostdesc_node.append_child(NAME_NAME, hostname)
        return

    def __append_condition_info(self, sysinfo_node):
        """Append "condition" into specified node.
           In addition, append "changelist" node if "condition" is "updated".
        """

        condition_node = XmlNode(CONDITION_NAME)
        sysinfo_node.append_child(condition_node)
        if os.path.exists(config.SYS_INSTALLED_FILE_PATH):
            condition_node.set_text(CONDITION_INSTALLED)
        elif os.path.exists(config.SYS_UPDATE_CHANGES_PATH):
            condition_node.set_text(CONDITION_UPDATED)
            try:
                sysinfo_node.append_child(
                    XmlData(config.SYS_UPDATE_CHANGES_PATH).get_root_node())
            except:
                # Ignore Error
                pass
        else:
            condition_node.set_text(CONDITION_NORMAL)

        return

    def __set_execute_bit(self, cmd_path):

        with RemountRoot():
            os.chmod(cmd_path,
                     os.stat(cmd_path).st_mode | stat.S_IXUSR)

    def __clear_execute_bit(self, cmd_path):

        with RemountRoot():
            os.chmod(cmd_path,
                     os.stat(cmd_path).st_mode & ~stat.S_IXUSR)

    def __get_uptime(self):

        def get_uptime():

            uptime_dict = {DEVICE_NAME: 0,
                           FAN_NAME: 0,
                           POWER_NAME: 0}

            # 装置稼働時間
            proc = self.__execute_cmd(True, False, PIPE, NULL,
                                      GET_DEVICE_UPTIME_ARG_FORMAT)
            uptime_dict[DEVICE_NAME] = int(proc.stdout.read().rstrip())
            if 0 != proc.wait():
                uptime_dict[DEVICE_NAME] = 0

            # FAN稼働時間
            proc = self.__execute_cmd(True, False, PIPE, NULL,
                                      GET_FAN_UPTIME_ARG_FORMAT)
            uptime_dict[FAN_NAME] = int(proc.stdout.read().rstrip())
            if 0 != proc.wait():
                uptime_dict[FAN_NAME] = 0

            # 電源稼働時間
            proc = self.__execute_cmd(True, False, PIPE, NULL,
                                      GET_POWER_UPTIME_ARG_FORMAT)
            uptime_dict[POWER_NAME] = int(proc.stdout.read().rstrip())
            if 0 != proc.wait():
                uptime_dict[POWER_NAME] = 0

            return uptime_dict

        uptime_dict = {DEVICE_NAME: 0,
                       FAN_NAME: 0,
                       POWER_NAME: 0}

        try:
            uptime_dict = get_uptime()
        except:
            self.exception("fail to get uptime")
            pass

        return uptime_dict

    def __append_uptime(self, sysinfo_node):

        UPTIME_XML = '<uptime>' + \
                     ' <device><time>' + \
                     '  <hour>%s</hour><minutes>%s</minutes>' + \
                     ' </time></device>' + \
                     ' <power><time>' + \
                     '  <hour>%s</hour><minutes>%s</minutes>' + \
                     ' </time></power>' + \
                     ' <fan><time>' + \
                     '  <hour>%s</hour><minutes>%s</minutes>' + \
                     ' </time></fan>' + \
                     '</uptime>'

        uptime_dict = self.__get_uptime()

        dev_hour = uptime_dict[DEVICE_NAME] / 60
        dev_min = uptime_dict[DEVICE_NAME] % 60

        fan_hour = uptime_dict[FAN_NAME] / 60
        fan_min = uptime_dict[FAN_NAME] % 60

        power_hour = uptime_dict[POWER_NAME] / 60
        power_min = uptime_dict[POWER_NAME] % 60

        uptime_node = XmlNode(UPTIME_XML % (str(dev_hour), str(dev_min),
                                            str(power_hour), str(power_min),
                                            str(fan_hour), str(fan_min)))

        sysinfo_node.append_child(uptime_node)

    def __try_clear_condition(self):
        """Try deleting files for determining the condition."""

        if os.path.exists(config.SYS_INSTALLED_FILE_PATH):
            os.remove(config.SYS_INSTALLED_FILE_PATH)

        if os.path.exists(config.SYS_UPDATE_CHANGES_PATH):
            os.remove(config.SYS_UPDATE_CHANGES_PATH)

    def __query_handler(self, entry):
        """Handle query request."""

        is_short_version = False
        is_detail = False
        try:
            if self.__is_exist_node(entry.get(), OPTION_NAME):
                opt_node = entry.get().get_node(OPTION_NAME)
                is_short_version = opt_node.is_set_to_true(SHORTVER_ATTR)
                is_detail = opt_node.is_set_to_true(DETAIL_ATTR)
        except:
            pass

        try:
            entry.set(XmlNode(SYSINFO_NAME))
            sysinfo_node = entry.get()

            append_node_func_list = [self.__append_series,
                                     self.__append_model,
                                     self.__append_fw_ver,
                                     self.__append_temperature_and_rpm,
                                     self.__append_host_info,
                                     self.__append_condition_info]
            if is_detail:
                append_node_func_list.append(self.__append_uptime)

            for append_node_func in append_node_func_list:
                append_node_func(sysinfo_node)

        except:
            self.error("Cannot get system info!!")
            raise

        if is_short_version:
            fwver = sysinfo_node.get_node(FWVERSION_NAME).get_text()
            match = re.match("(.*)\(.*\)", fwver)  # 1.00(yy/mm/dd) => 1.00
            if match:
                sysinfo_node.get_node(FWVERSION_NAME).set_text(match.group(1))

        entry.set_result(Results.OK)
        return entry

    def __apply_handler(self, entry):
        """Handle apply request."""

        self.__validate_root_tag_name(entry)
        sysinfo_node = entry.get()
        host_name = ""

        try:
            host_name = sysinfo_node.get_node(HOSTNAME_PATH).get_text()
        except:
            entry.set_result(Results.OK)
            return entry

        proc = self.__execute_cmd(True, True, PIPE, NULL,
                                  SET_HOSTNAME_ARG_FORMAT,
                                  SET_HOSTNAME_VAR_INDEX_LIST,
                                  host_name)
        if proc.returncode != 0:
            entry.set_result(Results.EXECUTE_ERROR)
            return entry

        with RemountRoot():
            with open(HOSTNAME_FILE, "w") as fh:
                fh.write(host_name)
            self.__execute_cmd(True, True, PIPE, NULL, config.UPDATE_HOSTS_CMD)

        with open(KERNEL_HOSTNAME_FILE, "w") as fh:
            fh.write(host_name)

        entry.set_result(Results.OK)
        return entry

    def __clear_handler(self, entry):
        """Handle clear request."""

        self.__validate_root_tag_name(entry)
        try:
            for child in entry.get().get_node_list(ANY_CHILD_NAME):
                if child.get_node_name() == CONDITION_NAME:
                    self.__try_clear_condition()

            entry.set_result(Results.OK)
        except:
            entry.set_result(Results.CLEAR_ERROR)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler("", self.__query_handler)
        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.CLEAR, self.__clear_handler)


#
# Main
#

if __name__ == "__main__":
    PalSysinfo(option=palcmd.PalCmdOption()).handle_request()
