#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Volume Manager."""

from __future__ import with_statement

import os
import glob
import shutil
import stat
import time
import re
import copy
import libpal.palcmd as palcmd
import pal_config as config
import libpal.util as util
import iobase.fileut as fileut
from libpal.util import RemountRoot
from iobase.exception import (NotFoundError, InvalidArgError, ExecError)
from pal_info import ReqTypes
from pal_info import Results
from iobase.xmldata import XmlNode
from iobase.cmdexec import (CmdExec, PIPE, NULL)
from pal_config import (DF_CMD, MOUNT_CMD, UMOUNT_CMD, CRYPTSETUP_CMD,
                        SWAPON_CMD, SWAPOFF_CMD, MKSWAP_CMD, BLKID_CMD,
                        MKFS_EXT3_CMD, MKFS_EXT4_CMD, MKFS_XFS_CMD,
                        MKFS_MSDOS_CMD, PARTED_CMD, LS_CMD, SYNC_CMD)
from albg_info import (ALBG_MKFS_PATH, ALBG_FSCK_PATH)
from pal_config import (DISK_X_SIZE_FILE, VFS_SYNC_PROGRESS)
from pal_config import (DEV_DIR, DEV_TO_VOL, DEV_TO_ENC_DEV, VOL_INFO,
                        STOR_IF_TO_VOL, MOUNT_OPT_INFO, EXT_RAID_DEV_INFO)
from pal_config import (LOCK_DIR, MOUNT_LOCK_FILE)

ENC_DIR = "mapper"

### X Path ###
PART_NAME_XPATH = "./name[@type='partition']"
ENC_NAME_XPATH = "./name[@type='crypt']"
DISK_NAME_XPATH = "./name[@type='disk']"

### XML Tag Name ###
VOLCTRL_NAME = "volctrl"
NAME_NAME = "name"
VOLUME_NAME = "volume"
KEY_PATH_NAME = "keypath"
DST_NAME = "dst"
SRC_NAME = "src"
CAPACITY_NAME = "capacity"
TOTAL_NAME = "total"
USE_NAME = "use"
OPTION_NAME = "option"
PARTITIONLIST_NAME = "partitionlist"
PARTITION_NAME = "partition"
TIMESTAMP_NAME = "timestamp"
UUID_NAME = "uuid"
MIRR_SET_NAME = "mirrset"

# XML Attribute
TYPE_ATTR = "type"
OPERATION_OPT = "operation"
WAIT_ATTR = "wait"
BIND_ATTR = "bind"
ENCRYPT_ATTR = "encrypt"
CLEAR_ATTR = "clear"
VFSREMOVE_ATTR = "vfsremove"
VFSADD_ATTR = "vfsadd"
VFS_MULTIPLE_ADD_ATTR = "vfsmultiadd"
MKNONRAID_ATTR = "mknonraid"
INCERRCNT_ATTR = "incerrcnt"
EXTERNAL_VFSEXT_ATTR = "externalvfsext"

# XML Attribute Value
PARTITION_ATTR_VAL = "partition"
SWAP_CREATE_ATTR_VAL = "swapcreate"
SWAP_ON_ATTR_VAL = "swapon"
SWAP_OFF_ATTR_VAL = "swapoff"
SWAP_GET_ATTR_VAL = "swapget"
ENC_START_ATTR_VAL = "encstart"
ENC_STOP_ATTR_VAL = "encstop"
WRITE_ENC_SIG_ATTR_VAL = "writeencsig"
CHK_ENC_SIG_ATTR_VAL = "chkencsig"
MOUNT_ATTR_VAL = "mount"
UNMOUNT_ATTR_VAL = "umount"
CREATE_ATTR_VAL = "create"
DELETE_ATTR_VAL = "delete"
START_ATTR_VAL = "start"
STOP_ATTR_VAL = "stop"
CRYPT_ATTR_VAL = "crypt"
CHK_FILE_EXIST_ATTR_VAL = "checkfileexist"
GET_DISK_LABEKL_ATTR_VAL = "getdisklabel"
SET_DISK_LABEKL_ATTR_VAL = "setdisklabel"
VFSEXT_ELEMENT_INFO_ATTR_VAL = "vfsextelementinfo"

### Command and argment ###
#
# XXXXX_ARG_FORMAT is the prototype list of the command which inclues VAR.
# VAR is replaced dynamically in __execute_cmd() method using
# XXXXX_VAR_INDEX_LIST and some variables._FORMAT
#
VAR = "%s"
# FORMAT: df --block-size=1 --portability [volume path]
DF_BLOCK_ARG_FORMAT = [DF_CMD, "--block-size=1", "--portability", VAR]
DF_BLOCK_VAR_INDEX_LIST = [3]

# FORMAT: mkswap [partition path]
MKSWAP_ARG_FORMAT = [MKSWAP_CMD, VAR]
MKSWAP_VAR_INDEX_LIST = [1]

# FORMAT: swapon [partition path]
SWAPON_ARG_FORMAT = [SWAPON_CMD, VAR]
SWAPON_VAR_INDEX_LIST = [1]

# FORMAT: swapoff [partition path]
SWAPOFF_ARG_FORMAT = [SWAPOFF_CMD, VAR]
SWAPOFF_VAR_INDEX_LIST = [1]

# FORMAT: cryptsetup.sh -h plain -d [keypath] create  [enc] [src]
CRYPT_START_ARG_FORMAT = \
    [CRYPTSETUP_CMD, "-h", "plain", "-d", VAR, "create", VAR, VAR]
CRYPT_SATRT_VAR_INDEX_LIST = [4, 6, 7]

# FORMAT: cryptsetup.sh -h plain remove [enc]
CRYPT_STOP_ARG_FORMAT = [CRYPTSETUP_CMD, "-h", "plain", "remove", VAR]
CRYPT_STOP_VAR_INDEX_LIST = [4]

# FORMAT: mount -t [type] -o [option] [partition path] [destination pth]
MOUNT_ARG_FORMAT = [MOUNT_CMD, "-t", VAR, "-o", VAR, VAR, VAR]
MOUNT_VAR_INDEX_LIST = [2, 4, 5, 6]

# FORMAT: mount -t [type] [partition path] [destination pth]
MOUNT_NON_OPTION_ARG_FORMAT = [MOUNT_CMD, "-t", VAR, VAR, VAR]
MOUNT_NON_OPTION_VAR_INDEX_LIST = [2, 3, 4]

# FORMAT: mount -t vfsext -o [elementlist] none [destination pth]
VFS_MOUNT_ARG_FORMAT = [MOUNT_CMD,  "-t", "vfsext", "-o", VAR, "none",
                        config.VOL_INFO["sataraid1"]["mnt_point"]]
VFS_MOUNT_VAR_INDEX_LIST = [4]

# FORMAT: mount -o remount,[elementlist] none [destination pth]
VFS_REMOUNT_ARG_FORMAT = [MOUNT_CMD,  "-o", "remount," + VAR, "none",
                          config.VOL_INFO["sataraid1"]["mnt_point"]]
VFS_REMOUNT_VAR_INDEX_LIST = [2]

# FORMAT: mount -n --bind [src path] [dst pth]
BIND_MOUNT_ARG_FORMAT = [MOUNT_CMD, "-n", "--bind", VAR, VAR]
BIND_MOUNT_VAR_INDEX_LIST = [3, 4]

# FORMAT: umount [destination pth]
UNMOUNT_ARG_FORMAT = [UMOUNT_CMD, VAR]
UNMOUNT_VAR_INDEX_LIST = [1]

# FORMAT: blkid [partition]
BLKID_ARG_FORMAT = [BLKID_CMD, "-p", VAR]
BLKID_VAR_INDEX_LIST = [2]

# FORMAT: mkfs.ext3 -m 0 -F [partition path]
MKFS_EXT3_ARG_FORMAT = [MKFS_EXT3_CMD, "-m", "0", "-F", VAR]
MKFS_EXT3_VAR_INDEX_LIST = [4]

MKFS_EXT4_ARG_FORMAT = [MKFS_EXT4_CMD, "-m", "0", "-F", VAR]
MKFS_EXT4_VAR_INDEX_LIST = [4]

# FORMAT: mkfs.xfs -f [partition path]
MKFS_XFS_ARG_FORMAT = [MKFS_XFS_CMD, "-f", VAR]
MKFS_XFS_VAR_INDEX_LIST = [2]

# FORMAT: mkfs.vfat -F 32 -I [partition path]
MKFS_FAT_ARG_FORMAT = [MKFS_MSDOS_CMD, "-F", "32", "-I", VAR]
MKFS_FAT_VAR_INDEX_LIST = [4]

# FORMAT: parted VAR unit s p
GET_PARTITION_ARG_FORMAT = [PARTED_CMD, VAR, "unit", "s", "print"]
GET_PARTITION_VAR_INDEX_LIST = [1]

# Some value
MOUNTED = "mounted"
UNMOUNTED = "unmounted"
UNFORMATED = "unformated"
STATUS_NAME = "status"
USE_SIZE = "use"
TOTAL_SIZE = "total"
MOUNT_DST = "mountdst"
FILESYSTEM = "filesystem"
EXT3_FSTYPE = "ext3"
EXT4_FSTYPE = "ext4"
XFS_FSTYPE = "xfs"
FAT_FSTYPE = "fat"
NTFS_FSTYPE = "ntfs"
NTFS_MOUNTING_FSTYPE = "ntfs-fuse"  # mount -t で指定するfstype
NTFS_MOUNTED_FSTYPE = "fuseblk"  # /proc/mountsで見えるfstype
VFSEXT_FSTYPE = "vfsext"

MBR_PART_START = 2048
GPT_PART_START = 40

FULL_ACCESS_FILE_MODE = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | \
                        stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP | \
                        stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH


class PalVolume(palcmd.PalCmd):
    """Pal Volume command implementation class."""

    #
    # Sub functions
    #

    class __CreateError(Exception):
        "Create Error"
        pass

    class __StartError(Exception):
        "Start Error"
        pass

    class __StopError(Exception):
        "Stop Error"
        pass

    class __DeviceNotFound(Exception):
        "Device Not Found"
        pass

    class __OnError(Exception):
        "On Error"
        pass

    class __OffError(Exception):
        "Off Error"
        pass

    class __MountError(Exception):
        "Mount Error"
        pass

    class __UnmountError(Exception):
        "Unmount Error"
        pass

    class __SignatureNotFound(Exception):
        "Signature Not Found"
        pass

    class __UnformatedPartition(Exception):
        "Unformated Partition"
        pass

    class __InternalError(Exception):
        "Internal Error"
        pass

    class __MakeError(Exception):
        "Make Error"
        pass

    class __FileNotExist(Exception):
        "File Not Exist"
        pass

    class __AlreadyMount(Exception):
        "Already Mount"
        pass

    class __AlreadyUnmount(Exception):
        "Already Unmount"
        pass

    def __validate_root_tag_name(self, entry):
        """Validate the root tag name."""

        node_name = entry.get().get_node_name()
        if VOLCTRL_NAME != node_name:
            self.error("Invalid root node name %s" % node_name)
            raise InvalidArgError("invalid root tag..")

    def __is_exist_node(self, node, node_path):
        """Check whether node_path exists in specified node"""

        try:
            node.get_node(node_path)
        except NotFoundError:
            return False
        return True

    def __load_value(self, file_name):
        """load value from specified file"""

        with open(file_name) as fh:
            return fh.read().rstrip()

    def __convert_real_devname(self, dev_name):
        """Convert synbol device name to real device name.
           If dev_name is already real this method just returns it.
        """

        dev_path = os.path.join(DEV_DIR, dev_name)
        return os.path.basename(os.path.realpath(dev_path))

    def __force_make_directory(self, dir_path):
        """make specified directory."""

        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

    def __copy_all_files(self, src_dir, dst_dir):

        file_list = glob.glob(os.path.join(src_dir, "*"))
        for file in file_list:
            if os.path.isdir(file):
                tmp_dir = os.path.join(dst_dir, os.path.basename(file))

                try:
                    shutil.copytree(file, tmp_dir)
                except:
                    pass

                continue

            try:
                shutil.copy(file, dst_dir)
            except:
                pass

        return

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_out, std_err,
                      cmd_fmt, cmd_var_idx_list=[], variable_list=[]):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_list - variables corresponding to cmd_var_idx_list
        """
        cmd = cmd_fmt[:]

        if len(cmd_var_idx_list) != len(variable_list):
            raise ExecError("")

        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            # cmd[var_idx] = variable_list[loop_idx]
            cmd[var_idx] = cmd[var_idx] % variable_list[loop_idx]

        return CmdExec(cmd,
                       stdout=std_out,
                       stderr=std_err,
                       wait=is_wait,
                       raise_exc=is_enable_exception,
                       close_fds=True).execute()

    def __analysis_df_result(self, result):
        """Analysis the result of df command"""

        info_dict = {TOTAL_SIZE: "0", USE_SIZE: "0"}
        for line in result.split("\n"):
            if re.match(r"^Filesystem", line):
                continue

            match = re.match("\S+"
                             "\s+"
                             "(\S+)"
                             "\s+"
                             "(\S+)"
                             "\s+"
                             "\S+"
                             "\s+"
                             "\S+"
                             "\s+"
                             "\S+", line)
            if not match:
                continue

            info_dict[TOTAL_SIZE] = match.group(1)
            info_dict[USE_SIZE] = match.group(2)

        return info_dict

    def __get_query_info_by_df(self, mnt_path):
        """Ge volume infomation by df command"""

        proc = self.__execute_cmd(False, False,
                                  PIPE, NULL,
                                  DF_BLOCK_ARG_FORMAT,
                                  DF_BLOCK_VAR_INDEX_LIST,
                                  [mnt_path])
        if 0 != proc.wait():
            raise

        result = proc.stdout.read()
        return self.__analysis_df_result(result)

    def __get_query_info_by_mount(self, part_path):
        """Get volume infomation by df command"""

        info_dict = {}
        try:
            (dev, mnt_point, fs, opts) = \
                util.get_mount_info(part_path)
            info_dict[STATUS_NAME] = MOUNTED
            info_dict[MOUNT_DST] = mnt_point
            if NTFS_MOUNTED_FSTYPE == fs:
                fs = NTFS_FSTYPE
            info_dict[FILESYSTEM] = fs
        except:
            info_dict[STATUS_NAME] = UNMOUNTED

        return info_dict

    def __get_partition_type(self, disk_path):
        """Get partition type."""

        if not os.path.exists(disk_path):
            raise self.__DeviceNotFound

        proc = self.__execute_cmd(False, False, PIPE, NULL,
                                  GET_PARTITION_ARG_FORMAT,
                                  GET_PARTITION_VAR_INDEX_LIST,
                                  [disk_path])

        part_table = ""
        for line in  proc.stdout.readlines():
            line = line.rstrip()
            if not part_table:
                match = re.match("Partition Table:\s*(\S+)", line)
                if match:
                    return match.group(1)
                continue

        return ""

    def __get_unmounted_volume_info(self, part_path, mount_dst_path):
        """return [status] [fstype] [totalcapa] [usecapa]"""

        fstype = self.__get_filesystem_type(part_path)
        if not fstype:
            raise self.__UnformatedPartition

        if fstype == NTFS_FSTYPE:
            fstype = NTFS_MOUNTING_FSTYPE

        if fstype not in MOUNT_OPT_INFO:
            raise self.__UnformatedPartition

        # マウントして情報取得
        mount_opt = MOUNT_OPT_INFO[fstype]
        proc = self.__execute_cmd(False, True, NULL, NULL,
                                  MOUNT_ARG_FORMAT, MOUNT_VAR_INDEX_LIST,
                                  [fstype, mount_opt,
                                  part_path, mount_dst_path])
        if 0 != proc.returncode:
            raise self.__UnformatedPartition

        query_info_dict = {}
        query_info_dict.update(self.__get_query_info_by_mount(part_path))

        mnt_path = query_info_dict.get(MOUNT_DST)
        if mnt_path:
            query_info_dict.update(self.__get_query_info_by_df(mnt_path))

        # 情報取得後はアンマウント
        self.__execute_cmd(False, True, NULL, NULL,
                           UNMOUNT_ARG_FORMAT, UNMOUNT_VAR_INDEX_LIST,
                           [mnt_path])

        # 重要
        query_info_dict[STATUS_NAME] = UNMOUNTED
        return query_info_dict

    def __get_internal_partition_name(self):

        return VOL_INFO[STOR_IF_TO_VOL["internal"][1]]["dev"]

    def __get_ptable_and_last_lba(self, disk_name):

        real_disk_name = self.__convert_real_devname(disk_name)
        disk_path = os.path.join(DEV_DIR, real_disk_name)
        ptable = self.__get_partition_type(disk_path)

        disk_size = \
                int(self.__load_value(DISK_X_SIZE_FILE % real_disk_name))

        if "gpt" == ptable:
            last_lba = disk_size - GPT_PART_START - 1 \
                - config.LANDISK_INFO_SECTOR_SIZE

        elif "msdos" == ptable:
            last_lba = (disk_size / 255 / 63) * 255 * 63 - 1 \
                - config.LANDISK_INFO_SECTOR_SIZE
        else:
            raise self.__InternalError

        return last_lba

    def __is_encrypted_volume(self, disk_name):

        real_disk_name = self.__convert_real_devname(disk_name)
        info_lba = self.__get_ptable_and_last_lba(real_disk_name) + 1

        disk_path = os.path.join(DEV_DIR, real_disk_name)

        with open(disk_path, "r") as disk_file:
            # FIXME sectorサイズ決めうち
            disk_file.seek(info_lba * 512)

            # FIXME ちゃんと構造体使って管理すべき
            signature = disk_file.read(len(config.ENCRYPT_SIGNATURE))
            if signature == config.ENCRYPT_SIGNATURE:
                return True

        return False

    def __append_partition_info(self, vol_node):
        """get volume info from tmp file"""

        part_name = vol_node.get_node(PART_NAME_XPATH).get_text()

        if part_name in DEV_TO_ENC_DEV.values():
            part_path = os.path.join(DEV_DIR, ENC_DIR, part_name)
        else:
            part_path = os.path.realpath(os.path.join(DEV_DIR, part_name))

        if not os.path.exists(part_path):
            raise self.__DeviceNotFound

        query_info_dict = {}
        query_info_dict.update(self.__get_query_info_by_mount(part_path))

        mnt_path = query_info_dict.get(MOUNT_DST)
        if mnt_path:
            query_info_dict.update(self.__get_query_info_by_df(mnt_path))

        capa_node = XmlNode(CAPACITY_NAME)
        for tag_name, value in query_info_dict.items():
            if TOTAL_SIZE == tag_name or USE_SIZE == tag_name:
                capa_node.append_child(tag_name, value)
                continue
            vol_node.append_child(tag_name, value)

        if 0 < len(capa_node.get_node_list("*")):
            vol_node.append_child(capa_node)

        return

    def __get_mounted_part_name(self, part_name_list):

        for part_name in part_name_list:

            query_info_dict = {}
            part_path = os.path.join(DEV_DIR, part_name)

            # mapper(暗号化デバイス)の実デバイス名は
            # mount情報に存在しないので
            if ENC_DIR not in part_path:
                part_path = os.path.realpath(part_path)

            if not os.path.exists(part_path):
                continue

            query_info_dict.update(self.__get_query_info_by_mount(part_path))

            if MOUNTED == query_info_dict.get(STATUS_NAME):
                return part_name

        return ""

    def __append_usable_partition_info(self, vol_node):
        """get usable partition info from tmp file"""

        disk_name = vol_node.get_value(DISK_NAME_XPATH, None)
        if not os.path.exists(os.path.join(DEV_DIR, disk_name)):
            raise self.__DeviceNotFound

        vol_name = DEV_TO_VOL[disk_name]

        part_name_list = []
        # 暗号化パーティション
        part_name_list.append(
            os.path.join(ENC_DIR, DEV_TO_ENC_DEV[disk_name]))

        # Super floppyか内蔵RAIDドライブ
        part_name_list.append(disk_name)

        # USBディスクの場合はusbdxp1~p15まで
        # 使えるパーティションをサーチ
        if VOL_INFO[vol_name]["part_base"]:
            for part_num in range(1, 16):
                part_name_list.append(
                    VOL_INFO[vol_name]["part_base"] + str(part_num))

        mouted_part_name = self.__get_mounted_part_name(part_name_list)
        if mouted_part_name:
            chk_tgt_part_name_list = [mouted_part_name]  # only
        else:
            chk_tgt_part_name_list = part_name_list

        for part_name in chk_tgt_part_name_list:

            query_info_dict = {}
            part_path = os.path.join(DEV_DIR, part_name)

            # mapper(暗号化デバイス)の実デバイス名は
            # mount情報に存在しないので
            if ENC_DIR not in part_path:
                part_path = os.path.realpath(part_path)

            if not os.path.exists(part_path):
                continue

            # マウント情報取得
            query_info_dict.update(self.__get_query_info_by_mount(part_path))

            # アンマウント状態の場合でもとりあえず
            # 情報取得を試みる
            if UNMOUNTED == query_info_dict.get(STATUS_NAME):
                tmp_mnt_path = VOL_INFO[vol_name]["mnt_point"]
                with RemountRoot():
                    try:
                        with fileut.FLock(LOCK_DIR, MOUNT_LOCK_FILE, True):
                            query_info_dict = \
                                self.__get_unmounted_volume_info(part_path,
                                                                 tmp_mnt_path)
                    except self.__UnformatedPartition:
                        continue
            else:
                mnt_path = query_info_dict.get(MOUNT_DST)
                if mnt_path:
                    query_info_dict.update(
                        self.__get_query_info_by_df(mnt_path))

            capa_node = XmlNode(CAPACITY_NAME)
            for tag_name, value in query_info_dict.items():
                if TOTAL_SIZE == tag_name or USE_SIZE == tag_name:
                    capa_node.append_child(tag_name, value)
                    continue
                vol_node.append_child(tag_name, value)

            if 0 < len(capa_node.get_node_list("*")):
                vol_node.append_child(capa_node)

            part_name_node = XmlNode('<name type="partition"/>')
            if ENC_DIR in part_path:
                part_name_node.set_text(os.path.basename(part_path))
                part_name_node.set_attrs({ENCRYPT_ATTR: "true"})
            else:
                part_name_node.set_text(part_name)
                part_name_node.set_attrs({ENCRYPT_ATTR: "false"})

            vol_node.append_child(part_name_node)
            return  # 最初にmount情報がとれればそれを採用

        # マウントされているパーティションが
        # 見つからなかった場合
        vol_node.append_child(STATUS_NAME, UNFORMATED)
        return

    def __get_filesystem_type(self, part_path):
        """Get filesystem type of specified partition."""

        for idx in range(2):
            proc = self.__execute_cmd(False, True,
                                      PIPE, NULL,
                                      BLKID_ARG_FORMAT,
                                      BLKID_VAR_INDEX_LIST,
                                      [os.path.realpath(part_path)])

            for line in proc.stdout:
                line = line.rstrip()
                match = re.search("\sTYPE=\"(\S*)\"", line)
                if match:
                    return match.group(1)
                time.sleep(1)

        return ""

    def __get_enable_vfs_element(self):

        try:
            (dev, mnt_point, fs, mount_opts) = \
                util.get_mount_info("/mnt/sataraid1")
        except:
            return []

        element_list = []
        for element, info in EXT_RAID_DEV_INFO.iteritems():
            if info["mnt_point"] in mount_opts:
                element_list.append(element)
        return element_list

    def __mount_vfs_element(self, is_enc, element_name, element_info):

        match = re.match("[a-zA-Z]+([0-9]+)p[0-9]+", element_name)
        if not match:
            raise self.__InternalError

        mount_dst_path = element_info["mnt_point"]

        if not os.path.isdir(mount_dst_path):
            os.makedirs(mount_dst_path)

        if is_enc:
            part_path = \
                os.path.join(DEV_DIR, ENC_DIR, DEV_TO_ENC_DEV[element_name])
        else:
            part_path = os.path.join(DEV_DIR, element_name)

        mount_opt = MOUNT_OPT_INFO[config.VFSEXT_INFO["filesystem"]]

        if util.is_mounted(mount_dst_path):
            raise self.__AlreadyMount

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  MOUNT_ARG_FORMAT,
                                  MOUNT_VAR_INDEX_LIST,
                                  [config.VFSEXT_INFO["filesystem"],
                                  mount_opt,
                                  part_path,
                                  mount_dst_path])
        if 0 != proc.returncode:
            raise self.__MountError
        return

    def __get_vfsext_mount_option(self, vfs_dev_info):

        max_mirrset_major = 0
        for info in vfs_dev_info.values():
            if max_mirrset_major < info["mirrset"][0]:
                max_mirrset_major = info["mirrset"][0]

        mirr_mnt_hash = {}
        for major in range(1, max_mirrset_major + 1):
            mirr_mnt_hash[major * 10 + 1] = ""
            mirr_mnt_hash[major * 10 + 2] = ""

        for info in vfs_dev_info.values():
            hash_key = info["mirrset"][0] * 10 + info["mirrset"][1]
            mirr_mnt_hash[hash_key] = info["mnt_point"]

        option_str = ""
        for major in range(1, max_mirrset_major + 1):
            option_str += "%s:%s," % \
                              (mirr_mnt_hash[major * 10 + 1],
                               mirr_mnt_hash[major * 10 + 2])

        return option_str + "usrquota,grpquota"

    def __do_vfsext_mount_by_specify_elements(self, entry):

        VFSEXT_PART_XPATH = "./volume/vfsext/partitionlist/partition"

        vol_node = entry.get().get_node(VOLUME_NAME)
        is_enc = vol_node.get_node(FILESYSTEM).is_set_to_true(ENCRYPT_ATTR)

        part_node_list = entry.get().get_node_list(VFSEXT_PART_XPATH)
        tgt_element_list = [node.get_text() for node in part_node_list]

        vfs_dev_info = copy.deepcopy(config.EXT_RAID_DEV_INFO)

        for element_name, element_info in config.EXT_RAID_DEV_INFO.iteritems():
            if element_name not in tgt_element_list:
                if DEV_TO_ENC_DEV[element_name] not in tgt_element_list:
                    vfs_dev_info[element_name]["mnt_point"] = ""
                    continue

            try:
                self.__mount_vfs_element(is_enc, element_name, element_info)
            except self.__AlreadyMount:
                pass
            except:
                self.exception("fail to __mount_vfs_element(%s)" % \
                               element_name)
                vfs_dev_info[element_name]["mnt_point"] = ""

        mount_option_str = self.__get_vfsext_mount_option(vfs_dev_info)
        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  VFS_MOUNT_ARG_FORMAT,
                                  VFS_MOUNT_VAR_INDEX_LIST,
                                  [mount_option_str])
        if 0 != proc.returncode:
            raise self.__MountError
        return

    def __simple_mount(self, part_path, mount_dst_path):

        filesystem_type = self.__get_filesystem_type(part_path)
        if not filesystem_type:
            raise self.__MountError

        # ntfsの場合だけ特別対応
        # (product_confでもっとうまいことやれるはず)
        if filesystem_type == NTFS_FSTYPE:
            filesystem_type = NTFS_MOUNTING_FSTYPE

        if filesystem_type not in MOUNT_OPT_INFO:
            raise self.__MountError

        mount_opt = MOUNT_OPT_INFO[filesystem_type]

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  MOUNT_ARG_FORMAT,
                                  MOUNT_VAR_INDEX_LIST,
                                  [filesystem_type,
                                  mount_opt,
                                  part_path,
                                  mount_dst_path])
        if 0 != proc.returncode:
            raise self.__MountError

        return

    def __simple_unmount(self, mount_dst_path):

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  UNMOUNT_ARG_FORMAT,
                                  UNMOUNT_VAR_INDEX_LIST,
                                  [mount_dst_path])
        if 0 != proc.returncode:
            raise self.__UnmountError

        return

    def __standard_mount(self, entry):

        vol_node = entry.get().get_node(VOLUME_NAME)
        part_name = vol_node.get_node(PART_NAME_XPATH).get_text()
        part_path = util.symbl_dev_name2dev_path(part_name)
        mount_dst_path = vol_node.get_node(DST_NAME).get_text()

        if util.is_mounted(mount_dst_path):
            raise self.__AlreadyMount

        # VFSEXTの場合は特殊なマウント
        if VFSEXT_FSTYPE == vol_node.get_value(FILESYSTEM, ""):
            self.__do_vfsext_mount_by_specify_elements(entry)
            return

        filesystem_type = self.__get_filesystem_type(part_path)
        if not filesystem_type:
            raise self.__MountError

        # ntfsの場合だけ特別対応
        # (product_confでもっとうまいことやれるはず)
        if filesystem_type == NTFS_FSTYPE:
            filesystem_type = NTFS_MOUNTING_FSTYPE

        if filesystem_type not in MOUNT_OPT_INFO:
            raise self.__MountError

        mount_opt = MOUNT_OPT_INFO[filesystem_type]

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  MOUNT_ARG_FORMAT,
                                  MOUNT_VAR_INDEX_LIST,
                                  [filesystem_type,
                                  mount_opt,
                                  part_path,
                                  mount_dst_path])
        if 0 != proc.returncode:
            raise self.__MountError

        time.sleep(0.5)
        if util.is_mounted_as_read_only(mount_dst_path):
            return

        os.chmod(mount_dst_path, FULL_ACCESS_FILE_MODE)

        return

    def __bind_mount(self, entry):

        vol_node = entry.get().get_node(VOLUME_NAME)
        mount_src_path = vol_node.get_node(SRC_NAME).get_text()
        mount_dst_path = vol_node.get_node(DST_NAME).get_text()

        if util.is_mounted(mount_dst_path):
            raise self.__AlreadyMount

        self.__force_make_directory(mount_dst_path)

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  BIND_MOUNT_ARG_FORMAT,
                                  BIND_MOUNT_VAR_INDEX_LIST,
                                  [mount_src_path, mount_dst_path])

        if 0 != proc.returncode:
            raise self.__MountError

        if util.is_mounted_as_read_only(mount_src_path):
            return

        os.chmod(mount_dst_path, FULL_ACCESS_FILE_MODE)
        return

    def __vfs_remove(self, entry):

        REMOVE_PART_XPATH = "./volume/partition/name"

        remove_element_name = \
            entry.get().get_node(REMOVE_PART_XPATH).get_text()

        enable_vfs_element_list = self.__get_enable_vfs_element()
        if remove_element_name in enable_vfs_element_list:
            enable_vfs_element_list.remove(remove_element_name)

        vfsext_dev_info = {}
        for element in enable_vfs_element_list:
            vfsext_dev_info[element] = EXT_RAID_DEV_INFO[element]

        mount_option_str = self.__get_vfsext_mount_option(vfsext_dev_info)
        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  VFS_REMOUNT_ARG_FORMAT,
                                  VFS_REMOUNT_VAR_INDEX_LIST,
                                  [mount_option_str])
        if 0 != proc.returncode:
            raise self.__MountError
        return

    def __vfs_add(self, entry):

        ADD_PART_XPATH = "./volume/partition/name"

        vol_node = entry.get().get_node(VOLUME_NAME)
        is_enc = vol_node.get_node(FILESYSTEM).is_set_to_true(ENCRYPT_ATTR)

        add_element_name = entry.get().get_node(ADD_PART_XPATH).get_text()

        enable_vfs_element_list = self.__get_enable_vfs_element()
        if add_element_name not in enable_vfs_element_list:
            enable_vfs_element_list.append(add_element_name)

        try:
            self.__mount_vfs_element(is_enc,
                                     add_element_name,
                                     EXT_RAID_DEV_INFO[add_element_name])
        except self.__AlreadyMount:
            pass
        except:
            self.exception("fail to __mount_vfs_element(%s)" % \
                           add_element_name)

        vfsext_dev_info = {}
        for element in enable_vfs_element_list:
            vfsext_dev_info[element] = EXT_RAID_DEV_INFO[element]

        mount_option_str = self.__get_vfsext_mount_option(vfsext_dev_info)
        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  VFS_REMOUNT_ARG_FORMAT,
                                  VFS_REMOUNT_VAR_INDEX_LIST,
                                  [mount_option_str])
        if 0 != proc.returncode:
            raise self.__MountError
        return

    def __vfs_multiple_add(self, entry):

        ADD_PART_XPATH = "./volume/partitionlist/partition/name"

        vol_node = entry.get().get_node(VOLUME_NAME)
        is_enc = vol_node.get_node(FILESYSTEM).is_set_to_true(ENCRYPT_ATTR)

        add_element_name_list = entry.get().get_value_list(ADD_PART_XPATH)

        enable_vfs_element_list = self.__get_enable_vfs_element()
        for add_element_name in add_element_name_list:

            if add_element_name in enable_vfs_element_list:
                continue

            enable_vfs_element_list.append(add_element_name)

            try:
                 self.__mount_vfs_element(is_enc, add_element_name,
                                          EXT_RAID_DEV_INFO[add_element_name])
            except self.__AlreadyMount:
                pass
            except:
                self.exception("fail to __mount_vfs_element(%s)" % \
                               add_element_name)

        vfsext_dev_info = {}
        for element in enable_vfs_element_list:
            vfsext_dev_info[element] = EXT_RAID_DEV_INFO[element]

        mount_option_str = self.__get_vfsext_mount_option(vfsext_dev_info)
        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  VFS_REMOUNT_ARG_FORMAT,
                                  VFS_REMOUNT_VAR_INDEX_LIST,
                                  [mount_option_str])
        if 0 != proc.returncode:
            raise self.__MountError
        return

    def __sync(self):

        self.__execute_cmd(False, True, NULL, NULL, [SYNC_CMD])

    def __mount(self, entry):
        """Do mount"""

        is_bind = False
        is_vfs_remove = False
        is_vfs_add = False
        is_vfs_multi_add = False
        if self.__is_exist_node(entry.get(), OPTION_NAME):
            is_bind = \
                entry.get().get_node(OPTION_NAME).is_set_to_true(BIND_ATTR)

            is_vfs_remove = \
                entry.get().get_node(OPTION_NAME).\
                    is_set_to_true(VFSREMOVE_ATTR)

            is_vfs_add = \
                entry.get().get_node(OPTION_NAME).\
                    is_set_to_true(VFSADD_ATTR)

            is_vfs_multi_add = \
                entry.get().get_node(OPTION_NAME).\
                    is_set_to_true(VFS_MULTIPLE_ADD_ATTR)

        if is_bind:
            return self.__bind_mount(entry)

        if is_vfs_remove:
            return self.__vfs_remove(entry)

        if is_vfs_add:
            return self.__vfs_add(entry)

        if is_vfs_multi_add:
            return self.__vfs_multiple_add(entry)

        return self.__standard_mount(entry)

    def __unmount(self, entry):
        """Do unmount"""

        is_bind = False
        if self.__is_exist_node(entry.get(), OPTION_NAME):
            is_bind = \
                entry.get().get_node(OPTION_NAME).is_set_to_true(BIND_ATTR)

        vol_node = entry.get().get_node(VOLUME_NAME)
        mount_dst_path = vol_node.get_node(DST_NAME).get_text()

        try:
            util.get_mount_info(mount_dst_path)
        except NotFoundError:
            raise self.__AlreadyUnmount

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  UNMOUNT_ARG_FORMAT,
                                  UNMOUNT_VAR_INDEX_LIST,
                                  [mount_dst_path])
        if 0 != proc.returncode:
            raise self.__UnmountError

        if is_bind:
            os.removedirs(mount_dst_path)

        return True

    def __convert_swap_part_name(self, swap_part_name):

        if swap_part_name in config.DEV_TO_ENC_DEV.values():
            return os.path.join(ENC_DIR, swap_part_name)

        return swap_part_name

    def __swap_create(self, entry):
        """Create swap."""

        vol_node = entry.get().get_node(VOLUME_NAME)
        part_name = vol_node.get_node(PART_NAME_XPATH).get_text()
        part_name = self.__convert_swap_part_name(part_name)
        part_path = os.path.join(DEV_DIR, part_name)

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  MKSWAP_ARG_FORMAT,
                                  MKSWAP_VAR_INDEX_LIST,
                                  [part_path])
        if 0 != proc.returncode:
            raise self.__CreateError

        return

    def __swap_on(self, entry):
        """Enable Swap partition"""

        vol_node = entry.get().get_node(VOLUME_NAME)
        part_name = vol_node.get_node(PART_NAME_XPATH).get_text()
        part_name = self.__convert_swap_part_name(part_name)
        part_path = os.path.join(DEV_DIR, part_name)

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  SWAPON_ARG_FORMAT,
                                  SWAPON_VAR_INDEX_LIST,
                                  [part_path])
        if 0 != proc.returncode:
            raise self.__StartError

        return

    def __swap_off(self, entry):
        """Disable Swap partition"""

        vol_node = entry.get().get_node(VOLUME_NAME)
        part_name = vol_node.get_node(PART_NAME_XPATH).get_text()
        part_name = self.__convert_swap_part_name(part_name)
        part_path = os.path.join(DEV_DIR, part_name)

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  SWAPOFF_ARG_FORMAT,
                                  SWAPOFF_VAR_INDEX_LIST,
                                  [part_path])
        if 0 != proc.returncode:
            raise self.__StopError

        return

    def __swap_get(self, entry):

        swaplist_node = XmlNode("<partitionlist/>")
        with open(config.SWAPS_FILE) as fh:
            for line in fh.readlines():
                match = re.match("/dev/(\S*)\s", line)
                if match:
                    real_part_name = match.group(1)
                    swaplist_node.append_child(PARTITION_NAME, real_part_name)

        entry.get().append_child(swaplist_node)
        return

    def __enc_start(self, entry):
        """Start encrypt partition"""

        vol_node = entry.get().get_node(VOLUME_NAME)
        key_path = entry.get_file()
        part_name = vol_node.get_node(PART_NAME_XPATH).get_text()
        part_path = util.symbl_dev_name2dev_path(part_name)
        if not os.path.exists(part_path):
            raise self.__DeviceNotFound

        enc_part_name = vol_node.get_node(ENC_NAME_XPATH).get_text()

        proc = self.__execute_cmd(False, True, NULL, NULL,
                                  CRYPT_START_ARG_FORMAT,
                                  CRYPT_SATRT_VAR_INDEX_LIST,
                                  [key_path, enc_part_name, part_path])
        if 0 != proc.returncode:
            self.error("encrypt start fail..")
            raise self.__StartError

    def __enc_stop(self, entry):
        """Stop encrypt partition"""

        vol_node = entry.get().get_node(VOLUME_NAME)
        enc_part_name = vol_node.get_node(ENC_NAME_XPATH).get_text()

        if not os.path.exists(os.path.join(DEV_DIR, ENC_DIR, enc_part_name)):
            raise self.__DeviceNotFound

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL,
                                  CRYPT_STOP_ARG_FORMAT,
                                  CRYPT_STOP_VAR_INDEX_LIST,
                                  [enc_part_name])
        if 0 != proc.returncode:
            self.error("encrypt stop fail..")
            raise self.__StopError

    def __write_enc_sig(self, entry):

        is_clear = False
        try:
            is_clear = \
                entry.get().get_node(OPTION_NAME).is_set_to_true(CLEAR_ATTR)
        except:
            pass

        vol_node = entry.get().get_node(VOLUME_NAME)
        disk_name = vol_node.get_node(DISK_NAME_XPATH).get_text()
        real_disk_name = self.__convert_real_devname(disk_name)
        disk_path = os.path.join(DEV_DIR, real_disk_name)

        info_lba = self.__get_ptable_and_last_lba(real_disk_name) + 1
        with open(disk_path, "w") as disk_file:
            # FIXME sectorサイズ決めうち
            disk_file.seek(info_lba * 512)

            # FIXME ちゃんと構造体使って管理すべき
            if is_clear:
                disk_file.write(" " * len(config.ENCRYPT_SIGNATURE))
            else:
                disk_file.write(config.ENCRYPT_SIGNATURE)
        return

    def __check_enc_sig(self, entry):

        vol_node = entry.get().get_node(VOLUME_NAME)
        disk_name = vol_node.get_node(DISK_NAME_XPATH).get_text()
        if not self.__is_encrypted_volume(disk_name):
            raise self.__SignatureNotFound
        return

    def __check_file_exist(self, entry):

        # <disk><name>usbd3<name><disk>
        # lsの出力が空ならファイルが存在しないとみなす
        # (隠しファイルは無視)
        # Windows 8.1 でフォーマットすると
        # 「System Volume Information」フォルダーが
        # できるので、これも無視する

        TMP_CHK_FILE_EXIST_DIR = "/var/tmp/checkfileexist"

        vol_node = entry.get().get_node(VOLUME_NAME)
        disk_name = vol_node.get_node(DISK_NAME_XPATH).get_text()

        if not os.path.exists(os.path.join(DEV_DIR, disk_name)):
            raise self.__DeviceNotFound

        mount_dst_path = TMP_CHK_FILE_EXIST_DIR

        if not os.path.isdir(mount_dst_path):
            os.makedirs(mount_dst_path)

        part_name_list = []
        # 暗号化パーティション
        part_name_list.append(
            os.path.join(ENC_DIR, DEV_TO_ENC_DEV[disk_name]))

        # Super floppyか内蔵RAIDドライブ
        part_name_list.append(disk_name)

        # USBディスクの場合はusbdxp1~p15まで
        # 使えるパーティションをサーチ
        for part_num in range(1, 16):
            part_name_list.append(disk_name + "p" + str(part_num))

        for part_name in part_name_list:

            part_path = os.path.join(DEV_DIR, part_name)

            if not os.path.exists(part_path):
                continue

            part_path = os.path.realpath(part_path)

            try:
                (dev, cur_mnt_point, fs, opts) = \
                    util.get_mount_info(part_path)
                if cur_mnt_point:
                    proc = self.__execute_cmd(False, False, PIPE, NULL,
                                              [LS_CMD, cur_mnt_point])
                    if "" == proc.stdout.read():
                        raise self.__FileNotExist
                    return
            except NotFoundError:
                pass

            try:
                self.__simple_mount(part_path, mount_dst_path)
            except:
                continue

            proc = self.__execute_cmd(False, False, PIPE, NULL,
                                      [LS_CMD, mount_dst_path])
            ls_output = proc.stdout.read()
            is_output_empty = False

            if ls_output == "":
                is_output_empty = True
            if ls_output == "System Volume Information\n":
                is_output_empty = True

            try:
                self.__simple_unmount(mount_dst_path)
            except:
                pass

            if is_output_empty:
                raise self.__FileNotExist

            return

        raise self.__DeviceNotFound

    def __get_disk_label(self, entry):

        DISK_LABEL_PART = "p4"
        TMP_DISK_LABEL_DIR = "/var/tmp/disklabel"
        DISK_LABEL_FILE_LIST = \
            ["errorcount", "magic", "md5sum", "mindisk",
             "raidmode", "usage", "version", "model", "default_swap_size"]
        DISK_LABEL_NAME = "disklabel"

        vol_node = entry.get().get_node(VOLUME_NAME)
        disk_name = vol_node.get_node(DISK_NAME_XPATH).get_text()

        if not os.path.exists(os.path.join(DEV_DIR, disk_name)):
            raise self.__DeviceNotFound

        if not os.path.isdir(TMP_DISK_LABEL_DIR):
            os.makedirs(TMP_DISK_LABEL_DIR)

        part_path = os.path.join(DEV_DIR, disk_name + DISK_LABEL_PART)
        if not os.path.exists(part_path):
            raise self.__DeviceNotFound
        part_path = os.path.realpath(part_path)

        self.__simple_mount(part_path,
                            TMP_DISK_LABEL_DIR)

        try:
            if not os.path.isdir(
                os.path.join(TMP_DISK_LABEL_DIR, "disklabel")):
                raise self.__DeviceNotFound

            vol_node.append_child(DISK_LABEL_NAME)
            disk_label_node = vol_node.get_node(DISK_LABEL_NAME)
            for file_name in DISK_LABEL_FILE_LIST:
                file_path = os.path.join(TMP_DISK_LABEL_DIR,
                                         "disklabel",
                                         file_name)
                try:
                    with open(file_path) as fh:
                        disk_label_node.append_child(file_name,
                                                     fh.read().rstrip())
                except:
                    continue
        finally:
            try:
                self.__simple_unmount(TMP_DISK_LABEL_DIR)
            except:
                self.exception("fail to __simple_mount")
        return

    def __set_disk_label(self, entry):

        DISK_LABEL_PART = "p4"
        TMP_DISK_LABEL_DIR = "/var/tmp/disklabel"
        DISK_LABEL_FILE_LIST = \
            ["errorcount", "magic", "md5sum", "mindisk",
             "raidmode", "usage", "version", "default_swap_size"]
        DISK_LABEL_NAME = "disklabel"

        is_inc_err = False
        if entry.get().get_node(OPTION_NAME).is_set_to_true(INCERRCNT_ATTR):
            is_inc_err = True

        vol_node = entry.get().get_node(VOLUME_NAME)
        disk_name = vol_node.get_node(DISK_NAME_XPATH).get_text()

        if not os.path.exists(os.path.join(DEV_DIR, disk_name)):
            raise self.__DeviceNotFound

        if not os.path.isdir(TMP_DISK_LABEL_DIR):
            os.makedirs(TMP_DISK_LABEL_DIR)

        part_path = os.path.join(DEV_DIR, disk_name + DISK_LABEL_PART)
        if not os.path.exists(part_path):
            raise self.__DeviceNotFound
        part_path = os.path.realpath(part_path)

        self.__simple_mount(part_path,
                            TMP_DISK_LABEL_DIR)

        try:
            if not os.path.isdir(
                os.path.join(TMP_DISK_LABEL_DIR, "disklabel")):
                raise self.__DeviceNotFound

            if is_inc_err:
                file_path = os.path.join(TMP_DISK_LABEL_DIR,
                                         "disklabel",
                                         "errorcount")
                with open(file_path, "r") as fh:
                    errcount = int(fh.read().rstrip()) + 1

                with open(file_path, "w") as fh:
                    fh.write(str(errcount))

                try:
                    self.__simple_unmount(TMP_DISK_LABEL_DIR)
                except:
                    self.exception("fail to __simple_mount")
                return

            vol_node.append_child(DISK_LABEL_NAME)
            disk_label_node = vol_node.get_node(DISK_LABEL_NAME)

            for file_name in DISK_LABEL_FILE_LIST:

                if not self.__is_exist_node(disk_label_node, file_name):
                    continue

                value = disk_label_node.get_value(file_name)

                file_path = os.path.join(TMP_DISK_LABEL_DIR,
                                         "disklabel",
                                         file_name)
                try:
                    with open(file_path, "w") as fh:
                        fh.write(value)
                except:
                    continue
        finally:
            try:
                self.__simple_unmount(TMP_DISK_LABEL_DIR)
            except:
                self.exception("fail to __simple_mount")

        return

    def __get_vfsext_element_info(self, entry):

        def unmount():

            if not is_already_mount:
                try:
                    self.__simple_unmount(mnt_point)
                except:
                    self.exception("fail to __simple_unmount(%s)" \
                                       % mnt_point)

        PARTITION_NAME_XPATH = "./volume/name[@type='partition']"
        UPDATE_FILE_RE = "^([0-9]+:[0-9]+),([0-9a-zA-Z]+)"
        info_table = {TIMESTAMP_NAME: "",
                      UUID_NAME: "",
                      MIRR_SET_NAME: ""}

        is_enc = \
            entry.get().get_node(VOLUME_NAME).is_set_to_true(ENCRYPT_ATTR)
        part_name = entry.get().get_node(PARTITION_NAME_XPATH).get_text()

        is_already_mount = False
        try:
            self.__mount_vfs_element(is_enc,
                                     part_name,
                                     EXT_RAID_DEV_INFO[part_name])
        except self.__AlreadyMount:
                is_already_mount = True
        except:
            pass

        mnt_point = EXT_RAID_DEV_INFO[part_name]["mnt_point"]
        update_file_path = os.path.join(mnt_point, config.VFSEXT_UPDATE_FILE)

        if not os.path.isfile(update_file_path):
            entry.set_result(Results.FILE_NOT_FOUND)
            unmount()
            raise self.__FileNotExist

        info_table[TIMESTAMP_NAME] = \
            str(int(os.stat(update_file_path).st_mtime))

        with open(update_file_path) as fh:
            update_info_str = fh.read().rstrip()

        unmount()

        match = re.match(UPDATE_FILE_RE, update_info_str)
        if not match:
            entry.set_result(Results.FILE_NOT_FOUND)
            raise self.__FileNotExist

        info_table[MIRR_SET_NAME] = match.group(1)
        info_table[UUID_NAME] = match.group(2)

        for tag, value in info_table.iteritems():
            entry.get().get_node(VOLUME_NAME).append_child(tag, value)

        return

    def __run_mkfs(self, real_part_name, filesystem):
        """Execute mkfs command."""

        part_path = os.path.join(config.DEV_DIR, real_part_name)

        if EXT3_FSTYPE == filesystem:
            return self.__execute_cmd(True, True, NULL, NULL,
                                      MKFS_EXT3_ARG_FORMAT,
                                      MKFS_EXT3_VAR_INDEX_LIST,
                                      [part_path])

        if EXT4_FSTYPE == filesystem:
            return self.__execute_cmd(True, True, NULL, NULL,
                                      MKFS_EXT4_ARG_FORMAT,
                                      MKFS_EXT4_VAR_INDEX_LIST,
                                      [part_path])

        if XFS_FSTYPE == filesystem:
            return self.__execute_cmd(True, True, NULL, NULL,
                                      MKFS_XFS_ARG_FORMAT,
                                      MKFS_XFS_VAR_INDEX_LIST,
                                      [part_path])

        if FAT_FSTYPE == filesystem:
            return self.__execute_cmd(True, True, NULL, NULL,
                                      MKFS_FAT_ARG_FORMAT,
                                      MKFS_FAT_VAR_INDEX_LIST,
                                      [part_path])

        return None

    def __is_vfs_query(self, entry):

        try:
            entry.get().get_node("volume[filesystem='vfsext']")
        except:
            return False

        return True

    def __append_vfs_query(self, entry):

        def get_vfs_sync_progress():

            NO_SYNC = "none"

            if not os.path.isfile(VFS_SYNC_PROGRESS):
                return ""

            with open(VFS_SYNC_PROGRESS, "r") as fh:
                progress = fh.read()
                if NO_SYNC == progress:
                    return ""

            match = re.match(r"([0-9]+) *\/ *([0-9]+)$", progress)
            if match:
                if 0 == int(match.group(2)):
                    return "0"

                return str((int(match.group(1)) * 100) / int(match.group(2)))

            return ""

        def is_exist_encrypted_partition():

            for element in config.EXT_RAID_DEV_INFO.keys():

                if not os.path.exists(os.path.join(DEV_DIR, element)):
                    continue

                part_path = \
                    os.path.join(DEV_DIR, ENC_DIR, DEV_TO_ENC_DEV[element])

                if os.path.exists(part_path):
                    return True

            return False

        vol_node = entry.get().get_node(VOLUME_NAME)

        if is_exist_encrypted_partition():
            vol_node.get_node("filesystem").set_attrs({ENCRYPT_ATTR: "true"})
        else:
            vol_node.get_node("filesystem").set_attrs({ENCRYPT_ATTR: "false"})

        try:
            (dev, mnt_point, fs, opts) = \
                util.get_mount_info("/mnt/sataraid1")
        except:
            vol_node.append_child("status", "unmounted")
            return

        proc = CmdExec(["df", "--block-size=1",
                        "--portability", mnt_point],
                       stdin=NULL, stdout=PIPE, stderr=NULL,
                       raise_exc=True, close_fds=True).execute()

        info_dict = self.__analysis_df_result(proc.stdout.read())
        capaxml = '<capacity><total>%s</total><use>%s</use></capacity>'
        capaxml = capaxml % (info_dict[TOTAL_SIZE], info_dict[USE_SIZE])
        capa_node = XmlNode(capaxml)

        vol_node = entry.get().get_node("volume")
        vol_node.append_child(capa_node)
        vol_node.append_child("status", "mounted")
        vol_node.append_child("mountdst", mnt_point)

        progress = ""
        try:
            progress = get_vfs_sync_progress()
        except:
            self.exception("fail to get_vfs_sync_progress()")

        if progress:
            vol_node.append_child("progress", progress)

        element_list = self.__get_enable_vfs_element()

        partlist_node = XmlNode(PARTITIONLIST_NAME)
        for element in element_list:
            partlist_node.append_child(PARTITION_NAME, element)
        vol_node.append_child(partlist_node)
        return

    def __make_non_raid_partition(self, src_disk_name, dst_disk_name,
                                  is_external_vfsext=False):

        TMP_PART1_DIR_SRC = "/var/tmp/noraid/" + dst_disk_name + "p1" + "/src"
        TMP_PART1_DIR_DST = "/var/tmp/noraid/" + dst_disk_name + "p1" + "/dst"
        TMP_PART4_DIR_SRC = "/var/tmp/noraid/" + src_disk_name + "p4" + "/src"
        TMP_PART4_DIR_DST = "/var/tmp/noraid/" + dst_disk_name + "p4" + "/dst"

        def mount_partition(fstype, src, dst):

            if not os.path.isdir(dst):
                os.makedirs(dst)

            proc = self.__execute_cmd(False, True,
                                      NULL, NULL,
                                      MOUNT_NON_OPTION_ARG_FORMAT,
                                      MOUNT_NON_OPTION_VAR_INDEX_LIST,
                                      [fstype, src, dst])
            if 0 != proc.returncode:
                raise self.__MountError

        def unmount_partition(mnt_point):

            proc = self.__execute_cmd(False, True,
                                      NULL, NULL,
                                      UNMOUNT_ARG_FORMAT,
                                      UNMOUNT_VAR_INDEX_LIST,
                                      [mnt_point])
            if 0 != proc.returncode:
                raise self.__UnmountError

        def make_partition_1():

            real_part_name = \
                self.__convert_real_devname(dst_disk_name + "p1")
            self.__run_mkfs(real_part_name, config.PART1_MKFS_TYPE)

            src_part_path = \
                os.path.join(config.DEV_DIR, src_disk_name + "p1")

            dst_part_path = \
                os.path.join(config.DEV_DIR, dst_disk_name + "p1")

            mount_partition(
                config.PART1_MOUNT_TYPE, src_part_path, TMP_PART1_DIR_SRC)
            mount_partition(
                config.PART1_MOUNT_TYPE, dst_part_path, TMP_PART1_DIR_DST)

            try:
                self.__copy_all_files(TMP_PART1_DIR_SRC, TMP_PART1_DIR_DST)
            except:
                unmount_partition(TMP_PART1_DIR_SRC)
                unmount_partition(TMP_PART1_DIR_DST)
                raise self.__MakeError

            unmount_partition(TMP_PART1_DIR_SRC)
            unmount_partition(TMP_PART1_DIR_DST)

        def make_partition_4():

            real_part_name = \
                self.__convert_real_devname(dst_disk_name + "p4")
            self.__run_mkfs(real_part_name, EXT4_FSTYPE)

            src_part_path = \
                os.path.join(config.DEV_DIR, src_disk_name + "p4")

            dst_part_path = \
                os.path.join(config.DEV_DIR, dst_disk_name + "p4")

            mount_partition("ext4", src_part_path, TMP_PART4_DIR_SRC)
            mount_partition("ext4", dst_part_path, TMP_PART4_DIR_DST)

            try:
                self.__copy_all_files(TMP_PART4_DIR_SRC, TMP_PART4_DIR_DST)
            except:
                unmount_partition(TMP_PART4_DIR_SRC)
                unmount_partition(TMP_PART4_DIR_DST)
                raise self.__MakeError

            unmount_partition(TMP_PART4_DIR_SRC)
            unmount_partition(TMP_PART4_DIR_DST)

        def make_partition_3():

            real_part_name = \
                self.__convert_real_devname(dst_disk_name + "p3")

            part_path = os.path.join(config.DEV_DIR, real_part_name)
            proc = self.__execute_cmd(False, True,
                                      NULL, NULL,
                                      MKSWAP_ARG_FORMAT,
                                      MKSWAP_VAR_INDEX_LIST,
                                      [part_path])
            if 0 != proc.returncode:
                raise self.__CreateError

            proc = self.__execute_cmd(False, True,
                                      NULL, NULL,
                                      SWAPON_ARG_FORMAT,
                                      SWAPON_VAR_INDEX_LIST,
                                      [part_path])
            if 0 != proc.returncode:
                raise self.__StartError

        if is_external_vfsext:
            make_partition_4()
        else:
            make_partition_1()
            make_partition_3()
            make_partition_4()
        return

    #
    # Request handlers
    #

    def __query_handler(self, entry):
        """Handle "query" request."""

        if self.__is_vfs_query(entry):
            self.__append_vfs_query(entry)
            entry.set_result(Results.OK)
            return entry

        self.__validate_root_tag_name(entry)
        vol_node = entry.get().get_node(VOLUME_NAME)

        disk_name = vol_node.get_value(DISK_NAME_XPATH, None)
        if disk_name:
            try:
                self.__append_usable_partition_info(vol_node)
                entry.set_result(Results.OK)
            except self.__DeviceNotFound:
                entry.set_result(Results.DEVICE_NOT_FOUND)
        else:
            try:
                self.__append_partition_info(vol_node)
                entry.set_result(Results.OK)
            except self.__DeviceNotFound:
                entry.set_result(Results.DEVICE_NOT_FOUND)

        return entry

    def __control_handler(self, entry):
        """Handle "control" request."""

        self.__validate_root_tag_name(entry)

        ctrl_funcs = {SWAP_CREATE_ATTR_VAL: self.__swap_create,
                      SWAP_ON_ATTR_VAL: self.__swap_on,
                      SWAP_OFF_ATTR_VAL: self.__swap_off,
                      SWAP_GET_ATTR_VAL: self.__swap_get,
                      ENC_START_ATTR_VAL: self.__enc_start,
                      ENC_STOP_ATTR_VAL: self.__enc_stop,
                      WRITE_ENC_SIG_ATTR_VAL: self.__write_enc_sig,
                      CHK_ENC_SIG_ATTR_VAL: self.__check_enc_sig,
                      CHK_FILE_EXIST_ATTR_VAL: self.__check_file_exist,
                      GET_DISK_LABEKL_ATTR_VAL: self.__get_disk_label,
                      SET_DISK_LABEKL_ATTR_VAL: self.__set_disk_label,
                      VFSEXT_ELEMENT_INFO_ATTR_VAL: \
                          self.__get_vfsext_element_info, }

        op = entry.get_options()[OPERATION_OPT]

        try:
            ctrl_funcs[op](entry)
            entry.set_result(Results.OK)
        except self.__CreateError:
            entry.set_result(Results.CREATE_ERROR)
        except self.__StartError:
            entry.set_result(Results.START_ERROR)
        except self.__StopError:
            entry.set_result(Results.STOP_ERROR)
        except self.__OnError:
            entry.set_result(Results.ENABLE_ERROR)
        except self.__OffError:
            entry.set_result(Results.DISABLE_ERROR)
        except self.__DeviceNotFound:
            entry.set_result(Results.DEVICE_NOT_FOUND)
        except self.__SignatureNotFound:
            entry.set_result(Results.CHECK_ERROR)
        except self.__FileNotExist:
            entry.set_result(Results.FILE_NOT_FOUND)
        except:
            raise

        return entry

    def __try_mount_with_retry(self, entry):

        for count in range(1, 4):
            try:
                self.__sync()
                self.__mount(entry)
            except self.__MountError:
                time.sleep(1)
                continue
            except:
                raise
            else:
                return

        raise self.__MountError

    def __mount_handler(self, entry):
        """Handle "mount" request."""

        self.__validate_root_tag_name(entry)

        try:
            with RemountRoot():
                with fileut.FLock(LOCK_DIR, MOUNT_LOCK_FILE, True):
                    self.__try_mount_with_retry(entry)
            entry.set_result(Results.OK)
        except self.__MountError:
            entry.set_result(Results.MOUNT_ERROR)
        except self.__AlreadyMount:
            entry.set_result(Results.ALREADY_MOUNT)
        except:
            raise

        return entry

    def __try_unmount_with_retry(self, entry):

        for count in range(1, 4):
            try:
                self.__sync()
                self.__unmount(entry)
            except self.__UnmountError:
                time.sleep(1)
                continue
            except:
                raise
            else:
                return

        raise self.__UnmountError

    def __unmount_handler(self, entry):
        """Handle "unmount" request."""

        self.__validate_root_tag_name(entry)

        try:
            with RemountRoot():
                self.__try_unmount_with_retry(entry)
            entry.set_result(Results.OK)
        except self.__UnmountError:
            entry.set_result(Results.UNMOUNT_ERROR)
        except self.__AlreadyUnmount:
            entry.set_result(Results.ALREADY_UNMOUNT)

        return entry

    def __mkfs_handler(self, entry):
        """Handle "mkfs" request."""

        self.__validate_root_tag_name(entry)

        is_wait = False
        is_mknonraid = False
        if self.__is_exist_node(entry.get(), OPTION_NAME):
            is_wait = \
                entry.get().get_node(OPTION_NAME).is_set_to_true(WAIT_ATTR)

            is_mknonraid = \
                entry.get().get_node(OPTION_NAME).\
                    is_set_to_true(MKNONRAID_ATTR)

        if is_mknonraid:

            SRC_DISK_NAME_XPATH = "./src/disk/name"
            DST_DISK_NAME_XPATH = "./dst/disk/name"

            is_external_vfsext = \
                 entry.get().get_node(OPTION_NAME).is_set_to_true(
                     EXTERNAL_VFSEXT_ATTR)

            src_disk_name = \
                entry.get().get_node(SRC_DISK_NAME_XPATH).get_text()

            dst_disk_name = \
                entry.get().get_node(DST_DISK_NAME_XPATH).get_text()

            self.__make_non_raid_partition(src_disk_name,
                                           dst_disk_name,
                                           is_external_vfsext)
            entry.set_result(Results.OK)
            return entry

        if is_wait:
            # waitモードでmkfs（albg使わない）
            PARTITION_NAME_XPATH = "./volume/name[@type='partition']"
            FILESYSTEM_XPATH = "./volume/filesystem"
            filesystem = entry.get().get_node(FILESYSTEM_XPATH).get_text()
            part_name = entry.get().get_node(PARTITION_NAME_XPATH).get_text()
            real_part_name = self.__convert_real_devname(part_name)

            proc = self.__run_mkfs(real_part_name, filesystem)
            if proc.wait() != 0:
                entry.set_result(Results.MKFS_ERROR)
            else:
                entry.set_result(Results.OK)
            return entry

        return self.send_entry(entry, ALBG_MKFS_PATH)

    def __fsck_handler(self, entry):
        """Handle "fsck" request."""

        self.__validate_root_tag_name(entry)
        return self.send_entry(entry, ALBG_FSCK_PATH)

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.CONTROL, self.__control_handler)
        self.set_handler(ReqTypes.MOUNT, self.__mount_handler)
        self.set_handler(ReqTypes.UNMOUNT, self.__unmount_handler)
        self.set_handler(ReqTypes.MKFS, self.__mkfs_handler)
        self.set_handler(ReqTypes.FSCK, self.__fsck_handler)

#
# Main
#

if __name__ == "__main__":
    PalVolume(option=palcmd.PalCmdOption()).handle_request()
