#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Winbind."""


import os
import shutil

import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)
from iobase.cmdexec import (CmdExec, PIPE, NULL)
from iobase.xmldata import (XmlNode)
from iobase.exception import (NotFoundError, InvalidArgError, ExecError)


import pal_config as config

TO_NULL_OPTS = {"stdout": NULL, "stderr": NULL}

MSNETWORK_NAME = "msnetwork"
ASSIGN_NAME = "assign"
NAME_NAME = "name"
USER_NAME = "user"
GROUP_NAME = "group"

WG_PATH = "assign/wg"
WORKGROUP_PATH = WG_PATH + "/workgroup"
NT_PATH = "assign/nt"
NT_DOMAIN_PATH = NT_PATH + "/domain"
NT_DCNAME_PATH = NT_PATH + "/dcname"
AD_PATH = "assign/ad"
NETBIOS_PATH = AD_PATH + "/netbios"
FULLDNS_PATH = AD_PATH + "/fulldns"
AD_DCNAME_PATH = AD_PATH + "/dcname"
AD_ADMINNAME_PATH = AD_PATH + "/adminname"
AD_ADMINPASSWD_PATH = AD_PATH + "/adminpasswd"
WINSSRV_PATH = "assign/winssrv"
NAME_PATH = "user/name"
PASSWD_PATH = "user/passwd"

METHOD_ATTR = "method"

SECURITY_KEY = "security"
WORKGROUP_KEY = "workgroup"
PASSWORD_SERVER_KEY = "password server"
WINBIND_RPC_ONLY_KEY = "winbind rpc only"
REALM_KEY = "realm"
WINS_SERVER_KEY = "wins server"
FULLDNS_KEY = "fulldns"
DOMAIN_CONTROLER_KEY = "domain_controler"
DOMAIN_NAME_KEY = "domain_name"

YES_STR = "yes"
USER_STR = "user"
DOMAIN_STR = "domain"
ADS_STR = "ads"

WRITE_MODE = "w"

SMB_CONF_ITEM_FMT_STR = "\t%s = %s\n"
KRB5_CONF_FMT_STR = \
        "[libdefaults]\n" + \
        "\tdefault_realm = %(" + FULLDNS_KEY + ")s\n" + \
        "\n" + \
        "[realms]\n" + \
        "\t%(" + FULLDNS_KEY + ")s = {\n" + \
        "\t\tkdc = %(" + DOMAIN_CONTROLER_KEY + ")s\n" + \
        "\t}\n" + \
        "\n" + \
        "[domain_realms]\n" + \
        "\t.%(" + DOMAIN_NAME_KEY + ")s = %(" + FULLDNS_KEY + ")s\n" + \
        "\t%(" + DOMAIN_NAME_KEY + ")s  = %(" + FULLDNS_KEY + ")s\n"
ADMIN_INFO_FMT_STR = "%s%%%s"

FILE_NAME_SEPARATOR = "."
DOMAIN_SEPARATOR = "."
DOMAIN_DETAIL_SEPARATOR = ":"

WG = "wg"
NT = "nt"
AD = "ad"

NT_OLDJOIN_ARGS = [config.NET_CMD, "rpc", "oldjoin"]
NT_TESTJOIN_ARGS = [config.NET_CMD, "rpc", "testjoin"]
AD_JOIN_ARGS = [config.NET_CMD, "ads", "join",
        "-S", "dc_name", "-U", "admin_info"]
DOMAIN_USER_LIST_SHOW_ARGS = [config.WBINFO_CMD, "-u"]
DOMAIN_GROUP_LIST_SHOW_ARGS = [config.WBINFO_CMD, "-g"]
DOMAIN_USER_DETAIL_SHOW_ARGS = [config.WBINFO_CMD, "-i"]
DOMAIN_GROUP_DETAIL_SHOW_ARGS = [config.WBINFO_CMD, "--group-info"]
DOMAIN_CHECK_ARGS = [config.WBINFO_CMD, "-t"]
DOMAIN_AUTH_ARGS = [config.WBINFO_CMD, "-a", ""]

DC_NAME_OPT_IDX = 4
ADMIN_INFO_OPT_IDX = 6
DOMAIN_USER_FULLNAME_IDX = 4
DOMAIN_USER_GID_IDX = 3
DOMAIN_GROUP_GID_IDX = 2
INFO_IDX = 2

ENV_LANG = "LANG"
ENV_C = "C"

USERLIST_NODE_STR = "<userlist/>"
USER_NODE_STR = "<user><name/></user>"
GROUPLIST_NODE_STR = "<grouplist/>"
GROUP_NODE_STR = "<group><name/></group>"
MSNETWORK_NODE_STR = "<msnetwork/>"
FULLNAME_NODE_STR = "<fullname/>"
GID_NODE_STR = "<gid/>"

NEW_LINE = "\n"

INFO_SEP = "%"

SUCCESS_CMD_RETURN_CODE = 0


class PalWinbind(palcmd.PalCmd):
    """Pal Winbind command implementation class."""

    def __check_root_node_name(self, entry, name):
        """Root node name check."""

        if entry.get().get_node_name() != name:
            raise InvalidArgError(
                "root_node_name = " + entry.get().get_node_name())

    def __create_smb_conf_file(self, entry):
        """Create smb.conf file."""
        #FIXME ファイル連結処理　ライブラリ使用予定
        pass

    def __get_wg_value_dict_setting(self, entry):
        """Get workgroup value dict set."""

        value_dict = {}
        value_dict[SECURITY_KEY] = USER_STR
        value_dict[WORKGROUP_KEY] = \
                entry.get().get_node(WORKGROUP_PATH).get_text()
        try:
            value_dict[WINS_SERVER_KEY] = \
                    entry.get().get_node(WINSSRV_PATH).get_text()

        # FIXME "wins server = "の場合があるが大丈夫か?
        except NotFoundError:
            pass

        return value_dict

    def __get_nt_value_dict_setting(self, entry):
        """Get ntdomain value dict setting."""

        value_dict = {}
        value_dict[SECURITY_KEY] = DOMAIN_STR
        value_dict[WORKGROUP_KEY] = \
                entry.get().get_node(NT_DOMAIN_PATH).get_text()
        value_dict[PASSWORD_SERVER_KEY] = \
                entry.get().get_node(NT_DCNAME_PATH).get_text()
        value_dict[WINBIND_RPC_ONLY_KEY] = YES_STR

        try:
            value_dict[WINS_SERVER_KEY] = \
                    entry.get().get_node(WINSSRV_PATH).get_text()

        except NotFoundError:
            pass

        return value_dict

    def __get_ad_dc_name(self, entry):
        """Get active directory domain controler name."""

        ad_dcname = entry.get().get_node(AD_DCNAME_PATH).get_text()

        try:
            dc_hostname, dc_domainname = ad_dcname.split(DOMAIN_SEPARATOR, 1)
            """ad_dcname is FQDN. """
            return ad_dcname.lower()

        except ValueError:
            """ad_dcname is short host name. """
            return ad_dcname + DOMAIN_SEPARATOR + \
                    entry.get().get_node(FULLDNS_PATH).get_text().lower()

    def __get_ad_value_dict_setting(self, entry):
        """Get active directory value dict setting."""

        value_dict = {}
        value_dict[SECURITY_KEY] = ADS_STR
        value_dict[WORKGROUP_KEY] = \
                entry.get().get_node(NETBIOS_PATH).get_text()
        value_dict[REALM_KEY] = \
                entry.get().get_node(FULLDNS_PATH).get_text().upper()
        value_dict[PASSWORD_SERVER_KEY] = self.__get_ad_dc_name(entry)

        try:
            value_dict[WINS_SERVER_KEY] = \
                    entry.get().get_node(WINSSRV_PATH).get_text()

        except NotFoundError:
            pass

        return value_dict

    def __get_domain_value_dict_setting(self, entry, method):
        """Get domain value dict setting."""

        if method == WG:
            return self.__get_wg_value_dict_setting(entry)
        if method == NT:
            return self.__get_nt_value_dict_setting(entry)
        if method == AD:
            return self.__get_ad_value_dict_setting(entry)

    def __create_domain_part_file(self, entry, method):
        """Create samba domain part file."""

        tmp_file_name = config.SMB_CONF_PART_GLOBAL_DOMAIN_FILE + \
                FILE_NAME_SEPARATOR + str(os.getpid())
        tmp_file_path = os.path.join(config.TEMP_DIR, tmp_file_name)

        tmp_file = open(tmp_file_path, WRITE_MODE)

        value_dict = self.__get_domain_value_dict_setting(entry, method)

        for key, value in value_dict.items():
            tmp_file.write(SMB_CONF_ITEM_FMT_STR % (key, value))
        tmp_file.write("\n")
        tmp_file.close

        # FIXME 部品ファイル保管場所はTBD
        file_dir = config.SMB_CONF_PART_DIR
        if not os.path.exists(file_dir):
            os.mkdir(file_dir)
        file_name = config.SMB_CONF_PART_GLOBAL_DOMAIN_FILE
        file_path = os.path.join(file_dir, file_name)

        shutil.move(tmp_file_path, file_path)

        return

    def __create_nsswitch_conf_nodomain_file(self):
        """Create nsswtich.conf for nodomain."""
        # FIXME 既存のnsswitch.conf.nodomainを
        # /etc/nsswitch.confへコピーするだけ
        pass

    def __create_nsswitch_conf_domain_file(self):
        """Create nsswtich.conf for domain."""
        # FIXME 既存のnsswitch.conf.domainを
        # /etc/nsswitch.confへコピーするだけ
        pass

    def __delete_krb5_conf_file(self):
        """Delete krb5.conf."""

        # FIXME 既存の/etc/krb5.confを削除するだけ
        pass

    def __get_kdc_name(self, entry):
        """Get kerberos domain controler name."""

        return self.__get_ad_dc_name(entry)

    def __get_domain_name(self, entry):
        """Get domain name."""

        ad_dcname = entry.get().get_node(AD_DCNAME_PATH).get_text()

        try:
            dc_hostname, dc_domainname = ad_dcname.split(DOMAIN_SEPARATOR, 1)
            """ad_dcname is FQDN. """
            return dc_domainname

        except ValueError:
            """ad_dcname is short host name. """
            return entry.get().get_node(FULLDNS_PATH).get_text().lower()

    def __get_krb5_value_dict_setting(self, entry):
        """Krb5 value dict set."""

        value_dict = {}
        value_dict[FULLDNS_KEY] = entry.get().get_node(FULLDNS_PATH).get_text()
        value_dict[DOMAIN_CONTROLER_KEY] = self.__get_kdc_name(entry)
        value_dict[DOMAIN_NAME_KEY] = self.__get_domain_name(entry)

        return value_dict

    def __create_krb5_conf_file(self, entry):
        """Create krb5.conf."""

        value_dict = self.__get_krb5_value_dict_setting(entry)

        krb5_conf_str = KRB5_CONF_FMT_STR % \
                {FULLDNS_KEY: value_dict[FULLDNS_KEY],
                DOMAIN_CONTROLER_KEY: value_dict[DOMAIN_CONTROLER_KEY],
                DOMAIN_NAME_KEY: value_dict[DOMAIN_NAME_KEY]}

        # FIXME 部品ファイルtmp保管場所はTBD
        tmp_file_dir = config.TEMP_DIR
        tmp_file_name = config.KRB5_CONF_FILE + \
                FILE_NAME_SEPARATOR + str(os.getpid())
        tmp_file_path = os.path.join(tmp_file_dir, tmp_file_name)
        tmp_file = open(tmp_file_path, WRITE_MODE)
        tmp_file.write(krb5_conf_str)
        tmp_file.close

        # FIXME 部品ファイル保管場所はTBD
        file_dir = config.REAL_CONF_DIR
        file_name = config.KRB5_CONF_FILE
        file_path = os.path.join(file_dir, file_name)

        shutil.move(tmp_file_path, file_path)

        return

    def __logon_to_nt(self):
        """Logon to NT domain."""

        try:
            CmdExec(NT_OLDJOIN_ARGS, **TO_NULL_OPTS).execute()

        except ExecError:   # FIXME XRはexec関数でやるせいなのか、
                            # 8ビットシフトしている
            CmdExec(NT_TESTJOIN_ARGS, **TO_NULL_OPTS).execute()

    def __logon_to_ad(self, entry):
        """Logon to Active Directory."""

        cmd_args = list(AD_JOIN_ARGS)
        cmd_args[DC_NAME_OPT_IDX] = self.__get_ad_dc_name(entry)
        cmd_args[ADMIN_INFO_OPT_IDX] = ADMIN_INFO_FMT_STR % \
                (entry.get().get_node(AD_ADMINNAME_PATH).get_text(),
                entry.get().get_node(AD_ADMINPASSWD_PATH).get_text())

        CmdExec(cmd_args, **TO_NULL_OPTS).execute()

    def __assign_to_workgroup(self, entry, method):
        """Assign to WorkGroup"""

        self.__create_domain_part_file(entry, method)

        self.__create_nsswitch_conf_nodomain_file()

        self.__delete_krb5_conf_file()

        return entry

    def __assign_to_ntdomain(self, entry, method):
        """Assign to NT Domain."""

        self.__create_domain_part_file(entry, method)
        #FIXME ファイル連結処理　ライブラリ使用予定

        self.__create_nsswitch_conf_domain_file()

        self.__delete_krb5_conf_file()

        self.__logon_to_nt()

        return entry

    def __assign_to_activedirectory(self, entry, method):
        """Assign to Active Directory."""

        self.__create_domain_part_file(entry, method)
        #FIXME ファイル連結処理　ライブラリ使用予定

        self.__create_nsswitch_conf_domain_file()

        self.__create_krb5_conf_file(entry)

        self.__logon_to_ad(entry)

        return entry

    def __get_lang_c_environ(self):
        """Get "LANG=C" environ."""

        environ = os.environ.copy()
        environ[ENV_LANG] = ENV_C

        return environ

    def __pipe_wait(self, proc, cmd_args):
        """Pipe wait."""

        proc.wait()
        if proc.returncode != SUCCESS_CMD_RETURN_CODE:
            raise ExecError(cmd_args, proc.returncode)

    def __get_domain_userlist_node(self):
        """Get domain userlist node."""

        userlist_node = XmlNode(USERLIST_NODE_STR)

        proc = CmdExec(DOMAIN_USER_LIST_SHOW_ARGS, stdout=PIPE, stderr=NULL,
            env=self.__get_lang_c_environ()).execute()

        for line in proc.stdout.readlines():
            user_node = XmlNode(USER_NODE_STR)
            user_node.get_node(NAME_NAME).set_text(line.rstrip(NEW_LINE))
            userlist_node.append_child(user_node)

        self.__pipe_wait(proc, DOMAIN_USER_LIST_SHOW_ARGS)

        return userlist_node

    def __get_domain_grouplist_node(self):
        """Get domain grouplist node."""

        grouplist_node = XmlNode(GROUPLIST_NODE_STR)

        proc = CmdExec(DOMAIN_GROUP_LIST_SHOW_ARGS, stdout=PIPE, stderr=NULL,
            env=self.__get_lang_c_environ()).execute()

        for line in proc.stdout.readlines():
            group_node = XmlNode(GROUP_NODE_STR)
            group_node.get_node(NAME_NAME).set_text(line.rstrip(NEW_LINE))
            grouplist_node.append_child(group_node)

        self.__pipe_wait(proc, DOMAIN_GROUP_LIST_SHOW_ARGS)

        return grouplist_node

    def __get_list(self):
        """Get list."""

        msnetwork_node = XmlNode(MSNETWORK_NODE_STR)
        msnetwork_node.append_child(self.__get_domain_userlist_node())
        msnetwork_node.append_child(self.__get_domain_grouplist_node())

        return msnetwork_node

    def __exec_get_detail_cmd(self, cmd_args):
        """Exec get detail cmd."""

        proc = CmdExec(cmd_args, stdout=PIPE, stderr=NULL,
            env=self.__get_lang_c_environ()).execute()

        self.__pipe_wait(proc, cmd_args)

        return proc.stdout.read().rstrip(NEW_LINE).split(
            DOMAIN_DETAIL_SEPARATOR)

    def __set_domain_user_detail(self, entry):
        """Set domain user detail."""

        node = entry.get().get_node(USER_NAME)

        user_name = node.get_node(NAME_NAME).get_text()
        cmd_args = DOMAIN_USER_DETAIL_SHOW_ARGS + [user_name]
        result = self.__exec_get_detail_cmd(cmd_args)

        fullname_node = XmlNode(FULLNAME_NODE_STR)
        fullname_node.set_text(result[DOMAIN_USER_FULLNAME_IDX])
        gid_node = XmlNode(GID_NODE_STR)
        gid_node.set_text(result[DOMAIN_USER_GID_IDX])

        node.append_child(fullname_node)
        node.append_child(gid_node)

    def __set_domain_group_detail(self, entry):
        """Set domain group detail."""

        node = entry.get().get_node(GROUP_NAME)
        group_name = node.get_node(NAME_NAME).get_text()
        cmd_args = DOMAIN_GROUP_DETAIL_SHOW_ARGS + [group_name]
        result = self.__exec_get_detail_cmd(cmd_args)

        gid_node = XmlNode(GID_NODE_STR)
        gid_node.set_text(result[DOMAIN_GROUP_GID_IDX])

        node.append_child(gid_node)

    def __set_detail(self, entry):
        """Set detail."""

        try:
            self.__set_domain_user_detail(entry)
            entry.set_result(Results.OK)
        except NotFoundError:
            try:
                self.__set_domain_group_detail(entry)
                entry.set_result(Results.OK)
            except NotFoundError:
                raise InvalidArgError(
                    'node struct must be "user/name" or "group/name"')
            except ExecError:
                entry.set_result(Results.DETAIL_ERROR)
                self.exception("Detail fail!!")
        except ExecError:
            entry.set_result(Results.DETAIL_ERROR)
            self.exception("Detail fail!!")

    def __apply_handler(self, entry):
        """Apply handler."""

        try:
            self.__check_root_node_name(entry, MSNETWORK_NAME)

            method = entry.get().\
                    get_node(ASSIGN_NAME).get_attr_value(METHOD_ATTR)

            if method == WG:
                self.__assign_to_workgroup(entry, method)
            elif method == NT:
                self.__assign_to_ntdomain(entry, method)
            elif method == AD:
                self.__assign_to_activedirectory(entry, method)
            else:
                raise InvalidArgError("method = " + method)

            self.__create_smb_conf_file(entry)

            entry.set_result(Results.OK)

        except (IOError, OSError, ExecError):
            entry.set_result(Results.APPLY_ERROR)
            self.exception("Apply fail!!")

        return entry

    def __list_handler(self, entry):
        """List handler."""

        try:
            entry.set(self.__get_list())
            entry.set_result(Results.OK)

        except (IOError, OSError, ExecError):
            entry.set_result(Results.LIST_ERROR)
            self.exception("List fail!!")

        return entry

    def __detail_handler(self, entry):
        """Detail handler."""

        self.__check_root_node_name(entry, MSNETWORK_NAME)
        self.__set_detail(entry)
        return entry

    def __check_handler(self, entry):
        """Check handler."""

        try:
            CmdExec(DOMAIN_CHECK_ARGS, **TO_NULL_OPTS).execute()
            entry.set_result(Results.OK)
        except ExecError:
            entry.set_result(Results.CHECK_ERROR)

        return entry

    def __auth_handler(self, entry):
        """Auth handler."""

        try:
            name = entry.get().get_node(NAME_PATH).get_text()
            passwd = entry.get().get_node(PASSWD_PATH).get_text()
            info = name + INFO_SEP + passwd

            cmd_args = list(DOMAIN_AUTH_ARGS)
            cmd_args[INFO_IDX] = info
            CmdExec(cmd_args, **TO_NULL_OPTS).execute()

            entry.set_result(Results.OK)

        except ExecError:
            entry.set_result(Results.AUTH_ERROR)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.LIST, self.__list_handler)
        self.set_handler(ReqTypes.DETAIL, self.__detail_handler)
        self.set_handler(ReqTypes.CHECK, self.__check_handler)
        self.set_handler(ReqTypes.AUTH, self.__auth_handler)


#
# Main
#

if __name__ == "__main__":
    PalWinbind(option=palcmd.PalCmdOption()).handle_request()
