#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Device Control class module.

Define for the Device Control class.
"""


import os
import sys
import json
import copy

from devctrl_config import (DEFAULT_LANG, DEFAULT_ENCODING)

from optparse import OptionParser


REQUEST_NAME = "request"
RESPONSE_NAME = "response"
LANG_NAME = "lang"
OPTIONS_NAME = "lang"
RESULT_NAME = "result"

CONFIG_NAME = "config"
CLOUD_NAME = "cloud"
ENABLED_NAME = "enabled"

APPLY_REQUEST = "apply"
SAVE_REQUEST = "save"
LOAD_REQUEST = "load"


class InvalidRequestData(Exception):
    """Class for invalid request data error."""

    pass


class RequestData(object):
    """Class for request to the Device Control class."""

    def __contains__(self, key):
        """Check if "key" is in dictionary."""

        return key in self.request_dict

    def __len__(self):
        """Return a length of dictionary."""

        return len(self.request_dict)

    def __getitem__(self, key):
        """Return value for "key"."""

        return self.request_dict[key]

    def __str__(self):
        return json.dumps(self.request_dict,
                            ensure_ascii=False).encode(self.encoding)

    def __init__(self, request_src, encoding=DEFAULT_ENCODING):
        """Constructor.
           "request_src" must be instance of RequestData, str, unicode or
           dict."
        """

        self.encoding = encoding

        if isinstance(request_src, RequestData):
            self.request_dict = copy.deepcopy(request_src.request_dict)
        elif isinstance(request_src, (str, unicode)):
            self.request_dict = json.loads(request_src)
        elif isinstance(request_src, dict):
            self.request_dict = copy.deepcopy(request_src)
        else:
            raise InvalidRequestData("Invalid data type!!")

        if REQUEST_NAME not in self.request_dict:
            raise InvalidRequestData("No request specified!!")

    @property
    def lang(self):
        if LANG_NAME in self.request_dict:
            return self.request_dict[LANG_NAME]
        return DEFAULT_LANG

    @property
    def request(self):
        return self.request_dict[REQUEST_NAME]

    @property
    def options(self):
        if OPTIONS_NAME in self.request_dict:
            return self.request_dict[OPTIONS_NAME]
        return {}


class ResponseData(object):
    """Class for response from the Device Control class."""

    def __contains__(self, key):
        """Check if "key" is in dictionary."""

        return key in self.response_dict

    def __getitem__(self, key):
        """Return value for "key"."""

        return self.response_dict[key]

    def __setitem__(self, key, value):
        """Set value to "key"."""

        self.response_dict[key] = value
        return self.response_dict[key]

    def __str__(self):
        return json.dumps(self.response_dict,
                            ensure_ascii=False).encode(self.encoding)

    def __init__(self, request=None, encoding=DEFAULT_ENCODING):
        """Constructor.
           "response" must be instance of "RequestData".
           If None is specified, then response data is created
           whose "response" value is a empty string.
        """

        self.encoding = encoding

        self.response_dict = {}
        if request is None:
            self.response_dict[RESPONSE_NAME] = ""
            self.response_dict[LANG_NAME] = DEFAULT_LANG
        else:
            self.response_dict[RESPONSE_NAME] = request.request
            self.response_dict[LANG_NAME] = request.lang

    @property
    def lang(self):
        return self.response_dict[LANG_NAME]

    @property
    def response(self):
        return self.response_dict[RESPONSE_NAME]

    @property
    def result(self):
        return self.response_dict[RESULT_NAME]

    @result.setter
    def result(self, result):
        self.response_dict[RESULT_NAME] = result
        return self.response_dict[RESULT_NAME]


class DevCtrlClass(object):
    """Class for the Device Control class."""

    def __init__(self):
        """Constructor."""

        pass

    def _json_load_handler(self, request_data, key_name, json_path):
        """Common handler to load a json data."""

        response_data = ResponseData(request_data)
        try:
            response_data[key_name] = json.load(open(json_path))
            response_data.result = 0
        except:
            response_data.result = 1

        return str(response_data)

    def _json_save_handler(self, request_data, key_name, json_path):
        """Common handler to save a json data."""

        response_data = ResponseData(request_data)
        try:
            json.dump(request_data[key_name], open(json_path, "w"))
            response_data.result = 0
        except:
            response_data.result = 1

        return str(response_data)

    def handle_request(self, request):
        """Handle request to the device control class."""

        raise NotImplementedError
