#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Do something for installing the remotelink3."""

import os

from iobase import xmldata
from iobase import cmdexec
from iobase.cmdexec import (PIPE, NULL)
from iobase.exception import (IoBaseException,)
from ionas.request import (Request, Entry)
from product_conf import (BIN_DIR, PACKAGES_DIR, XML_CONF_DIR, NET_IF_LIST,)
import fal_msginfo
from fal_info import (FAL_NETWORK_PATH, FAL_REMOTELINK3_PATH,
                        FAL_FEATURES_PATH,
                        Results, ReqTypes)
from pkgaddon.remotelink3_config import (PORTGEN_CMD,)

import manualctrl


ENCODING = "UTF-8"
CATEGORYLIST_PATH = "categorylist"
CATEGORY_NAME = "category"
CATEGORY_KEY_PATH = "code"
CATEGORY_FIND_XPATH_FMT = """category[code="%s"]"""

MESSAGEINFOLIST_PATH = "messageinfolist"
MESSAGEINFO_NAME = "messageinfo"
MESSAGEINFO_KEY_PATH = "code"
MESSAGEINFO_FIND_XPATH_FMT = """messageinfo[code="%s"]"""

MESSAGES_FILE = os.path.join(XML_CONF_DIR, "messages.xml")
PACKAGE_MESSAGES_FILE = \
        os.path.join(PACKAGES_DIR, "remotelink3", "messages.xml")
BUILD_MSGINFO_CMD = os.path.join(BIN_DIR, "build_msginfo")

PYC_SUFFIX = ".pyc"
PY_SUFFIX = ".py"

NETWORK_NAME = "network"
NET_IF_MAC_XPATH = """netiflist/netif[@name="%s"]/mac""" % (NET_IF_LIST[0],)

REMOTELINK3_NAME = "remotelink3"
CONNECTION_NAME = "connection"
HTTP_PORT_XPATH = """portlist/port[@type="http"]"""
HTTPS_PORT_XPATH = """portlist/port[@type="https"]"""

PORTGEN_ARGS_BASE = [PORTGEN_CMD, "--port-num", "2"]

FEATURES_REQ_XML = "<features><functionlist>" + \
                "<function>remotelink3</function></functionlist></features>"


class RequestError(IoBaseException):
    """Exception Class for request error."""

    pass


def __send_request(target, type, body="", err_msg=""):
    """Send request entry to "target"."""

    entry = Entry(type, body)
    result_entry = Request(target).request(entry).get_entry()
    if (result_entry.get_result() != Results.OK and
        result_entry.get_result() != Results.FILE_NOT_FOUND):

        raise RequestError("%s - (%s)" % (err_msg, result_entry.get_result()))

    try:
        return result_entry.get()
    except:
        return ""


def __merge_list(new_messages_xml, messages_xml, list_xpath,
        item_name, key_in_item_xpath, find_xpath_fmt):
    """Merge a list in "new_mssages_xml" into "messages_xml"."""

    new_list_node = new_messages_xml.get_node(list_xpath)
    list_node = messages_xml.get_node(list_xpath)

    for item_node in new_list_node.get_node_list(item_name):
        key = item_node.get_value(key_in_item_xpath)
        try:
            cur_node = list_node.get_node(find_xpath_fmt % (key,))
            if (cur_node.compare(item_node, ignore_order=True,
                                    ignore_indent=True)):
                cur_node.update(item_node, override=True)
        except:
            list_node.append_child(item_node.get_clone())


def __merge_categorylist(new_messages_xml, messages_xml):
    """Merge "categorylist" tree in "new_mssages_xml" into "messages_xml"."""

    __merge_list(new_messages_xml, messages_xml, CATEGORYLIST_PATH,
        CATEGORY_NAME, CATEGORY_KEY_PATH, CATEGORY_FIND_XPATH_FMT)


def __merge_messageinfolist(new_messages_xml, messages_xml):
    """Merge "messageinfolist" tree in "new_mssages_xml"
       into "messages_xml".
    """

    __merge_list(new_messages_xml, messages_xml, MESSAGEINFOLIST_PATH,
        MESSAGEINFO_NAME, MESSAGEINFO_KEY_PATH, MESSAGEINFO_FIND_XPATH_FMT)


def __update_msginfo(messages_file):
    """Update a "fal_msginfo" for "message_file"."""

    fal_msginfo_path = fal_msginfo.__file__

    (base, suffix) = os.path.splitext(fal_msginfo_path)
    if suffix == PYC_SUFFIX:
        fal_msginfo_path = base + PY_SUFFIX

    cmd_args = [BUILD_MSGINFO_CMD, messages_file, fal_msginfo_path]
    cmdexec.CmdExec(cmd_args).execute()


def __get_mac_address():
    """Return MAC Address."""

    network_info = __send_request(FAL_NETWORK_PATH, ReqTypes.QUERY,
                                    NETWORK_NAME,
                                    err_msg="Error: call fal_network")
    return network_info.get_node(NET_IF_MAC_XPATH).get_text()


def __get_remotelink3_port_pair(mac_address):
    """Return port number pair for Remote Link 3 connection."""

    cmd_args = PORTGEN_ARGS_BASE + [mac_address]
    proc = cmdexec.CmdExec(cmd_args).execute(stdout=PIPE, stderr=NULL)
    base_port = proc.stdout.read().strip()
    proc.wait()
    return (base_port, str(int(base_port) + 2))


def update_messages(messages_file, package_messages_file):
    """Update "messages_file" by "package_messages_file"."""

    messages_xml = xmldata.XmlConf(messages_file, encoding=ENCODING)
    package_messages_xml = xmldata.XmlConf(package_messages_file,
                                encoding=ENCODING)

    __merge_categorylist(package_messages_xml, messages_xml)
    __merge_messageinfolist(package_messages_xml, messages_xml)
    tmp_path = messages_file + "." + str(os.getpid())
    messages_xml.save(tmp_path)
    os.rename(tmp_path, messages_file)

    __update_msginfo(messages_file)


def update_remotelink3_port_config():
    """Update Remote Link 3 port configuration."""

    try:
        remotelink3_config = __send_request(FAL_REMOTELINK3_PATH,
                                    ReqTypes.LOAD,
                                    err_msg="Error: call fal_remotelink3")

        remotelink3 = xmldata.XmlConfNode(REMOTELINK3_NAME)
        connection = remotelink3_config.get_node(CONNECTION_NAME)

        http_port, https_port = \
                __get_remotelink3_port_pair(__get_mac_address())
        connection.get_node(HTTP_PORT_XPATH).set_text(http_port)
        connection.get_node(HTTPS_PORT_XPATH).set_text(https_port)

        remotelink3.append_child(connection)
        __send_request(FAL_REMOTELINK3_PATH, ReqTypes.APPLY, remotelink3,
                                        err_msg="Error: call fal_remotelink3")
    except:
        # Ignore error
        pass


def add_remotelink3_feature():
    """Add Remote Link 3 feature info."""

    try:
        __send_request(FAL_FEATURES_PATH, ReqTypes.ADD, FEATURES_REQ_XML)
    except:
        # Ignore error
        pass

#
# Main
#

add_remotelink3_feature()
update_messages(MESSAGES_FILE, PACKAGE_MESSAGES_FILE)
update_remotelink3_port_config()

manualctrl.update()
